/**
* crypt.c - Cryptographic operations module
*
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*
* In addition, as a special exception, the copyright holders
* give permission to link the code of this program with the
* OpenSSL toolkit (or with modified versions of OpenSSL that use the
* same license as OpenSSL, version 0.9.8a), and distribute linked
* combinations including the two.  You must obey the GNU General Public License
* in all respects for all of the code used other than OpenSSL.  If you modify
* this file, you may extend this exception to your version of the
* file, but you are not obligated to do so.  If you do not wish to
* do so, delete this exception statement from your version.
*/

#include "crypt.h"
//Convert to big endian form
unsigned char *endian_conv(unsigned char* bytes, int bytelen) {
	unsigned char *retval = NULL;
	int i = 0;
	int j = 0;
	ALLOC_AND_CLEAN(retval, bytelen, unsigned char*);
	for (i = bytelen; i>0; i--) {
		retval[j] = bytes[i - 1];
		j++;
	}
	return retval;
}

//################################################################################
//Generates a pub/priv key pair and returns the PWNC_KEYPAIR structure
//algorithm: PWNC_ALGO_ECC or PWNC_ALGO_RSA
//ECC Supported curves:
//- secp160k1 (koblitz curve - 160 bits)
//-secp192k1 (koblitz curve - 192 bits)
//RSA public exponent is always RSA_F4, so we don't have to include it in the pub/priv key data
PWNC_KEYPAIR* generate_key_pair(unsigned short algorithm, unsigned short bits) {
	PWNC_KEYPAIR *pair = NULL;
	PWNC_PUBKEY *pub = NULL;
	PWNC_PRKEY *priv = NULL;
	unsigned char *pubdata = NULL;
	unsigned char *privdata = NULL;
	int datalen = 0;

#ifdef _ECC_
	//ECC key pair structure
	EC_KEY *eckey = NULL;
#endif
	//RSA key pair structure
	RSA *rsakey = NULL;

	switch (algorithm) {

		case PWNC_ALGO_RSA: {

			rsakey = RSA_new();
			if (rsakey == NULL) {
				return NULL;
			}
			rsakey = RSA_generate_key(bits, RSA_F4, NULL, NULL);
			if (!rsakey) {
				return NULL;
			}

			ALLOC_AND_CLEAN(pair, sizeof(PWNC_KEYPAIR), PWNC_KEYPAIR*);
			pub = (PWNC_PUBKEY*)malloc(sizeof(PWNC_PUBKEY));
			priv = (PWNC_PRKEY*)malloc(sizeof(PWNC_PRKEY));
			pub->algorithm = PWNC_ALGO_RSA;
			pub->bits = bits;
			priv->algorithm = PWNC_ALGO_RSA;
			priv->bits = bits;

			//public key data: modulus
			//private key data: modulus+private exponent
			//we have to extract the above fields from the RSA structure and convert them from BIGNUM to byte arrays
			ALLOC_AND_CLEAN( pubdata, bits/8, unsigned char*); //public key datalen = bits/8
			//convert from bignum to to byte array
			datalen = BN_bn2bin(rsakey->n, pubdata);

			ALLOC_AND_COPY( pub, pubdata, datalen);

			ALLOC_AND_CLEAN( privdata, bits/4, unsigned char*); //+++++++CHECK it ==> bits/4(???)
			//convert from bignum to to byte array
			datalen = BN_bn2bin(rsakey->n, privdata);
			datalen += BN_bn2bin(rsakey->d, privdata + datalen);
			ALLOC_AND_COPY( priv, privdata, datalen);

			RSA_free(rsakey);
			SAFE_FREE(pubdata);
			SAFE_FREE(privdata);

			pair->pubkey = pub;
			pair->privkey = priv;

			return pair;
		}

#ifdef _ECC_
		case PWNC_ALGO_ECC: {

			ALLOC_AND_CLEAN(pair, sizeof(PWNC_KEYPAIR), PWNC_KEYPAIR*);
			pub = (PWNC_PUBKEY*)malloc(sizeof(PWNC_PUBKEY));
			priv = (PWNC_PRKEY*)malloc(sizeof(PWNC_PRKEY));
			pub->algorithm = PWNC_ALGO_ECC;
			pub->bits = bits;
			priv->algorithm = PWNC_ALGO_ECC;
			priv->bits = bits;

			eckey = EC_KEY_new();
			if (eckey == NULL) {
				SAFE_FREE(pub);
				SAFE_FREE(priv);
				SAFE_FREE(pair);
				return NULL;
			}

			EC_GROUP *group = NULL;
			switch (bits) {
				case 160: {
					//secp 160 koblitz curve
					//eckey->group = EC_GROUP_new_by_nid(NID_secp160k1);
					group = EC_GROUP_new_by_curve_name(NID_secp160k1);
					break;
				}
				case 192: {
					//secp192 koblitz curve
					group = EC_GROUP_new_by_curve_name(NID_secp192k1);
					break;
				}
				case 224: {
					//secp 160 koblitz curve
					group = EC_GROUP_new_by_curve_name(NID_secp224k1);
					break;
				}
				case 256: {
					//secp 160 koblitz curve
					group = EC_GROUP_new_by_curve_name(NID_secp256k1);
					break;
				}
				default: {
					break;
				}
			} //endsw (bits)
			if (group == NULL) {
				EC_KEY_free(eckey);
				SAFE_FREE(pub);
				SAFE_FREE(priv);
				SAFE_FREE(pair);
				return NULL;
			}
			EC_KEY_set_group(eckey, group);
			
			if (!EC_KEY_generate_key(eckey)) {
				SAFE_FREE(pub);
				SAFE_FREE(priv);
				SAFE_FREE(pair);
				EC_KEY_free(eckey);
				return NULL;
			}

			//public key data
			ALLOC_AND_CLEAN( pubdata, bits/4+1, unsigned char*); //2*sizeof(coordinate) + 1 (first byte)
			//convert from EC_POINT structure to byte array (not using point compression)
			datalen =  EC_POINT_point2oct(EC_KEY_get0_group(eckey), EC_KEY_get0_public_key(eckey), POINT_CONVERSION_UNCOMPRESSED, pubdata, bits/4+1, NULL);
			ALLOC_AND_COPY( pub, pubdata, datalen);
			//private key data
			ALLOC_AND_CLEAN( privdata, bits/8, unsigned char*); //bytes = bits/8
			datalen = BN_bn2bin(EC_KEY_get0_private_key(eckey), privdata); //convert from BIGNUM --> byte array
			ALLOC_AND_COPY( priv, privdata, datalen);

			EC_KEY_free(eckey);
			SAFE_FREE(pubdata);
			SAFE_FREE(privdata);

			pair->pubkey = pub;
			pair->privkey = priv;

			return pair;
		}
#endif /* _ECC_*/
		default: {
			return NULL;
		}
	}//endsw (algo)

	return NULL;
}

//################################################################################
//converts from internal representation (pwnc_key struct) to openssl specific representation (RSA, or ECKEY)
//returns pointer to openssl-specific struct
//apps that use this function must cast return value to appropriate type
//pointer must be freed by the application
void *struct_to_ossl_key(struct pwnc_key *key, unsigned short is_public) {
	RSA *rsakey;

#ifdef _ECC_
	EC_KEY *eckey;
#endif

	unsigned long pubmod = RSA_F4;
	unsigned char *pubmod_be = NULL;//big endian form of pub mod

	if (!key) {
		return NULL;
	}
	switch (key->algorithm) {
		case PWNC_ALGO_RSA: {
			rsakey = RSA_new();
			if (!rsakey) {
				//error creating key structure
				return NULL;
			}
			//set the public exponent: it is always RSA_F4 (65537)
			pubmod_be = endian_conv((void*)&pubmod, sizeof(unsigned long));
			rsakey->e = BN_bin2bn( pubmod_be, sizeof(unsigned long), NULL);
			SAFE_FREE(pubmod_be);
			if (is_public) {
				//if it is a public key, set the public modulus (n)
				rsakey->n = BN_bin2bn(key->data, key->datalen, NULL);
				rsakey->p = NULL;
				rsakey->q = NULL;
			}
			else {
				//if it is a private key, set the public modulus(n) and private exponent (d)
				rsakey->n = BN_bin2bn(key->data, key->datalen/2, NULL);
				rsakey->d = BN_bin2bn(key->data + key->datalen/2, key->datalen/2, NULL);
				rsakey->p = NULL;
				rsakey->q = NULL;
			}
			return rsakey;
			break;
		}
#ifdef _ECC_
		case PWNC_ALGO_ECC: {
			eckey = EC_KEY_new();
			if (eckey == NULL) {
				//error creating key structure
				return NULL;
			}
			EC_GROUP *group = NULL;
			switch(key->bits) {
				case 160: {
					group = EC_GROUP_new_by_curve_name(NID_secp160k1);
					break;
				}
				case 192: {
					group = EC_GROUP_new_by_curve_name(NID_secp192k1);
					break;
				}
				case 224: {
					group = EC_GROUP_new_by_curve_name(NID_secp224k1);
					break;
				}
				case 256: {
					group = EC_GROUP_new_by_curve_name(NID_secp224k1);
					break;
				}
				default: {
					break;
				}
			} //endsw
			if (group == NULL) {
				//error
				if (eckey) {
					EC_KEY_free(eckey);
				}
				return NULL;
			}
			EC_KEY_set_group(eckey, group);	

			EC_POINT *the_pub_key = EC_POINT_new(EC_KEY_get0_group(eckey));;
			BIGNUM *the_priv_key = NULL;
			if (is_public) {
				EC_POINT_oct2point(EC_KEY_get0_group(eckey), the_pub_key, key->data, key->datalen, NULL);
				EC_KEY_set_public_key(eckey, the_pub_key);
			}
			else {
				EC_KEY_set_public_key(eckey, the_pub_key); //maybe not necessary
				the_priv_key = BN_bin2bn(key->data, key->datalen, NULL);
				EC_KEY_set_private_key(eckey, the_priv_key);
			}
			return (void*)eckey;
			break;
		}
#endif /* _ECC_*/
		default: {
			return NULL;
		}
	}//endsw
	return NULL;
}
//################################################################################

#ifdef _ECC_
//################################################################################
//Converts from pwnc internal ecdsa signature representation to ECDSA_SIG
ECDSA_SIG *struct_to_ossl_ecdsa_sig(PWNC_SIG *signature) {
	ECDSA_SIG *ecdsa_sig = NULL;
	if (signature->algorithm != PWNC_ALGO_ECC) {
		return NULL;
	}
	if ( (signature->bits != 160) && (signature->bits != 192) && (signature->bits != 224) && (signature->bits != 256)) {
		return NULL;
	}
	if (!signature || !(signature->data) ) {
		return NULL;
	}
	ecdsa_sig = ECDSA_SIG_new();
	BN_bin2bn(signature->data, signature->bits/8, ecdsa_sig->r);
	BN_bin2bn(signature->data + signature->bits/8, signature->bits/8, ecdsa_sig->s);

	return ecdsa_sig;
}
//################################################################################
#endif /* _ECC_*/

//################################################################################
//Signs the "message" using the "key"
//Signature operation makes use of SHA-1
//PWNC_SIG* pwnc_sign_msg(struct pwnc_key* key, unsigned char *message, int mlen, int is_public) {
PWNC_SIG* pwnc_sign_msg(struct pwnc_key *key, unsigned char *message, int mlen, int is_public) {
	unsigned char *message_digest = NULL; //sha-1 hash of "message"
	unsigned char *encrypted_digest = NULL; //rsa encrypted message digest
	unsigned char *sigdata = NULL; //ecdsa signature data
	PWNC_SIG *signature = NULL; //signature structure (return value)
	RSA *rsakey = NULL; //returned from struct_to_ossl_key

#ifdef _ECC_
	EC_KEY *eckey = NULL; //returned from struct_to_ossl_key
	ECDSA_SIG *ecdsa_sig = NULL; //ecdsa openssl specific signature data structure
#endif /* _ECC_*/

	int siglen = 0; //length of signature data

	//allocate 20 bytes for the SHA-1 hash (it is always 160bits)
	ALLOC_AND_CLEAN(message_digest, 20, unsigned char*);

	//calculate message(mlen bytes long) SHA-1 hash
	if (!SHA1(message, mlen, message_digest)) {
		fprintf (stderr, "Error generating SHA1 digest\n");
		SAFE_FREE(message_digest);
		return NULL;
	}

	switch (key->algorithm) {
		case PWNC_ALGO_RSA: {
			siglen = key->bits/8;
			ALLOC_AND_CLEAN(encrypted_digest, siglen, unsigned char*);
			rsakey = (RSA*)struct_to_ossl_key(key, is_public);
			if (!rsakey) {
				SAFE_FREE(encrypted_digest);
				SAFE_FREE(message_digest);
				return NULL;
			}
			if (is_public) {
				//encrypt using public key
				siglen = RSA_public_encrypt(20, message_digest, encrypted_digest, rsakey, RSA_PKCS1_PADDING);
			}
			else {
				//encrypt using private key
				siglen = RSA_private_encrypt(20, message_digest, encrypted_digest, rsakey, RSA_PKCS1_PADDING);
			}
			if (siglen == -1) {
				//error generating signature - data encryption
				fprintf(stderr, "pwnc_sign_msg: %s\n", ERR_error_string(ERR_get_error(), NULL));
				SAFE_FREE(encrypted_digest);
				SAFE_FREE(message_digest);
				if (rsakey) {
					RSA_free(rsakey);
				}
				return NULL;
			}
			//put signature inside PWNC_SIG struct
			ALLOC_AND_CLEAN(signature, sizeof(PWNC_SIG), PWNC_SIG*);
			signature->algorithm = PWNC_ALGO_RSA;
			signature->bits = key->bits;
			ALLOC_AND_COPY( signature, encrypted_digest, siglen);
			SAFE_FREE(encrypted_digest);
			SAFE_FREE(message_digest);
			if (rsakey) {
				RSA_free(rsakey);
			}
			return signature;
		}

#ifdef _ECC_
		case PWNC_ALGO_ECC: {
			eckey = (EC_KEY*)struct_to_ossl_key(key, is_public);
			if (!eckey) {
				SAFE_FREE(message_digest);
				return NULL;
			}

			if (! (ecdsa_sig = ECDSA_do_sign(message_digest, 20 , eckey)) ) {
				fprintf(stderr, "pwnc_sign_msg: %s\n", ERR_error_string(ERR_get_error(), NULL));
				SAFE_FREE(message_digest);
				return NULL;
			}

			ALLOC_AND_CLEAN(sigdata, key->bits/4, unsigned char*);
			
			//convert from bignum to byte array...
			BN_bn2bin(ecdsa_sig->r, sigdata);
			BN_bn2bin(ecdsa_sig->s, sigdata + key->bits/8);

			ALLOC_AND_CLEAN(signature, sizeof(PWNC_SIG), PWNC_SIG*);
			signature->algorithm = PWNC_ALGO_ECC;
			signature->bits = key->bits;
			ALLOC_AND_COPY( signature, sigdata, key->bits/4);
			ECDSA_SIG_free(ecdsa_sig);
			SAFE_FREE(message_digest);
			SAFE_FREE(sigdata);
			return signature;
		}
#endif /* _ECC_*/

		default: {
			SAFE_FREE(message_digest);
			return NULL;
		}
	}//endsw
	return NULL;
}

//################################################################################
//Verifies a signature using the key "key"
//First converts from internal representation to openssl specific structs and then carries out signature verification
//returns 0 if the signature could not be verified, -1 on error
//"message" refers to the actual data to be signed
//The function first calcutates message's SHA-1 hash, decrypts the signature and compares the decrypted digest to the message digest
//mlen: message length
int verify_pwnc_signature(PWNC_SIG *signature, struct pwnc_key *key, unsigned char *message, int mlen, int is_public) {
	RSA *rsakey = NULL; //returned from struct_to_ossl_key
#ifdef _ECC_
	EC_KEY *eckey = NULL; //returned from struct_to_ossl_key
	ECDSA_SIG *ecdsa_sig = NULL; //ecdsa openssl specific signature data structure
#endif /* _ECC_*/

	unsigned char *decrypted_digest = NULL; //sha-1 hash of the message, as this was decrypted from signature->data
	unsigned char *message_digest = NULL; //sha-1 hash of the message (calculated here)
	int len = 0;
	int verified = 0;

	//allocate 20 bytes for the SHA-1 hash (it is always 160bits)
	ALLOC_AND_CLEAN(message_digest, 20, unsigned char*);

	//calculate message(mlen bytes long) SHA-1 hash
	if (!SHA1(message, mlen, message_digest)) {
		fprintf(stderr, "Error generating SHA1 digest\n");
		SAFE_FREE(message_digest);
		return -1;
	}

	switch (signature->algorithm) {
		case PWNC_ALGO_RSA: {
			//rsakey = (RSA*)struct_to_ossl_key(key, public);
			rsakey = (RSA*)struct_to_ossl_key(key, is_public);
			if (!rsakey) {
				SAFE_FREE(message_digest);
				return -1;
			}
			//digest is always sha-1 (20bytes)
			ALLOC_AND_CLEAN(decrypted_digest, 20, unsigned char*);
			if (is_public) {
				//printf("siglen: %d\n", signature->datalen);
				len = RSA_public_decrypt(signature->datalen, signature->data, decrypted_digest, rsakey, RSA_PKCS1_PADDING);
			}
			else {
				len = RSA_private_decrypt(signature->datalen, signature->data, decrypted_digest, rsakey, RSA_PKCS1_PADDING);
			}
			if (len == -1) {
				printf("verify_pwnc_signature: %s\n", ERR_error_string(ERR_get_error(), NULL));
				SAFE_FREE(message_digest);
				return -1;
			}
			else {
				//compare digests, if they are the same ==> signature verified ==> return 1
				if (!memcmp(message_digest, decrypted_digest, 20)) {
					verified = 1;
				}
				else {
					verified = 0;
				}
			}

			SAFE_FREE(message_digest);
			SAFE_FREE(decrypted_digest);
			if (rsakey) {
				RSA_free(rsakey);
			}

			return verified;
		}

#ifdef _ECC_
		case PWNC_ALGO_ECC: {
			//convert internal==>openssl format
			ecdsa_sig = struct_to_ossl_ecdsa_sig(signature);
			if (!ecdsa_sig) {
				SAFE_FREE(message_digest);
				return -1;
			}
			//convert key to openssl struct
			eckey = (EC_KEY*)struct_to_ossl_key(key, is_public);
			if (!eckey) {
				if (ecdsa_sig) {
					ECDSA_SIG_free(ecdsa_sig);
				}
				SAFE_FREE(message_digest);
				return -1;
			}

			verified = ECDSA_do_verify(message_digest, 20, ecdsa_sig, eckey);
			return verified; //0: wrong sig, 1: OK, -1:err
		}
#endif /* _ECC_*/

		default: {
			SAFE_FREE(message_digest);
			return -1;
		}
	}
	return -1;
}
//################################################################################

//################################################################################
//serializes the receipt info
//timestamp and value are serialized in BIG ENDIAN form
//returns the serialized receipt info and sets info_len to the length of the serialized rcpt
unsigned char* serialize_receipt_info(PWNC_CERT *cert, PWNC_PUBKEY *provpk, unsigned long timestamp, unsigned long weight, int *info_len) {
	unsigned char *rinfo = NULL;
	int rlen = 0;
	int pos = 0;
	unsigned char *ts = NULL; //big endian timestamp representation
	unsigned char *w = NULL; //big endian weight representation
	//calculate length of serialized receipt info

	rlen = cert->TeamPubKey->datalen + cert->UserPubKey->datalen + cert->TeamSig->datalen + provpk->datalen + 2*sizeof(unsigned long);
	ALLOC_AND_CLEAN(rinfo, rlen, unsigned char*);
	memcpy(rinfo, cert->TeamPubKey->data, cert->TeamPubKey->datalen);
	pos += cert->TeamPubKey->datalen;
	memcpy(rinfo + pos, cert->UserPubKey->data, cert->UserPubKey->datalen);
	pos += cert->UserPubKey->datalen;
	memcpy(rinfo + pos, cert->TeamSig->data, cert->TeamSig->datalen);
	pos += cert->TeamSig->datalen;
	memcpy(rinfo + pos, provpk->data, provpk->datalen);
	pos += provpk->datalen;

	ts = endian_conv( (unsigned char*)&timestamp, sizeof(unsigned long));
	w = endian_conv( (unsigned char*)&weight, sizeof(unsigned long));
	
	memcpy(rinfo + pos, ts, sizeof(unsigned long));
	pos += sizeof(unsigned long);
	memcpy(rinfo + pos, w, sizeof(unsigned long));

	SAFE_FREE(ts);
	SAFE_FREE(w);
	*info_len = rlen;
	return rinfo;
}
//################################################################################

//################################################################################
//Constructs a receipt using the arguments as the receipt contents
//copies these fields into a PWNC_RCPT struct, serializes them, signs them using "key" and adds the signature to the receipt
PWNC_RCPT* construct_receipt(PWNC_CERT *cert, PWNC_PUBKEY *provpk, unsigned long timestamp, unsigned long weight, struct pwnc_key *key, int is_public) {
	PWNC_RCPT *rcpt = NULL;
	PWNC_SIG *rcpt_sig = NULL;
	unsigned char *rinfo = NULL;
	int info_len = 0;
	ALLOC_AND_CLEAN(rcpt, sizeof(PWNC_RCPT), PWNC_RCPT*); //allocates/cleans rcpt mem
	//rcpt = PWNC_RCPT_new();
	rcpt->ConsumerCert = cert;
	rcpt->ProviderPubKey = provpk;
	rcpt->timestamp = timestamp;
	rcpt->weight = weight;
	rinfo = serialize_receipt_info(cert, provpk, timestamp, weight, &info_len); //serializes receipt info
	rcpt_sig = pwnc_sign_msg(key, rinfo, info_len, is_public); //signs receipt using key
	rcpt->RcptSig = rcpt_sig;
	SAFE_FREE(rinfo);
	return rcpt;
}
//################################################################################

//################################################################################
//serializes a certificate's info(team pk + user pk) to be signed
unsigned char* serialize_cert_info(PWNC_PUBKEY *team, PWNC_PUBKEY *user, int *len) {
	unsigned char *cert_info = NULL;
	ALLOC_AND_CLEAN(cert_info, team->datalen + user->datalen, unsigned char*);
	memcpy(cert_info, team->data, team->datalen);
	memcpy(cert_info + team->datalen, user->data, user->datalen);
	*len = team->datalen + user->datalen;
	return cert_info;
}
//################################################################################

//################################################################################
// a new member certificate
//suppose key pair is already generated
PWNC_CERT* create_member_cert(PWNC_KEYPAIR *team_pair, PWNC_PUBKEY *user_pub) {
	PWNC_CERT *cert = NULL;
	unsigned char *cert_info = NULL;
	int len = 0;

	cert = PWNC_CERT_new();
	cert->TeamPubKey->bits = team_pair->pubkey->bits;
	cert->TeamPubKey->algorithm = team_pair->pubkey->algorithm;
	//write team pubkey
	cert->TeamPubKey->datalen = team_pair->pubkey->datalen;
	ALLOC_AND_CLEAN(cert->TeamPubKey->data, team_pair->pubkey->datalen, unsigned char*);
	memcpy(cert->TeamPubKey->data, team_pair->pubkey->data, team_pair->pubkey->datalen);

	//add user pub key
	cert->UserPubKey->bits = user_pub->bits;
	cert->UserPubKey->algorithm = user_pub->algorithm;
	cert->UserPubKey->datalen = user_pub->datalen;
	ALLOC_AND_CLEAN(cert->UserPubKey->data, user_pub->datalen, unsigned char*);
	memcpy(cert->UserPubKey->data, user_pub->data, cert->UserPubKey->datalen);

	//serialize cert info
	cert_info = serialize_cert_info(cert->TeamPubKey, cert->UserPubKey, &len);
	//sign cert info using team's private key
	cert->TeamSig = pwnc_sign_msg(team_pair->privkey, cert_info, len, 0);
	return cert;
}
//################################################################################

//################################################################################
//seed the prng
void init_prng() {
	ERR_load_crypto_strings();
	unsigned char buff[2048];
	int i = 0;
	srand((unsigned int)time(NULL));
	for (i=0;i<2048;i++) {
		buff[i] = (unsigned char)(255.0*rand()/(RAND_MAX+1.0));
	}
	RAND_seed(&buff, 2048);
}
