/**
* threads.c - Thread pool implementation based on the POSIX threads API
*
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "threads.h"

void init_thread_pool(int max_threads) {
	int i = 0;
	pool_size = max_threads;
	thread_pool = (CTHREAD *)malloc(pool_size * sizeof(CTHREAD));
	for (i=0; i<pool_size; i++) {
		thread_pool[i].is_free = 1;
	}
}

//linear search in the thread pool array to find the first thread marked as "free"
int find_first_free_thread() {
	int i = 0;
	for (i=0; i<pool_size; i++) {
		if (thread_pool[i].is_free) {
			return i;
		}
	}
	return -1; //no free threads
}

//this function is called from the main thread
//Starts a new thread
//First looks for the first free thread, and then starts it
void  start_new_thread(void *thread_func, int sockfd) {
	THREAD_INFO *ti = (THREAD_INFO*)malloc(sizeof(THREAD_INFO));
	memset(ti, 0, sizeof(THREAD_INFO));
	int tindex = 0;

	//attempt to lock mutex so that no other thread can be created or exit
	pthread_mutex_lock(&thread_pool_mutex);

	tindex = find_first_free_thread();
	if (tindex == -1) {
		fprintf(stderr, "start_new_thread: No threads free\n");
		close(sockfd);
		pthread_mutex_unlock(&thread_pool_mutex);
		return;
	}
	ti->client_socket = sockfd;
	ti->thread_index = tindex;
	//start new thread
	fprintf(stderr, "start_new_thread: starting new thread %d - %d\n", ti->thread_index, ti->client_socket);
	pthread_create(&thread_pool[tindex].client_thread, NULL, thread_func, (void*)ti);
	thread_pool[tindex].is_free = 0; //mark thread as "used"
	//unlock mutex
	pthread_mutex_unlock(&thread_pool_mutex);
}

//this func is called from within the client thread
//when its about to exit
void exit_client_thread(int thread_index) {
	//update the value of the global "last_thread_exited" variable
	fprintf(stderr, "exit_client_thread: func started\n");
	pthread_mutex_lock(&thread_pool_mutex);
	//mark thread as "free"
	thread_pool[thread_index].is_free = 1;
	fprintf(stderr, "exit_client_thread: Thread %d exiting\n", thread_index);
	//unlock mutex and exit
	pthread_mutex_unlock(&thread_pool_mutex);
	pthread_exit(NULL);
}

//cancels a thread of the thread pool
void cancel_thread(int thread_index) {
	pthread_mutex_lock(&thread_pool_mutex);
	if (!thread_pool[thread_index].is_free) {
		pthread_cancel(thread_pool[thread_index].client_thread);
	}
	pthread_mutex_unlock(&thread_pool_mutex);
}

//blocks sigint, sigterm, sigkill signals
int block_signals() {
	sigset_t sig_mask;
	sigemptyset(&sig_mask);
	sigaddset(&sig_mask, SIGINT);
	sigaddset(&sig_mask, SIGTERM);
	sigaddset(&sig_mask, SIGKILL);
	return pthread_sigmask(SIG_BLOCK, &sig_mask, NULL);
}

void thread_pool_cleanup() {
	if (thread_pool) {
		free(thread_pool);
	}
}
