/**
* conversions.h - Type conversions module 
* (internal <--> network (ascii) representation, message generation/parsing, etc)
*
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef _PWNC_CONVERSIONS_H
#define _PWNC_CONVERSIONS_H

//#ifndef _PWNC_TYPES_H
#include "types.h"
//#define _PWNC_TYPES_H
//#endif

#include <time.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#ifndef B64_LINE_LEN
#define B64_LINE_LEN 64
#endif

/*
The macro below operates on the string pointed to by pos
(which normally is part of a bigger text message)
field: which field of a struct the operations are about (e.g. "rcpt->ConsumerCert.TeamPubKey")
pos: start of the string
crlfpos: start of the crlf sequence that follows the b64 string
raw: pointer to the raw (decoded) data
b64: pointer to the b64-encoded data
rawlen: length of the raw data byte array
tofree: the pointer that has to be freed in case of error (the macro is used is several functions, where different structs have to be freed)
first allocates / cleans the "raw" byte array
then allocates and cleans the "b64" byte array
next, decodes the b64 data, stores the length of the raw data to rawlen and the actual data to "raw"
finally, copies the data to the "data" field of "field" (and adjusts the ->datalen fields, etc accordingly) and frees raw and b64
failed: 1 if something failed, 0 if all ok...this value has  to be checked from the calling function
*/
#define DECODE_AND_COPY(field, pos, crlfpos, raw, b64, rawlen, tofree, failed) \
	ALLOC_AND_CLEAN(raw, crlfpos - pos + 1, unsigned char*); \
	ALLOC_AND_CLEAN(b64, crlfpos - pos + 1, unsigned char*); \
	strncpy(b64, pos, crlfpos - pos); \
	if ( (rawlen = b64_decode(b64, raw)) == -1) { \
		SAFE_FREE(tofree); \
		SAFE_FREE(raw); \
		SAFE_FREE(b64); \
		failed = 1; \
	} \
	else { \
		ALLOC_AND_COPY(field, raw, rawlen); \
		SAFE_FREE(b64); \
		SAFE_FREE(raw); \
		failed = 0; \
	} \

static const char EncTable[]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

unsigned long calculate_b64_size(unsigned long len);
void encode_block(unsigned char in[3], unsigned char out[4], int len);
void decode_block(unsigned char in[4], unsigned char out[3]);
int b64_encode(unsigned char *raw, unsigned char* b64, int datalen);
int b64_decode(unsigned char *b64, unsigned char* raw);

unsigned char* construct_pem_description(unsigned short algo, unsigned short keytype, unsigned short head);
unsigned char *b64_2_pem(unsigned char *data, unsigned short pemtype, unsigned short algo);
unsigned char *pem_2_b64(unsigned char *data);

unsigned char* receipt_2_text(PWNC_RCPT *rcpt);
unsigned char* construct_conn_msg(PWNC_CERT *ClientCert);
unsigned char* construct_rreq(PWNC_PUBKEY *pkey, unsigned long weight);
unsigned char* construct_conn_ack(unsigned long timestamp, unsigned short mode);
int get_prov_pklen(unsigned char *text);
int get_cons_pklen(unsigned char *text);
unsigned char* construct_query(PWNC_PUBKEY* provider, PWNC_PUBKEY* consumer);
unsigned char* construct_qrsp(unsigned short action);
unsigned char* construct_rrsp(unsigned short mcode);

unsigned char *serialize_receipt_enc_data(PWNC_RCPT *rcpt, int *rawlen);
PWNC_RCPT *deserialize_receipt_enc_data(PWNC_RCPT *rcpt, unsigned char *data, int len, unsigned short receipt_algorithm, unsigned short bits);
unsigned char *serialize_cert_enc_data(PWNC_CERT *cert, int *rawlen);
PWNC_CERT *deserialize_cert_enc_data(unsigned char *data, int len, unsigned short algorithm, unsigned short bits);

PWNC_RCPT *text_2_rcpt(unsigned char* text);
PWNC_CERT* parse_conn_msg(unsigned char *text);
int parse_cack_msg(unsigned char *text, unsigned long *timestamp);
PWNC_PUBKEY* parse_rreq(unsigned char* text, unsigned long *weight);
int parse_query_msg(unsigned char *text, PWNC_PUBKEY* provider, PWNC_PUBKEY* consumer);

int parse_qrsp(unsigned char *text);
int parse_rrsp(unsigned char *text);

time_t my_timegm (struct tm *tm);

unsigned char *make_updt_req_data(unsigned long timestamp, int *mlen);
unsigned char *make_updt_end_data();
unsigned char *construct_updt_msg(unsigned char *data, int datalen);
int parse_updt_req(unsigned char *text, unsigned long *timestamp);

#endif /* _PWNC_CONVERSIONS_H */
