/**
* rdb_engine.h - Receipt database module
*
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*
* In addition, as a special exception, the copyright holders
* give permission to link the code of this program with the
* OpenSSL toolkit (or with modified versions of OpenSSL that use the
* same license as OpenSSL, version 0.9.8a), and distribute linked
* combinations including the two.  You must obey the GNU General Public License
* in all respects for all of the code used other than OpenSSL.  If you modify
* this file, you may extend this exception to your version of the
* file, but you are not obligated to do so.  If you do not wish to
* do so, delete this exception statement from your version.
*/

#ifndef _PWNC_RDB_ENGINE_H
#define _PWNC_RDB_ENGINE_H

#include <sys/times.h>

#include "io.h"
#include "crypt.h"
#include "threads.h"
#include "rbtree.h"
#include "fifo.h"
#include "storage.h"

#include <signal.h>
#include <errno.h>
#include <dirent.h>

static int INFINITY = 2147483647;

#ifndef RECEIPT_THRESHOLD
#define RECEIPT_THRESHOLD 15000
#endif

#ifndef TEAM_THRESHOLD
#define TEAM_THRESHOLD 100
#endif

//maximum receipt size (that is, using rsa2048 keys/sigs)
int size_of_rcpt = sizeof(unsigned char)  + 2*sizeof(unsigned short) + 2*sizeof(unsigned long) + 5*256;

//Number of receipts currently stored. This can reach up to RECEIPT_THRESHOLD.
//If receipt threshold has been reached and a new receipt is to be stored,
//then the oldest receipt is removed and the new one is added
static int num_of_receipts_stored = 0;
static int num_of_teams_stored = 0;

//################################################################################################//
//The definitions below are used to wrap in a type-safe manner the callbacks used to control insertion/deletion/etc from a hash table
//They will be used in the hash tables that store team and receipt info
 #define DECLARE_LHASH_HASH_FN(f_name,o_type) \
         unsigned long f_name##_LHASH_HASH(const void *);
 #define IMPLEMENT_LHASH_HASH_FN(f_name,o_type) \
         unsigned long f_name##_LHASH_HASH(const void *arg) { \
                 o_type a = (o_type)arg; \
                 return f_name(a); }
 #define LHASH_HASH_FN(f_name) f_name##_LHASH_HASH

 #define DECLARE_LHASH_COMP_FN(f_name,o_type) \
         int f_name##_LHASH_COMP(const void *, const void *);
 #define IMPLEMENT_LHASH_COMP_FN(f_name,o_type) \
         int f_name##_LHASH_COMP(const void *arg1, const void *arg2) { \
                 o_type a = (o_type)arg1; \
                 o_type b = (o_type)arg2; \
                 return f_name(a,b); }
 #define LHASH_COMP_FN(f_name) f_name##_LHASH_COMP

//generates a hash - djb2 algorithm
//takes a byte array of arbitrary length(datalen) and maps it to an unsigned long
//this func is used to produce a hash for hash table entries
unsigned long GenerateHash(unsigned char *tobehashed, int datalen);
#define RHASH_LEN 20 //sha1
//################################################################################################//


//################################################################################################//
//RECEIPT GRAPH
//This graph will be used for performing maxflow operations to check whether a consumer is to be granted access or not
//The graph is represented by an adjacency list data structure
//This list is composed of head nodes (each team is a head node) where team info are stored. Head nodes are in a singly connected list.
//There is a hash table where pointers to the  heads are stored, for O(1) access
//Each head has a list of its neighbors in the graph. For each receipt, there is an edge in the graph.
//An edge is directed from the consumer to the provider (therefore, heads are consumers)

#define MF_NODE_COLOR_BLACK 10
#define MF_NODE_COLOR_WHITE 20
#define MF_NODE_COLOR_GRAY 30

//this represents a head entry in the adjacency list (represents a team) - corresponds to a service provider
struct adjacency_list_head_t {
	int label;
	int color;
	int height;
	int discharged;
	long excess;
	struct adjacency_list_rcpt_node_t *neighbors;
	struct adjacency_list_rcpt_node_t *last_neighbor;
	struct adjacency_list_head_t *next_head;
#ifdef _USE_HT_
	LHASH *neighbors_ht;
#else
	struct neighbor_matrix_entry_t *neighbors_m;
#endif

#ifdef _GLOBAL_RELABEL_
	unsigned short node_state;
	struct incoming_edge_list_node_t *incoming;
	struct incoming_edge_list_node_t *last_incoming;
#endif

	struct __pwnc_team *team;
};
typedef struct adjacency_list_head_t AL_HEAD;

//represents a node in the list of a head node's neighbors (this corresponds to a receipt's consumer)
struct adjacency_list_rcpt_node_t {
	//RBTNODE *tree_node;
	struct adjacency_list_rcpt_node_t *next;
	struct adjacency_list_rcpt_node_t *prev;
	struct adjacency_list_head_t *myhead; //this is the head that represents this node
	struct adjacency_list_head_t *consumer; //this is the node in the neighbor list of whom this node belongs
	int saturated; //indicates if the edge is saturated (=> doesn't exist the residual graph)
	long weight; //original edge weight
	long residual_weight;
	int is_foo; //true if its not a real edge, but one that was added while running maxflow - foo edges do not have a pointer to the tree (== NULL)

#ifdef _GLOBAL_RELABEL_
	struct incoming_edge_list_node_t *incoming_ptr;
#endif
};
typedef struct adjacency_list_rcpt_node_t AL_NODE;

#ifndef _USE_HT_
typedef struct neighbor_matrix_entry_t {
	AL_NODE *node;
	unsigned short removed;
} NMATRIX_ENTRY;
#endif

#ifdef _GLOBAL_RELABEL_
typedef struct incoming_edge_list_node_t {
	AL_NODE *node;
	struct incoming_edge_list_node_t *next;
	struct incoming_edge_list_node_t *prev;
} INCOMING;

#define INCOMING_HEAD(inc) inc->node->consumer

void print_inverse_graph();

#endif /* _GLOBAL_RELABEL_ */

AL_HEAD *receipt_graph;

void graph_add_team(struct __pwnc_team *team);
void graph_add_rcpt(RBTNODE *tree_node);
AL_NODE *graph_remove_rcpt(AL_NODE *rcpt_node, unsigned long rweight);
//TODO: Delete TEAM
long maxflow_run(AL_HEAD *source, AL_HEAD *sink);
void print_graph();

//callbacks used for storing pointers to a node's neighbors in its hash table (AL_NODE->neighbors_ht)
unsigned long neighbor_hash(const AL_NODE *to_hash);
int neighbor_cmp(const AL_NODE *arg1, const AL_NODE *arg2);
static IMPLEMENT_LHASH_HASH_FN(neighbor_hash, const AL_NODE*);
static IMPLEMENT_LHASH_COMP_FN(neighbor_cmp, const AL_NODE*);

//################################################################################################//
//The following refer to the data structures used for storing receipts
//Receipts are stored in nodes of a red black tree. There are pointers to these receipts stored in a hash table
//This was, it takes O(1) time to retrieve a receipt (e.g. check if a receipt exists)  because the hash table is used. Also, it takes O(logn)
//time to replace the oldest receipt with a new one (it might take less if we have a pointer to the oldest receipt - TODO).
//Range queries take O(n) time
//For each tree node, there is a pointer to and from the correspondent entry in the receipt graph

//hash table where receipts are stored
static LHASH *receipts_ht;
static pthread_mutex_t rcpt_mutex;

//receipt handling callbacks for the HASH TABLE
unsigned long receipt_hash(const PWNC_RCPT *to_hash);
int receipt_cmp(const PWNC_RCPT *arg1, const PWNC_RCPT *arg2);
static IMPLEMENT_LHASH_HASH_FN(receipt_hash, const PWNC_RCPT *);
static IMPLEMENT_LHASH_COMP_FN(receipt_cmp, const PWNC_RCPT *);

//the root of a red black tree storing the receipts
RBTNODE *rcpt_tree;

//this is the "data" field that will be placed in an RBTNODE
//rcpt is a pointer to a receipt
//graph_entry is a pointer to the correspondent graph data structure node (this may be null if there is no such graph
typedef struct rbtree_rcpt_data_t {
	PWNC_RCPT *rcpt;
	//struct receipt_graph_node_t *graph_entry;
	AL_NODE *graph_entry;
	int disk_pos; //position of this record in the file in disk
} RBTREE_RCPT;

//constructs a new RBTREE_RCPT object and returns it
//this is to be placed in an rbtnode->data field
RBTREE_RCPT* RBTREE_RCPT_new(PWNC_RCPT *rcpt, AL_NODE *gnode);

//receipt RB tree CALLBACKS
//returns the key for an RBTNODE storing receipts
void *rcpt_key_func(RBTNODE *node);

//sets key value to an RBTNODE storing receipts
void rcpt_set_key(RBTNODE *node, void *keyval);

//compares (the keys of) two RBTNODEs storing receipts
int rcpt_comp(RBTNODE *node1, RBTNODE *node2);

//constructs a new RBTNODE and returns it
RBTNODE *RBTNODE_new(PWNC_RCPT *key, const void *key_func, const void *set_key_func, const void *comp_func);
//################################################################################################//

//################################################################################################//
//The following refer to the hash table storing team info
// Teams are stored in a hash table in memory
//team struct
typedef struct __pwnc_team {
	PWNC_PUBKEY *team_pk; //team public key
	double credits; //Team credits (for the service it has provided
//	double score; //credits omadas / max credits
	int served; //how many times the team has served
	int consumed; //how many times a team has consumed service
	AL_HEAD *graph_node;
}PWNC_TEAM;

//hash table where teams are stored
static LHASH *teams_ht;
pthread_mutex_t teams_mutex;

//team handling callbacks for the HASH TABLE
unsigned long team_hash(const PWNC_TEAM *to_hash);
int team_cmp(const PWNC_TEAM *arg1, const PWNC_TEAM *arg2);
static IMPLEMENT_LHASH_HASH_FN(team_hash, const PWNC_TEAM *);
static IMPLEMENT_LHASH_COMP_FN(team_cmp, const PWNC_TEAM *);

unsigned char *serialize_rcpt_identifier(const PWNC_RCPT *rcpt, int *len);
//calculates, allocates space for and returns a receipt's SHA-1 hash
unsigned char *rcpt_SHA1(const PWNC_RCPT *rcpt);
//################################################################################################//


//################################################################################################//
//Receipt db operations

//initializes hash tables and mutices
void rdb_init(const char *receipts_db_file, const char *receipts_db_free_list); //done
int insert_receipt(PWNC_RCPT *rcpt, int store_it); //done
int rdb_judge(PWNC_PUBKEY *consumer, PWNC_PUBKEY *provider); //done

//adds a team to the teams hash table
void rdb_add_team(PWNC_PUBKEY *teampk);
//looks up a team in the hash table
//void rdb_team_lookup(PWNC_PUBKEY *teampk);//TODO

int rdb_team_exists(PWNC_PUBKEY *teampk);

//handles a new receipt
//checks receipt validity
//checks receipts_arr load
//adds it in the array
//(adds it in the hash table)
//(adjusts weights)
void handle_new_receipt(PWNC_RCPT *rcpt); //TODO...(mallon de xreiazetai...)

//traverses the tree, and replaces the a tree node ('s data) with the rcpt
//If there are more than one receipts with the same timestamp, it checks the receipt's sha1 hash and replaces this one
void replace_rcpt(RBTNODE *tree, PWNC_RCPT *rcpt, int store_it);

//checks receipt validity
int check_rcpt(PWNC_RCPT *rcpt);

//checks the provider and consumer teams
//in case they already exists, update records, else add new teams, etc
void check_rcpt_teams(PWNC_RCPT *rcpt);

//return 1 if provider==consumer, else return 0
int check_equal_teams(PWNC_RCPT *rcpt);

void remove_rcpt_credits(PWNC_RCPT *rcpt);
void add_rcpt_credits(PWNC_RCPT *rcpt);

//calls rbtree's range_query
//On each node, do_stuff is applied and do_stuff_args are the data used by the callback func
//e.g. do_stuff_args may contain port, ipaddr, etc to send a receipt to, etc
void rdb_range_query( void (*do_stuff)(RBTNODE*, void *), void *do_stuff_args, unsigned long timestamp);

#ifdef PERM_STORAGE

char RECEIPTS_DB_FILE[256];
char RECEIPTS_DB_FREE_LIST[256];

#define RDB_ACTION_INSERT 10
#define RDB_ACTION_REMOVE 20

int handle_receipt_storage(RBTREE_RCPT *to_be_handled, int action);
void rdb_load_data();
#endif

#endif /* _PWNC_RDB_ENGINE_H */
