/**
* types.h - Data types representing P2PWNC entities 
*
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef _PWNC_TYPES_H
#define _PWNC_TYPES_H

#include "proto.h"

#define SAFE_FREE(x) \
	if(x) { \
		free(x); \
		x = NULL; \
	} \

#define SAFE_FREE_RCPT(r) \
	SAFE_FREE_CERT(r->ConsumerCert); \
	SAFE_FREE_KEY(r->ProviderPubKey); \
	SAFE_FREE_KEY(r->RcptSig); \
	SAFE_FREE(r); \

#define SAFE_FREE_KEY(x) \
	SAFE_FREE(x->data); \
	SAFE_FREE(x); \

#define SAFE_FREE_KEY_PAIR(x) \
	SAFE_FREE_KEY(x->pubkey); \
	SAFE_FREE_KEY(x->privkey); \

#define SAFE_FREE_CERT(x) \
	SAFE_FREE_KEY(x->TeamPubKey); \
	SAFE_FREE_KEY(x->UserPubKey); \
	SAFE_FREE_KEY(x->TeamSig); \

//macro used to allocate len bytes for a struct's field, set struct's datalen field to len and memcpy the data
#define ALLOC_AND_COPY(field, bytes, len) \
	field->datalen = len; \
	ALLOC_AND_CLEAN(field->data , len, unsigned char*); \
	memcpy(field->data, bytes, len); \

//allocates len bytes of memory, casts them to "type", cleans memory (sets bytes to 0)
#define ALLOC_AND_CLEAN(x, len, type) \
	x = (type)malloc(len+1); \
	memset(x, 0, len+1);

typedef struct pwnc_key {
	unsigned short algorithm; //encryption algorithm (RSA,ECC)
	unsigned short bits; //pk len (e.g. 1024, 160)
	unsigned int datalen;
	unsigned char *data; //public key data byte array
} PWNC_PUBKEY, PWNC_PRKEY, PWNC_SIG;

typedef struct pwnc_key_pair {
	PWNC_PUBKEY *pubkey;
	PWNC_PRKEY *privkey;
} PWNC_KEYPAIR;

typedef struct pwnc_certificate {
	PWNC_PUBKEY *TeamPubKey;
	PWNC_PUBKEY *UserPubKey;
	PWNC_SIG *TeamSig;
} PWNC_CERT;

typedef struct pwnc_receipt {
	PWNC_CERT *ConsumerCert;
	PWNC_PUBKEY *ProviderPubKey;
	unsigned long timestamp;
	unsigned long weight;
	PWNC_SIG *RcptSig;
} PWNC_RCPT;

struct pwnc_key *PWNC_KEY_new();
PWNC_CERT *PWNC_CERT_new();
PWNC_KEYPAIR *PWNC_KEYPAIR_new();
PWNC_RCPT * PWNC_RCPT_new();

#endif /* _PWNC_TYPES_H */
