/**
* conffile.c - Routines/Structures for parsing configuration files
* and command line arguments
*
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "conffile.h"

CMD_ARG *get_arg_by_name(unsigned char *name, CMD_ARG *args, int num_args) {
	int i = 0;
	for (i=0; i<num_args; i++) {
		if (!strcmp(name, args[i].full_name)) {
			return &args[i];
		}
	}
	return NULL;
}

CMD_ARG *get_arg_by_switch(unsigned char *swtch, CMD_ARG *args, int num_args) {
	int i = 0;
	for (i=0; i<num_args; i++) {
		if (!strcmp(swtch, args[i].swtch)) {
			return &args[i];
		}
	}
	return NULL;
}

CMD_ARG *make_valid_args(unsigned char **arg_names, unsigned char **switches, void **vars, unsigned short *vtypes, int argnum) {
	int i = 0;
	CMD_ARG *args = (CMD_ARG*)malloc(argnum * sizeof(CMD_ARG) );
	memset(args, 0, argnum*sizeof(CMD_ARG));

	for (i=0; i<argnum; i++) {
		strcpy(args[i].full_name, arg_names[i]);
		args[i].vtype = vtypes[i];
		args[i].var = vars[i];
		strcpy(args[i].swtch, switches[i]);
	}
	return args;
}

//parses command line arguments
//for each cmd line arg:
//First checks if it a long option (--xyz) and looks it up in args (using get_arg_by_name
//Then, checks if it is a switch (e.g -c) and looks it up in args (using get_arg_by_switch)
//In such a case, the following arg is the value of the long opt/switch, which will be stored in args
//Malformed input does not produce error, it is ignored
//only valid iput is stored in args
void parse_cmd_line_args(int argc, char **argv, CMD_ARG *args, int num_args) {
	int i  = 1;
	CMD_ARG *cur;
	while ( i < argc ) {
		//cmd line switch
		if (!strncmp(argv[i], "--", 2)) {
			//long option (e.g --xyz)
			if ( (cur = get_arg_by_name( argv[i]+2, args, num_args) ) ) {
				if (cur->vtype == VTYPE_INT) {
					if (i+1 >= argc) {
						*(int*)cur->var = 0;
					}
					else {
						*(int*)cur->var = atoi(argv[i + 1]);
					}
					i += 2;
				}
				else
				if (cur->vtype == VTYPE_STRING) {
					if (i+1 >= argc) {
						cur->var = NULL;
					}
					else {
						strcpy(cur->var, argv[i + 1]);
					}
					i += 2;
				}
				else
				if (cur->vtype == VTYPE_NOVAL) {
					//when type is VTYPE_NOVAL, then the cur->var field
					//is boolean (either arg exists or not...)
					*(int*)cur->var = 1;
					i++;
				}
			}
			else i++;
//			if (i > argc - 2) break;
		}
		else if (*argv[i] == '-') {
			//cmd line switch
			if ( (cur = get_arg_by_switch( argv[i]+1, args, num_args) ) ) {
				if (cur->vtype == VTYPE_INT) {
					if (i+1 >= argc) {
						*(int*)cur->var = 0;
					}
					else {
						*(int*)cur->var = atoi(argv[i + 1]);
					}
					i += 2;
				}
				else
				if (cur->vtype == VTYPE_STRING) {
					if (i+1 >= argc) {
						cur->var = NULL;
					}
					else {
						strcpy(cur->var, argv[i + 1]);
					}
					i += 2;
				}
				else
				if (cur->vtype == VTYPE_NOVAL) {
					*(int*)cur->var = 1;
					i++;
				}
			} //if this is a valid switch
			else i++;
//			if (i > argc - 2) break;
		}
		else {
			i++;
			continue; //ignore it if its not a switch, e.g. -xyz
		}
	} //loop
}

//Searches for the entry with name "name" (name: the id of an entry) in "entries" (which has "num_entries" elements)
//search is case insensitive
CENTRY *get_entry_by_name(unsigned char *name, CENTRY *entries, int num_entries) {
	int i = 0;
	for (i=0; i<num_entries; i++) {
		if (!strcasecmp(name, entries[i].var_name)) {
			return &entries[i];
		}
	}
	return NULL;
}

//constructs the conf entries table
//first mallocs space for varnum entries
//then, copies var_names to the correspondent entry elements
//extra space for the actual variables is not allocated here (we suppose that it is preallocated by the app)
CENTRY *make_valid_entries(unsigned char **var_names, void **vars, unsigned short *vtypes, int varnum) {
	int i = 0;
	CENTRY *entries = (CENTRY*)malloc(varnum*sizeof(CENTRY));
	memset(entries, 0, varnum*sizeof(CENTRY));

	for (i=0; i<varnum; i++) {
		strcpy(entries[i].var_name, var_names[i]);
		entries[i].var = vars[i];
		entries[i].vtype = vtypes[i];
	}
	return entries;
}

//parses a line:
//removes whitespace before-after the first word (variable name), sets key to this word and sets the rest of the line to value
int parse_line(unsigned char *line, unsigned char *key, unsigned char *value) {
	int len = strlen(line);

	int i = 0;
	int j = 0;
	int k = 0;
	//removes space from the beginning of line
	for (i=0; i<len; i++) {
		if ( !isspace(line[i]) ) {
			break;
		}
	}

	//if its a comment ---> do nothing
	if (line[i] == '#') {
		return 0;
	}

	//gets key
	for (j=i; j<len; j++) {
		if ( isspace(line[j]) ) {
			break;
		}
		key[k] = line[j];
		k++;
	}

	//removes trailing space
	while (isspace(line[j]) && j<len) {
		j++;
	}
	//key without value
	if (j == len) {
		*value = '\0';
		return 1;
	}

	//the rest of it is the value
	strncpy(value, line + j, len - j);

	//remove trailing whitespace from value (replace them with \0)
	for (i=(strlen(value) - 1); i>=0; i--) {
		if (isspace(value[i])) {
			value[i] = '\0';
		}
		else {
			break;
		}
	}
	return 1;
}

//parses a conf file
//reads file line after line
//for each variable name read, it looks it up in entries and then stores its value there (checking the var type)
void parse_conf_file(CENTRY *entries, int entry_num, FILE *fp) {
	char *line = NULL;
	unsigned char *key = (unsigned char*)malloc(256);
	unsigned char *value = (unsigned char*)malloc(256);

	CENTRY *cur = NULL;
	int linelen = 256;
	if (!fp) {
		printf("parse_conf_file: Null file descriptor\n");
		return;
	}

	line = (char*)malloc(256);
	memset(line, 0, 256);
	memset(key, 0, 256);
	memset(value, 0, 256);
	//read line after line
	//if  line starts with #, ignore it
	while(getline(&line, &linelen, fp) != -1) {
		if (!parse_line(line, key, value) ) {
			continue;
		}
		if ( (cur = get_entry_by_name(key, entries, entry_num)) ) {
			if (cur->vtype == VTYPE_INT) {
				*(int*)cur->var = atoi(value);
			}
			else
			if (cur->vtype == VTYPE_STRING) {
				strncpy(cur->var, value, strlen(value));
			}
			else
			if (cur->vtype == VTYPE_NOVAL) {
				*(int*)cur->var = 1;
			}
		}
		memset(line, 0, 256);
		memset(key, 0, 256);
		memset(value, 0, 256);
	}
}

/*
int main(int argc, char ** argv) {
	char str_tca_ipaddr[20];
	char str_pub_key_path[80];
	char str_conf_file_path[80];
	char str_iptables_path[80];
	char str_subnet[80];
	char str_mask[4];
	int mq_key = 0;
	int server_port = 0;

	int update = 0;

	memset(&str_tca_ipaddr, 0, 20);
	memset(&str_pub_key_path, 0, 80);
//	memset(&str_priv_key_path, 0, 80);
	memset(&str_conf_file_path, 0, 80);
	memset(&str_iptables_path, 0, 80);
	memset(&str_subnet, 0, 80);
	memset(&str_mask, 0, 4);

	CMD_ARG *cmd_line_args = NULL;
	CENTRY *conf_entries = NULL;
	unsigned char *valid_conf_names[] = {
		"public-key-path",
//		"secret-key-path",
		"server-port",
		"tca-addr",
		"mq-key",
		"iptables-path",
		"subnet",
		"mask"
		};
	unsigned char *valid_cmd_line_names[] ={
		"conf",
		"public-key-path",
//		"secret-key-path",
		"server-port",
		"tca-addr",
		"mq-key",
		"iptables-path",
		"subnet",
		"mask",
		"update"
		};
	unsigned char* cmd_line_switches[] = {"c", "pk", "sp", "tca", "mq", "ipt", "n", "m", "u"};
	unsigned short conf_var_types[] = {VTYPE_STRING, VTYPE_INT, VTYPE_STRING, VTYPE_INT, VTYPE_STRING, VTYPE_STRING, VTYPE_STRING};
	unsigned short cmd_line_var_types[] = {VTYPE_STRING, VTYPE_STRING, VTYPE_INT, VTYPE_STRING, VTYPE_INT, VTYPE_STRING, VTYPE_STRING, VTYPE_STRING, VTYPE_NOVAL};
	void *conf_variables[] = {&str_pub_key_path, &server_port, &str_tca_ipaddr, &mq_key, &str_iptables_path, &str_subnet, &str_mask};
	void *cmd_line_variables[] = {&str_conf_file_path, &str_pub_key_path, &server_port, &str_tca_ipaddr, &mq_key, &str_iptables_path, &str_subnet, &str_mask, &update};
	conf_entries = make_valid_entries(valid_conf_names, conf_variables, conf_var_types, 7);
	cmd_line_args = make_valid_args(valid_cmd_line_names, cmd_line_switches, cmd_line_variables, cmd_line_var_types, 9);

		//parse command line
		//if there is a -c flag, ignore all other flags and check the specified conf file
		parse_cmd_line_args(argc, argv, cmd_line_args, 9);


	printf("%s\n%s\n%d\n%d\n%s\n%s\n%s\n%d\n", str_pub_key_path, str_tca_ipaddr, server_port, mq_key, str_iptables_path, str_subnet, str_mask, update);
}

*/
