/**
* conversions.c - Type conversions module
* (internal <--> network (ascii) representation, 
* message generation/parsing, etc)
*
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "conversions.h"

//################################################################################
//Convert from broken time format to a time_t UTC value
//It is the inverse of gmtime
time_t my_timegm (struct tm *tm) {
	time_t ret;
	char *tz;

	tz = getenv("TZ");
	setenv("TZ", "", 1);
	tzset();
	ret = mktime(tm);
	if (tz)
		setenv("TZ", tz, 1);
	else
		unsetenv("TZ");
	tzset();
	return ret;
}
//################################################################################

//################################################################################
//Calculates the size of a b64 encoded structure (after the encoding the size is about 33% higher...)
unsigned long calculate_b64_size(unsigned long len) {
	switch (len % 3) {
		case 1:
			return (4.0/3.0*(len+2) + 2*(( 4.0/3.0*(len+2) )/(double)B64_LINE_LEN + 1)); //+1: final crlf, 2*...because crlf=2bytes
		case 2:
			return (4.0/3.0*(len+1) + 2*(( 4.0/3.0*(len+1) )/(double)B64_LINE_LEN + 1));
		case 0:
			return (4*(len/3) + 2*(4.0*(len/3)/(double)B64_LINE_LEN + 1));
	}
	return 0;
}
//################################################################################

//################################################################################
//Returns true if c is a character that should be ignored during b64 decoding
//that is, tab, space, cr, lf, =
int should_be_ignored(unsigned char c) {
	if ( (c == 10) || (c == 13) || (c == 32) || (c == 9) ) {
		return 1;
	}
	return 0;
}
//################################################################################

//################################################################################
//Reads b64 data and removes blank/crfl characters & returns translated b64 byte array
//b64 bytes are translated to the original byte values (e.g. "A" ---> 0, "B" ---> 1, "0" ---> 52)
//in case of error (e.g. non-ascii char, char that doesn't belong to the translation table) returns NULL
//else, sets outlen to the size of the returned array
unsigned char* translate_b64(const unsigned char* data, int inlen, int *outlen) {
	int i = 0;
	int j = -1;
	unsigned char* translated = (unsigned char*)malloc(inlen);
	memset(translated, 0, inlen);
	for (i=0;i<inlen;i++) {
		if (!should_be_ignored(data[i])) {
			//not a whitespace
			j++;

			//we could use something like a lookup table,etc...
			if (  (data[i] >= 65) && (data[i] <= 90) ) {
				//capital A-Z
				translated[j] = data[i] - 65;
			}
			else
			if  ((data[i] >= 97) && (data[i] <= 122)) {
				//lowercase a-z
				translated[j] = data[i] - 71;
			}
			else
			if ((data[i] >= 48) && (data[i] <= 57)) {
				//0-9
				translated[j] = data[i] + 4;
			}
			else
			if (data[i] == 43) {
				//+
				translated[j] = 62;
			}
			else
			if (data[i] == 47) {
				// /
				translated[j] =63;
			}
			else
			if (data[i] == 61) {
				//padding(=)
				translated[j] = 0;
			}
			else {
				//data error
				SAFE_FREE(translated);
				return NULL;
			}
		} //else ignore it...
	}
	*outlen = j + 1;
	return translated;
}
//################################################################################

//################################################################################
//Encodes a 3byte block to a 4-byte b64-encoded block
//len: 1,2,3: if len<3 we must pad with =
void encode_block(unsigned char in[3], unsigned char out[4], int len) {
	out[0] = EncTable[ in[0] >> 2 ];
	out[1] = EncTable[ ((in[0] & 0x03) << 4) | ((in[1] & 0xf0) >> 4) ];
	out[2] = (unsigned char) (len > 1 ? EncTable[ ((in[1] & 0x0f) << 2) | ((in[2] & 0xc0) >> 6) ] : '=');
	out[3] = (unsigned char) (len > 2 ? EncTable[ in[2] & 0x3f ] : '=');
}
//################################################################################

//################################################################################
//Decodes a 4byte b64 block to raw bytes
void decode_block(unsigned char in[4], unsigned char out[3]) {
	out[0] = (unsigned char ) (in[0] << 2 | in[1] >> 4);
	out[1] = (unsigned char ) (in[1] << 4 | in[2] >> 2);
	out[2] = (unsigned char ) (((in[2] << 6) & 0xc0) | in[3]);
}
//################################################################################

//################################################################################
//b64-encodes a raw byte array of length == datalen. Data is output in lines of B64_LINE_LEN characters
//datalen: length of raw data byte array
int b64_encode(unsigned char *raw, unsigned char* b64, int datalen) {
	int i = 0;
	int curpos = 0; //current position in b64 byte array (i.e. how many bytes we have appended to b64)
	int charsinline = 0; //how many chars are in the current line
	for (i = 0; i<datalen-3; i+=3) {
		encode_block(raw + i, b64 + curpos, 3);
		curpos+=4;
		charsinline += 4;
		if (charsinline == B64_LINE_LEN) {
			//we suppose that B64_LINE_LEN is multiple of 4
			charsinline = 0;
			b64[curpos] = '\r';
			b64[curpos+1] = '\n';
			curpos+=2; //added crlf
		}
	}
	if (datalen % 3) {
		//if we have spare characters because datalen is not multiple of 3...
		encode_block(raw + i, b64 + curpos, datalen % 3);
	}
	else {
		encode_block(raw + i, b64 + curpos, 3);
	}
	return strlen(b64); //returns length of b64
}
//################################################################################

//################################################################################
//Decodes the b64 data ---> raw
//("raw" MUST have been allocated before the call...)
int b64_decode(unsigned char *b64, unsigned char* raw) {
	//Call translate_b64 to remove whitespace+crlf characters and padding
	//and to convert b64 ---> original byte values
	//and then operate on "translated"
	int rawpos = 0;
	int i = 0;
	int len = 0; //length of the "processed" b64 array
	int inlen = strlen(b64);
	unsigned char *translated = translate_b64((const unsigned char*)b64, inlen, &len);

	if (!translated) {
		//data error - translate_b64  returned null...
		return -1;
	}

	for (i=0; i<len; i+=4) {
		decode_block(translated + i, raw + rawpos);
		//b64pos+=4;
		rawpos+=3;
	}

	if (strstr(b64, "==")) {
		rawpos-=2;
	}
	else
	if (strstr(b64, "=")) {
		rawpos--;
	}

	return rawpos;
}
//################################################################################

//################################################################################
//Converts from internat pwnc representation (structs)
//to ascii text to be transmitted through the network
unsigned char* receipt_2_text(PWNC_RCPT *rcpt) {
	unsigned char *text; //final text message

	//total length of text message that will be constructed from rcpt
	int rlen = 0;
	//content length;
	int clen = 0;

	unsigned char stralgo[5]; //string repr. of encryption algorithm
	memset(stralgo, 0, 5);
	unsigned char strbits[5]; //str representation of  key length in bits
	memset(strbits, 0, 5);
	unsigned char strts[32]; //string representation of timestamp (UTC)
	memset(strts, 0, 32);
	unsigned char strweight[20]; //string representation of weight
	memset(strweight, 0, 20);
	unsigned char strcl[10]; //string repr. of content length
	memset(strcl, 0, 10);

	struct tm utctime; //timestamp in UTC

	rlen += strlen("RCPT P2PWNC/1.0") + 2; //+2 : crlf
	clen += strlen("Algorithm: ") + 2;//+2 crlf
	switch (rcpt->RcptSig->algorithm) {
		case PWNC_ALGO_ECC:
			clen += strlen("ECC");
			strcpy(stralgo, "ECC");
			break;
		case PWNC_ALGO_RSA: default:
			clen += strlen("RSA");
			strcpy(stralgo, "RSA");
			break;
	}
	sprintf(strbits, "%hd", rcpt->RcptSig->bits);
	clen += strlen(strbits);
/*
	TODO: CHECK TIMESTAMP!!!!
	localtime_r( (const time_t*)&(rcpt->timestamp), &utctime );
*/
	gmtime_r((const time_t*)&(rcpt->timestamp), &utctime);
	strftime(strts, 32, "%a, %d %b %Y %H:%M:%S %z", &utctime); //string representation - RFC 822
	clen += strlen("Timestamp: ") + strlen(strts) + 2;//+2: crlf

	//construct string representation of weight
	sprintf(strweight, "%lu", rcpt->weight);
	clen += strlen("Weight: ") + strlen(strweight) + 2;//+2: crlf

	//now serialize receipt info (cert, key, signature)
	int rawlen = 0;
	unsigned char *serialized = serialize_receipt_enc_data(rcpt, &rawlen);

	unsigned char *b64data = NULL;
	int b64len = calculate_b64_size(rawlen);
	ALLOC_AND_CLEAN(b64data, b64len, unsigned char*);

	b64len = b64_encode( serialized, b64data, rawlen);

	//Calculate "content-length" field
	//content-length is the first field after the msg "title" and contains the length of the remaining message
	clen += b64len;
	sprintf(strcl, "%d", clen);
	//Calculate total length of text message (add the remaining pieces' length to "rlen" (plus crlfs)
	rlen += ( strlen("Content-length: ") + strlen(strcl) + 2 + clen + 1);
	//finally, allocate mem for actual text message data

	text = (unsigned char*)malloc(rlen);
	memset(text, 0, rlen);
	//now sprintf data to text
	sprintf(
		text,
		"RCPT P2PWNC/1.0\r\nContent-length: %s\r\nAlgorithm: %s%s\r\nTimestamp: %s\r\nWeight: %s\r\n%s",
		strcl, stralgo, strbits, strts, strweight, b64data
		);

	SAFE_FREE(b64data);

	return text;
}
//################################################################################

//################################################################################
//Constructs a "CONNECT" message that a client sends to an access point
unsigned char* construct_conn_msg(PWNC_CERT *ClientCert) {
	unsigned char *text; //text message

	unsigned char *b64_data = NULL; //b64 encoded serialized cert data
	unsigned char *raw_data = NULL; //packed cert data
	unsigned int b64_len = 0;
	unsigned int raw_len = 0;

	//total length of text message that will be constructed from rcpt
	int mlen = 0;
	//content length
	int clen = 0;

	unsigned char stralgo[5]; //string repr. of ancryption algorithm
	memset(stralgo, 0, 5);
	unsigned char strbits[5]; //str representation of  key length in bits
	memset(strbits, 0, 5);
	unsigned char strcl[10]; //string repr. of content length
	memset(strcl, 0, 10);

	mlen += strlen("CONN P2PWNC/1.0") + 2; //+2 : crlf
	clen += strlen("Algorithm: ") + 2;//+2 crlf
	switch (ClientCert->TeamPubKey->algorithm) {
		case PWNC_ALGO_ECC:
			clen += strlen("ECC");
			strcpy(stralgo, "ECC");
			break;
		case PWNC_ALGO_RSA: default:
			clen += strlen("RSA");
			strcpy(stralgo, "RSA");
			break;
	}
	sprintf(strbits, "%hd", ClientCert->TeamPubKey->bits);
	clen += strlen(strbits);

	raw_data = serialize_cert_enc_data(ClientCert, &raw_len);
	b64_len = calculate_b64_size(raw_len);
	ALLOC_AND_CLEAN(b64_data, b64_len, unsigned char*);

	//Generate b64 encoded byte array
	b64_len = b64_encode( raw_data, b64_data, raw_len);

	//Calculate "content-length" field
	clen += b64_len;
	sprintf(strcl, "%d", clen);

	//Calculate total length of text message (add the remaining pieces' length to "mlen" (plus crlfs)
	mlen += (strlen("Content-length: ") + strlen(strcl) + 2 + clen);

	//finally, allocate mem for actual text message data
	ALLOC_AND_CLEAN(text, mlen, unsigned char*);

	//now sprintf data to text
	sprintf(
		text,
		"CONN P2PWNC/1.0\r\nContent-length: %s\r\nAlgorithm: %s%s\r\n%s",
		strcl, stralgo, strbits, b64_data
		);

	SAFE_FREE(b64_data);
	SAFE_FREE(raw_data);

	return text;
}
//################################################################################

//################################################################################
//Constructs and returns a Receipt Request message (RREQ)
unsigned char* construct_rreq(PWNC_PUBKEY *pkey, unsigned long weight) {
	unsigned char *text; //text message
	unsigned char *serverpub = NULL; //server pubkey

	unsigned int keylen = 0; //length of the (b64 encoded) pubkey data

	//total length of text message
	int mlen = 0;
	//content length
	int clen = 0;

	unsigned char stralgo[5]; //string repr. of ancryption algorithm
	memset(stralgo, 0, 5);
	unsigned char strbits[5]; //str representation of  key length in bits
	memset(strbits, 0, 5);
	unsigned char strcl[10]; //string repr. of content length
	memset(strcl, 0, 10);
	unsigned char strweight[15];
	memset(strweight, 0, 15);

	mlen += strlen("RREQ P2PWNC/1.0") + 2; //+2 : crlf
	clen += strlen("Algorithm: ") + 2;//+2 crlf
	switch (pkey->algorithm) {
		case PWNC_ALGO_ECC:
			clen += strlen("ECC");
			strcpy(stralgo, "ECC");
			break;
		case PWNC_ALGO_RSA: default:
			clen += strlen("RSA");
			strcpy(stralgo, "RSA");
			break;
	}
	sprintf(strbits, "%hd", pkey->bits);
	clen += strlen(strbits);

	//weight string representation
	sprintf(strweight, "%ld", weight);
	clen += strlen("Weight: ") + strlen(strweight) + 2;

	//server public key representation (base64 encoded, separated from the msg headers by a crlf)
	keylen = calculate_b64_size(pkey->datalen);
//	serverpub = (unsigned char*)malloc( keylen );
//	memset(serverpub, 0, keylen);

	ALLOC_AND_CLEAN(serverpub, keylen + 100, unsigned char*);

	//Generate b64 encoded byte arrays
	keylen = b64_encode( pkey->data, serverpub, pkey->datalen );
	clen += keylen;

	//Calculate "content-length" field
	sprintf(strcl, "%d", clen);

	//Calculate total length of text message (add the pubkey length + "content-length" to "mlen")
	mlen += strlen("Content-length: ") + strlen(strcl) + 2 + clen;
	//allocate mem for actual text message data
	text = (unsigned char*)malloc(mlen+100);
	memset(text, 0, mlen);

	// sprintf data to text
	sprintf(
		text,
		"RREQ P2PWNC/1.0\r\nContent-length: %s\r\nAlgorithm: %s%s\r\nWeight: %s\r\n%s",
		strcl, stralgo, strbits, strweight, serverpub
		);

	//free pointers
	SAFE_FREE(serverpub);
	return text;
}
//################################################################################

//################################################################################
//Constructs a "CACK" message
//this includes a timestamp to be used in receipts
//and the port number to which clients send messages
unsigned char* construct_conn_ack(unsigned long timestamp, unsigned short mode) {
	unsigned char* text; //text message
	struct tm utctime; //timestamp in UTC
	int mlen = 0; //length of text message
	int clen = 0; //content length

	//string repesentation of content length
	unsigned char *strcl = NULL;
	ALLOC_AND_CLEAN(strcl, 10, unsigned char*);
	unsigned char strts[100];

	//if mode is MODE_RCPT_REQ_ON then construct a cack message asking for receipts
	//CACK\r\n
	//Content-length: 8\r\n
	//rcpt req
	if (mode == MODE_RCPT_REQ_ON) {
		//TODO: mallon akyro, mallon de 8a xreiastei
		clen = 8; //strlen("rcpt req");
		mlen = strlen("CACK P2PWNC/1.0\r\nContent-length: 8") + 10 + 1;//+1: null byte
		text = (unsigned char*)malloc(mlen);
		sprintf(text, "CACK P2PWNC/1.0\r\nContent-length: 8\r\nRCPT REQ");
	}
	else {
		//string representation of timestamp
		memset(strts, 0, 100);
		gmtime_r((const time_t*)&timestamp, &utctime);
		strftime(strts, 100, "%a, %d %b %Y %H:%M:%S %z", &utctime); //string representation - RFC 822

		//calculate + write content-length to string
		clen = strlen("Timestamp: ") + strlen(strts);
		sprintf(strcl, "%d", clen);

		mlen = strlen("CACK P2PWNC/1.0\r\nContent-length: ") + strlen(strcl) + 2 + strlen("Timestamp: ") + strlen(strts);
		text = (unsigned char*)malloc(mlen + 100);
		sprintf(
			text,
			"CACK P2PWNC/1.0\r\nContent-length: %s\r\nTimestamp: %s",
			strcl, strts
			);
	}
	SAFE_FREE(strcl);

	return text;
}
//################################################################################

//################################################################################
//Parses a "RECEIPT" text message
//and returns a PWNC_RCPT struct
PWNC_RCPT *text_2_rcpt(unsigned char* text) {
	PWNC_RCPT *rcpt = NULL;

	unsigned short receipt_algorithm = 0;
	unsigned short bits = 0;
	unsigned int clen = 0; //content length
	unsigned long weight = 0;
	unsigned long timestamp = 0;

	unsigned char strts[32]; //date in rfc822 format (28bytes)
	memset(strts, 0, 32);
	unsigned char checkstr[32]; //used when checking time conversion
	memset(checkstr, 0, 32);
	struct tm utctime; //timestamp in broken-down format
	memset(&utctime, 0, sizeof(struct tm));

	//this is our current position inside the "text" structure
	unsigned char *pos = text;
	//pointer to the start of the message-specific data (that is data which are counted by content-length)
	unsigned char *startofdata = NULL;

	if ( strstr( pos, "RCPT P2PWNC/1.0\r\n") != (char*)pos) {
		//if msg doesn't start with "RECEIPT" --->malformed msg
		return NULL;
	}

	//check content length
	if (!(pos = strstr(pos, "\r\n")) ) {
		return NULL;
	}
	pos += 2;
	if (strncasecmp(pos, "Content-length: ", strlen("Content-length: "))) {
		//malformed header
		return NULL;
	}
	//move pos...
	pos += strlen("Content-length: ");
	//read content length
	if (!sscanf(pos, "%u", &clen)) {
		return NULL;
	}
	//move pos to the next crlf (right before "Algorithm:"
	if (!(pos = strstr(pos, "\r\n")) ) {
		return NULL;
	}
	pos+=2;

	startofdata = pos;

	//check whether the next header field is "Algorithm: " - else ---> malformed msg, return NULL
	if ( strncasecmp (pos, "Algorithm: ", strlen("Algorithm: ")) ) {
		return NULL;
	}
	//move pos to the beginning of the algorithm value ("RSA...\r\n")
	pos += strlen("Algorithm: ");
	//read algoritmh
	if (strncasecmp(pos, "RSA", 3)) {
		if (strncasecmp(pos, "ECC", 3)) {
			//unknown algorithm identifier or just grabage...
			return NULL;
		}
		else {
			receipt_algorithm = PWNC_ALGO_ECC;
		}
	}
	else {
		receipt_algorithm = PWNC_ALGO_RSA;
	}
	pos += 3;
	//now read bits
	if (!sscanf(pos, "%hd", &bits)) {
		//error reading bits
		return NULL;
	}

	//move pos after the next crlf
	if (!(pos = strstr(pos, "\r\n"))) {
		return NULL;
	}
	pos += 2;

	//now read timestamp
	if (strncasecmp(pos, "Timestamp: ", strlen("Timestamp: ")) ) {
		//header err, etc
		return NULL;
	}
	pos += strlen("Timestamp: ");
	//copy message timestamp to "strts" string
	if (!strncpy(strts, pos, 31) ) {
		//copy 31 bytes (see RFC822 date format...)
		return NULL;
	}

	//read rfc822 timestamp from "strts" to a broken-down time format
	strptime(strts, "%a, %d %b %Y %H:%M:%S %z", &utctime);
	//set timestamp
	//rcpt->timestamp = (unsigned long)mktime(&utctime);
	timestamp = (unsigned long)my_timegm(&utctime);
	//timestamp check
/*
	//TODO: AYTO PREPEI NA TO KOITAKSW!!!
	localtime_r( (const time_t*)&(rcpt->timestamp), &utctime );
	strftime(checkstr, 32, "%a, %d %b %Y %H:%M:%S %z", &utctime);
	//if  strts==checkstr then conversion has succeeded
	//else maybe strts was malformed and checkstr==default <> strts ==> error
	if (strcasecmp(strts, checkstr)) {
		//strts<>checkstr
		SAFE_FREE_RCPT(rcpt);
		return NULL;
	}
*/
	//move pos after the next crlf
	if (! (pos = strstr(pos, "\r\n")) ) {
		return NULL;
	}
	pos += 2;

	//now we are at the "Weight: " header
	if (strncasecmp(pos, "Weight: ", strlen("Weight: "))) {
		//malformed header
		return NULL;
	}
	//move pointer to the start of the weight value
	pos += strlen("Weight: ");
	//read the actual weight (unsigned long)
	if (!sscanf(pos, "%lu", &weight)) {
//		SAFE_FREE_RCPT(rcpt);
		return NULL;
	}

	//now we're ready to read the b64-eecoded byte arrays
	//move pos to start of data
	if (!(pos = strstr(pos, "\r\n")) ) {
		return NULL;
	}
	pos += 2;

	//allocate new receipt
	rcpt = PWNC_RCPT_new();
	rcpt->timestamp = timestamp;
	rcpt->weight = weight;

	//decode b64 data
	unsigned char *raw_data = NULL;
	unsigned char *b64_data = NULL;
	int rawlen = 0;
	ALLOC_AND_CLEAN(b64_data, clen - (pos - startofdata - 1), unsigned char*);
	ALLOC_AND_CLEAN(raw_data, clen - (pos - startofdata - 1), unsigned char*); //alloc more space than necessary, but anyway....
	strncpy(b64_data, pos, clen - (pos - startofdata - 1));
	if ( (rawlen = b64_decode(b64_data, raw_data)) == -1) {
		//actual decoding
		SAFE_FREE_RCPT(rcpt);
		SAFE_FREE(raw_data);
		SAFE_FREE(b64_data);
		return NULL;
	}
	SAFE_FREE(b64_data);

	if (!deserialize_receipt_enc_data(rcpt, raw_data, rawlen, receipt_algorithm, bits)) {
		SAFE_FREE(raw_data);
		SAFE_FREE(rcpt);
		return NULL;
	}
	return rcpt;
}
//################################################################################

//################################################################################
//Parses a text "CONNECT" message and returns a pointer to a PWNC_CERT struct (the certificate of
//the consumer
PWNC_CERT* parse_conn_msg(unsigned char *text) {
	PWNC_CERT *cert = NULL;
	unsigned char *raw_data = NULL; //raw bytes (will be used later in decoding b64 strings)
	unsigned char *b64_data = NULL;//b64 encoded bytes (as copied from "text")
	unsigned int raw_len = 0; //length of the decoded data (will be used while decoding from b64)

	unsigned short algorithm = 0;
	unsigned short bits = 0;
	unsigned int clen = 0; //content length

	//this is our current position inside the "text" structure
	unsigned char *pos = text;
	//start of message specific data
	unsigned char *startofdata = NULL;

	if (strncasecmp(pos, "CONN P2PWNC/1.0\r\n", strlen("CONN P2PWNC/1.0\r\n"))) {
		//malformed header
		return NULL;
	}

	//now check content length
	if (!(pos = strstr(pos, "\r\n")) ) {
		return NULL;
	}
	pos += 2;
	if (strncasecmp(pos, "Content-length: ", strlen("Content-length: "))) {
		//malformed header
		return NULL;
	}
	//move pos...
	pos += strlen("Content-length: ");
	//read content length
	if (!sscanf(pos, "%u", &clen)) {
		return NULL;
	}

	//move pos to the next crlf (right before "Algorithm:")
	if (!(pos = strstr(pos, "\r\n")) ) {
		return NULL;
	}
	pos+=2;
	startofdata = pos;

	//check whether the next header field is "Algorithm: " - else ---> malformed msg, return NULL
	if ( strncasecmp (pos, "Algorithm: ", strlen("Algorithm: ")) ) {
		return NULL;
	}

	//move pos to the beginning of the algorithm value ("RSA...\r\n")
	pos += strlen("Algorithm: ");
	//read algoritmh
	if (strncasecmp(pos, "RSA", 3)) {
		if (strncasecmp(pos, "ECC", 3)) {
			//unknown algorithm identifier or just grabage...
			return NULL;
		}
		else {
			algorithm = PWNC_ALGO_ECC;
		}
	}
	else {
		algorithm = PWNC_ALGO_RSA;
	}
	pos += 3;

	//now read bits
	if (!sscanf(pos, "%hd", &bits)) {
		//error reading bits
		return NULL;
	}

	//move pos after the next crlf
	if (!(pos = strstr(pos, "\r\n"))) {
		return NULL;
	}
	pos += 2;

	//now decode data and deserialize them
	ALLOC_AND_CLEAN(b64_data, clen - (pos - startofdata), unsigned char*);
	ALLOC_AND_CLEAN(raw_data, clen - (pos - startofdata), unsigned char*);
	strncpy(b64_data, pos, clen - (pos - startofdata)); //copy the b64 enc data to b64_data
	if ( (raw_len = b64_decode(b64_data, raw_data)) == -1) { //decode them
		SAFE_FREE(raw_data);
		SAFE_FREE(b64_data);
		return NULL;
	}
	SAFE_FREE(b64_data);

//TODO: isws to BUG na einai kapou EDW!
	if (! (cert = deserialize_cert_enc_data(raw_data, raw_len, algorithm, bits)) ) {
		SAFE_FREE(raw_data);
		return NULL;
	}
	return cert;
}
//################################################################################

//################################################################################
//Parses a cack message. Extracts timestamp from message
int parse_cack_msg(unsigned char *text, unsigned long *timestamp) {
//	unsigned char strts[32];
	unsigned char checkstr[32];
	unsigned char *pos = NULL;
	struct tm utctime;
	memset(&utctime, 0, sizeof(struct tm));
//	memset(strts, 0, 32);
	memset(checkstr, 0, 32);

	if (strncasecmp(text, "cack P2PWNC/1.0\r\n", 6 + strlen(" P2PWNC/1.0") )) {
		return 0;
	}
	pos = text + strlen("cack P2PWNC/1.0\r\n");

	//now read and move past content length
	if (strncasecmp(pos, "content-length: ", strlen("content-length: "))) {
		return 0;
	}
	pos = strstr(pos, "\r\n") + 2;

	if (!strncasecmp(pos, "rcpt req", 8)) {
		return MODE_RCPT_REQ_ON;
	}

	//check header and move past "Timestamp: "
	if (strncasecmp(pos, "timestamp: ", strlen("timestamp: "))) {
		return 0;
	}
	pos += strlen("timestamp: ");

	//now pos = rfc822 timestamp
	strptime(pos, "%a, %d %b %Y %H:%M:%S %z", &utctime);
	//timestamp check
	//*timestamp = (unsigned long)mktime(&utctime);
	*timestamp = (unsigned long)my_timegm(&utctime);

/*	localtime_r( (const time_t*)timestamp, &utctime );
	//gmtime_r((const time_t*)timestamp, &utctime);
	strftime(checkstr, 32, "%a, %d %b %Y %H:%M:%S %z", &utctime);
	//if  strts==checkstr then conversion has succeeded
	//else maybe strts was malformed and checkstr==default <> strts ==> error
	if (strncasecmp(pos, checkstr, 31)) {
		//strts<>checkstr
		printf("%s\n%s\n", pos, checkstr);
		return 0;
	}
*/
	return MODE_RCPT_REQ_OFF;
}
//################################################################################

//################################################################################
//Reads the "server's" public key from a RREQ text message
PWNC_PUBKEY* parse_rreq(unsigned char* text, unsigned long *weight) {
	PWNC_PUBKEY *pkey = (PWNC_PUBKEY*)malloc(sizeof(PWNC_PUBKEY));
	unsigned char *raw = NULL; //raw bytes (will be used later in decoding b64 strings)
	unsigned char *b64 = NULL;//b64 encoded bytes (as copied from "text"
	unsigned int rawlen = 0; //length of the decoded data (will be used while decoding from b64)
	unsigned short bits = 0;
	unsigned int clen = 0; //content length

	//this is our current position inside the "text" structure
	unsigned char *pos = text;
	unsigned char *startofdata = NULL;

	int failed = 0;

	if ( strstr( pos, "RREQ P2PWNC/1.0\r\n") != (char*)pos) {
		//if msg doesn't start with "RECEIPT" --->malformed msg
		SAFE_FREE(pkey);
		return NULL;
	}

	//check content length
	if (!(pos = strstr(pos, "\r\n")) ) {
		SAFE_FREE(pkey);
		return NULL;
	}
	pos += 2;
	if (strncasecmp(pos, "Content-length: ", strlen("Content-length: "))) {
		//malformed header
		SAFE_FREE(pkey);
		return NULL;
	}
	//move pos...
	pos += strlen("Content-length: ");
	//read content length
	if (!sscanf(pos, "%u", &clen)) {
		SAFE_FREE(pkey);
		return NULL;
	}


	//move pos to the next crlf (right before "Algorithm:")
	if (!(pos = strstr(pos, "\r\n")) ) {
		SAFE_FREE(pkey);
		return NULL;
	}
	pos+=2;
	startofdata = pos;

	//check whether the next header field is "Algorithm: " - else ---> malformed msg, return NULL
	if ( strncasecmp (pos, "Algorithm: ", strlen("Algorithm: ")) ) {
		SAFE_FREE(pkey);
		return NULL;
	}

	//move pos to the beginning of the algorithm value ("RSA...\r\n")
	pos += strlen("Algorithm: ");
	//read algoritmh
	if (strncasecmp(pos, "RSA", 3)) {
		if (strncasecmp(pos, "ECC", 3)) {
			//unknown algorithm identifier or just grabage...
			SAFE_FREE(pkey);
			return NULL;
		}
		else {
			pkey->algorithm = PWNC_ALGO_ECC;
		}
	}
	else {
		pkey->algorithm = PWNC_ALGO_RSA;
	}
	pos += 3;

	//now read bits
	if (!sscanf(pos, "%hd", &bits)) {
		//error reading bits
		SAFE_FREE(pkey);
		return NULL;
	}
	pkey->bits = bits;

	//move pos after the next crlf
	if (!(pos = strstr(pos, "\r\n"))) {
		SAFE_FREE(pkey);
		return NULL;
	}
	pos += 2;

	//read weight
	if (strncasecmp(pos, "Weight: ", strlen("Weight: "))) {
		//malformed hdr
		SAFE_FREE(pkey);
		return NULL;
	}
	pos += strlen("Weight: ");
	if (!sscanf(pos, "%ld", weight)) {
		//error reading weight
		SAFE_FREE(pkey);
		return NULL;
	}
	//move pos after the next crlf - start of pk data
	if (!(pos = strstr(pos, "\r\n"))) {
		SAFE_FREE(pkey);
		return NULL;
	}
	pos += 2;

	clen -= (pos - startofdata);
	DECODE_AND_COPY(pkey, pos, pos + clen, raw, b64, rawlen, pkey, failed);
	if (failed) {
		SAFE_FREE_KEY(pkey);
		return NULL;
	}

	return pkey;
}
//################################################################################

//################################################################################
//Returns the length of the b64 encoded provider public key which is included in a QUERY text message
int get_prov_pklen(unsigned char *text) {
	unsigned char *pos = text;
	unsigned char *crlfpos = NULL;

	//check msg header
	if ( strstr( pos, "QUER P2PWNC/1.0\r\n") != (char*)pos) {
		return -1;
	}

	//go to the beginning of the content-length field
	if (!(pos = strstr(pos, "\r\n")) ) {
		return -1;
	}

	pos += 2;
	if (strncasecmp(pos, "Content-length: ", strlen("Content-length: "))) {
		//malformed header
		return -1;
	}
	//move pos to the beginning of provider-pubkey
	if (!(pos = strstr(pos, "\r\n")) ) {
		return -1;
	}
	pos += 2;

	//find crlfcrlf (end of provider pubkey)
	if ( !(crlfpos = strstr(pos, "\r\n\r\n")) ) {
		return -1;
	}

	return (crlfpos - pos);
}
//################################################################################

//################################################################################
//Returns the length of the b64 encoded consumer public key which is included in a QUERY text message
int get_cons_pklen(unsigned char *text) {
	unsigned int clen = 0;
	unsigned int provpklen = 0;

	unsigned char *pos = text;

	//check msg header
	if ( strstr( pos, "QUER P2PWNC/1.0\r\n") != (char*)pos) {
		return -1;
	}

	//go to the beginning of the content-length field
	if (!(pos = strstr(pos, "\r\n")) ) {
		return -1;
	}

	pos += 2;
	if (strncasecmp(pos, "Content-length: ", strlen("Content-length: "))) {
		//malformed header
		return -1;
	}
	pos += strlen("Content-length: ");
	//read content length
	if (!sscanf(pos, "%u", &clen)) {
		return -1;
	}

	if ( (provpklen = get_prov_pklen(text)) == -1) {
		return -1;
	}

	//consumer pklen = contentlength - providerpklen - 2*|crlf|
	return (clen - provpklen - 4);
}
//################################################################################

//################################################################################
//Constructs a QUERY text message
//provider: public key of the provider
//consumer: public key of the consumer
unsigned char* construct_query(PWNC_PUBKEY* provider, PWNC_PUBKEY* consumer) {
	unsigned char *text; //text message

	unsigned char *consumerpk = NULL; //consumer public key
	unsigned char *providerpk = NULL; //provider pubkey

	int providerpklen, consumerpklen;

	//message total length
	int mlen = 0;
	//content length
	int clen = 0;

	unsigned char strcl[10]; //string repr. of content length
	memset(strcl, 0, 10);

	mlen += strlen("QUER P2PWNC/1.0") + 2; //+2 : crlf

	providerpklen = calculate_b64_size(provider->datalen);
	consumerpklen = calculate_b64_size(consumer->datalen);

	providerpk = (unsigned char*)malloc( providerpklen );
	consumerpk = (unsigned char*)malloc( consumerpklen );

	memset(providerpk, 0, providerpklen);
	memset(consumerpk, 0, consumerpklen);

	//Generate b64 encoded byte arrays
	providerpklen = b64_encode( provider->data, providerpk, provider->datalen );
	consumerpklen = b64_encode( consumer->data, consumerpk, consumer->datalen );

	//Calculate "content-length" field
	clen += providerpklen + 4 + consumerpklen;
	sprintf(strcl, "%d", clen);

	//Calculate total length of text message
	mlen += (strlen("Content-length: ") + strlen(strcl) + 2 + clen);

	//allocate mem for actual text message data
	text = (unsigned char*)malloc(mlen);
	memset(text, 0, mlen);
	//now sprintf data to text
	sprintf(
		text,
		"QUER P2PWNC/1.0\r\nContent-length: %s\r\n%s\r\n\r\n%s",
		strcl, providerpk, consumerpk
		);

	SAFE_FREE(providerpk);
	SAFE_FREE(consumerpk);

	return text;
}
//################################################################################

//################################################################################
//Constructs a query response message
//action: "Grant"/"Forbid"
unsigned char* construct_qrsp(unsigned short action) {
	//text message
	unsigned char *text = NULL;
	unsigned char straction[20];
	int mlen = 0;
	int clen = 0; //content length
	int actlen = 0; //len of action string ("Grant" or "Forbid")

	unsigned char strcl[10];
	memset(strcl, 0, 10);

	mlen += strlen("QRSP P2PWNC/1.0\r\nContent-length: ");
	switch (action) {
		case PWNC_ACTION_GRANT:
			actlen = strlen("Grant");
			memset(&straction, 0, 20);
			strcpy(straction, "Grant\0");
			break;
		case PWNC_ACTION_FORBID:default:
			actlen = strlen("Forbid");
			memset(&straction, 0, 20);
			strcpy(straction, "Forbid\0");
			break;
	}

	clen += strlen("Action: ") + actlen;
	sprintf(strcl, "%d", clen);

	mlen += strlen(strcl) + 2 + actlen;
	text = (unsigned char*)malloc(1000);
	if (text) {
		memset(text, 0, 1000);
		sprintf(
			text,
			"QRSP P2PWNC/1.0\r\nContent-length: %s\r\nAction: %s",
			&strcl, &straction
			);
		return text;
	}
	else {
		//SAFE_FREE(text);
		return NULL;
	}
}

//################################################################################

//################################################################################
//Constructs an "RRSP" message (receipt response..)
//mcode: 0 if its OK, else the error code
unsigned char* construct_rrsp(unsigned short mcode) {
	unsigned char *text = NULL;
	unsigned char *errstring = NULL;
	unsigned char *strcl = NULL; //content length string
	int mlen = 0;
	int clen = 0;

	ALLOC_AND_CLEAN(strcl, 10, unsigned char*);

	mlen += strlen("RRSP P2PWNC/1.0\r\nContent-length: ");

	if (mcode) {
		//error code != 0
		ALLOC_AND_CLEAN(errstring, 10, unsigned char*);
		sprintf(errstring, "ERR %hd", mcode);
	}
	else {
		ALLOC_AND_CLEAN(errstring, 3, unsigned char*);
		strcpy(errstring, "OK");
		//OK message
	}
	clen += strlen(errstring);
	mlen += strlen(errstring);

	sprintf(strcl, "%d", clen);

	text = (unsigned char*)malloc(mlen + 100);
	if (text) {
		memset(text, 0, mlen + 100);
		sprintf(
			text,
			"RRSP P2PWNC/1.0\r\nContent-length: %s\r\n%s",
			strcl, errstring
			);
	}
	else {
		SAFE_FREE(errstring);
		SAFE_FREE(strcl);
		return NULL;
	}
	SAFE_FREE(errstring);
	SAFE_FREE(strcl);
	return text;
}
//################################################################################

//################################################################################
//Parses a "QUER" text message ("text") and stores the pubkey values read in the provider/consumer structs
//provider and consumer MUST have been allocated prior to a call to parse_query_msg
//Memory allocation of the above must be done using get_prov_pklen and get_cons_pklen
//if OK --> return 0, else -1
int parse_query_msg(unsigned char *text, PWNC_PUBKEY* provider, PWNC_PUBKEY* consumer) {
	unsigned char *raw = NULL; //raw bytes (will be used later in decoding b64 strings)
	unsigned char *b64 = NULL;//b64 encoded bytes (as copied from "text")
	unsigned int rawlen = 0; //length of the decoded data (will be used while decoding from b64)
	unsigned int clen = 0; //content length

	//this is our current position inside the "text" structure
	unsigned char *pos = text;
	unsigned char *crlfpos = NULL;
	unsigned char *startofdata = NULL;

	int failed = 0;

	if ( strstr( pos, "QUER P2PWNC/1.0\r\n") != (char*)pos) {
		//if msg doesn't start with "RECEIPT" --->malformed msg
		return -1;
	}

	//check content length
	if (!(pos = strstr(pos, "\r\n")) ) {
		return -1;
	}
	pos += 2;
	if (strncasecmp(pos, "Content-length: ", strlen("Content-length: "))) {
		//malformed header
		return -1;
	}
	//move pos...
	pos += strlen("Content-length: ");
	//read content length
	if (!sscanf(pos, "%u", &clen)) {
		return -1;
	}

	//move pos to the next crlf
	if (!(pos = strstr(pos, "\r\n")) ) {
		return -1;
	}
	pos+=2;
	startofdata = pos;

	if (!(crlfpos = strstr(pos, "\r\n\r\n"))) {
		return -1;
	}
	//careful: if DECODE_AND_COPY fails, provider will be freed)
	//DECODE_AND_COPY( provider, pos, crlfpos - 1, raw, b64, rawlen, provider, failed);
	DECODE_AND_COPY( provider, pos, crlfpos, raw, b64, rawlen, provider, failed);
	if (failed) {
		return -1;
	}

	pos = crlfpos + 4; //move pos to the start of the consumer public key
	clen -= (pos - startofdata);
	DECODE_AND_COPY( consumer, pos, pos + clen, raw, b64, rawlen, consumer, failed);
	if (failed) {
		return -1;
	}
	return 0;
}
//################################################################################

//################################################################################
//Parses a QRSP text message
//return values:
//0: forbid
//1: grant
//-1: error
int parse_qrsp(unsigned char *text) {
	unsigned char *pos = NULL;

	if (strncasecmp(text, "qrsp P2PWNC/1.0\r\n", 6 + strlen(" P2PWNC/1.0") )) {
		return -1;
	}
	pos = text + 6 + strlen(" P2PWNC/1.0");

	//read and move past content length
	if (strncasecmp(pos, "content-length: ", strlen("content-length: "))) {
		return -1;
	}
	pos = strstr(pos, "\r\n") + 2;

	if (strncasecmp(pos, "action: ", strlen("action: "))) {
		return -1;
	}
	pos += strlen("action: ");

	if (strncasecmp(pos, "forbid", strlen("forbid"))) {
		if (strncasecmp(pos, "grant", strlen("grant"))) {
			return -1; //error
		}
		else {
			return 1; //grant access
		}
	}
	else {
		return 0; //forbid;
	}

	return 0;
}
//################################################################################

//################################################################################
//parses a RRSP message (from TCA to team)
//return values:
//1: success
//or other error code (see pwnc_proto.h)
int parse_rrsp(unsigned char *text) {
	unsigned char *pos = NULL;
	unsigned short ecode = 0;
	if (strncasecmp(text, "rrsp P2PWNC/1.0\r\n", 6 + strlen(" P2PWNC/1.0") )) {
		return -1;
	}
	pos = text + 6 + strlen(" P2PWNC/1.0");

	//read content length
	if (strncasecmp(pos, "content-length: ", strlen("content-length: "))) {
		return -1;
	}
	pos = strstr(pos, "\r\n") + 2;

	if (strncasecmp(pos, "ok", 2)) {
		if (strncasecmp(pos, "err ", 3)) {
			return -1;
		}
		else {
			//ERR xyz
			pos += 4;
			if (!sscanf(pos, "%hd", &ecode)) {
				return -1;
			}
			return ecode;
		}
	}
	else {
		//ok
		return 1;
	}
	return -1;
}
//################################################################################

//################################################################################
//constructs the content of an UPDT message a client sends to its home
//this message must contain the client's certificate and the timestamp.
//Cert will be b64 encoded and timestamp will be represented according to the RFC822 format.
//returns the text data and sets mlen to their length
//algorithm id not needed
//message format:
//UPDT\r\n
//Content-length: ...\r\n
//Timestamp: .....\r\n
unsigned char *make_updt_req_data(unsigned long timestamp, int *mlen) {
	unsigned char *text = NULL;
	
	struct tm utctime; //timestamp in UTC

	//string representation of timestamp
	unsigned char strts[100];
	memset(strts, 0, 100);
	gmtime_r((const time_t*)&timestamp, &utctime);
	strftime(strts, 100, "%a, %d %b %Y %H:%M:%S %z", &utctime); //string representation - RFC 822

	*mlen = strlen("Timestamp: ") + strlen(strts);
	ALLOC_AND_CLEAN(text, *mlen, unsigned char*);
	sprintf(text, "Timestamp: %s", strts);

	return text;
}

//Constructs the body of the UPDT message sent to the client which signals that the transaction has ended
//Such a message constists of a single "END" string
//However, for reasons of homogeneity, it has the same format as other updt messages:
//UPDT\r\nContent-Length: 3\r\nEND
unsigned char *make_updt_end_data() {
	unsigned char *text = NULL;
	ALLOC_AND_CLEAN(text, 3, unsigned char*);
	strcpy(text, "END");
	return text;
}

//constructs an UPDT message
//data must have already been prepared (e.g. b64 encoded, etc) - ready to be appended to a UPDT message
unsigned char *construct_updt_msg(unsigned char *data, int datalen) {
	unsigned char *text = NULL;
	unsigned char strcl[10]; //string representation of content length
	if (!data) {
		return NULL;
	}

	memset(strcl, 0, 10);
	sprintf(strcl, "%d", datalen);

	ALLOC_AND_CLEAN(text, strlen("UPDT P2PWNC/1.0\r\nContent-length: ") + strlen(strcl) + 2 + datalen, unsigned char *); //???+1: null byte in the end
	sprintf(text, "UPDT P2PWNC/1.0\r\nContent-length: %d\r\n%s", datalen, data);
	return text;
}

//parses an updt request message sent by a client
//returns -1 on error, sets the timestamp provided by the message
int parse_updt_req(unsigned char *text, unsigned long *timestamp) {
	unsigned int clen = 0; //content length

	unsigned char strts[32]; //date in rfc822 format (28bytes)
	memset(strts, 0, 32);
	struct tm utctime; //timestamp in broken-down format
	memset(&utctime, 0, sizeof(struct tm));

	//this is our current position inside the "text" structure
	unsigned char *pos = text;

	if ( strstr( pos, "UPDT P2PWNC/1.0\r\n") != (char*)pos) {
		//check message header - must be UPDT
		return -1;
	}

	//now check content length
	if (!(pos = strstr(pos, "\r\n")) ) {
		return -1;
	}
	pos += 2;
	if (strncasecmp(pos, "Content-length: ", strlen("Content-length: "))) {
		//malformed header
		return -1;
	}
	//move pos...
	pos += strlen("Content-length: ");
	//read content length
	if (!sscanf(pos, "%u", &clen)) {
		return -1;
	}

	//move pos after the next crlf
	if (!(pos = strstr(pos, "\r\n"))) {
		return -1;
	}
	pos += 2;

	//read timestamp
	//now read timestamp
	if (strncasecmp(pos, "Timestamp: ", strlen("Timestamp: ")) ) {
		return -1;
	}
	pos += strlen("Timestamp: ");
	//copy message timestamp to "strts" string
	if (!strncpy(strts, pos, 31) ) {
		//copy 31 bytes (see RFC822 date format...)
		return -1;
	}

	//read rfc822 timestamp from "strts" to a broken-down time format
	strptime(strts, "%a, %d %b %Y %H:%M:%S %z", &utctime);
	//set timestamp
	*timestamp = (unsigned long)my_timegm(&utctime);

	return 0;
}

//##################################################################################
//receipt info serialization
//Takes a PWNC_RCPT and returnes a byte array of the cryptography data in a serialized form in the order that follows:
//Cert Team Pub Key - Cert User Pub Key - Cert Signature - Provider Pub Key - Receipt Signature
unsigned char *serialize_receipt_enc_data(PWNC_RCPT *rcpt, int *rawlen) {
	unsigned char *retval = NULL;
	unsigned char *pos = NULL;

	*rawlen =
		rcpt->ConsumerCert->TeamPubKey->datalen +
		rcpt->ConsumerCert->UserPubKey->datalen +
		rcpt->ConsumerCert->TeamSig->datalen +
		rcpt->ProviderPubKey->datalen +
		rcpt->RcptSig->datalen,

	ALLOC_AND_CLEAN(
		retval,
		*rawlen,
		unsigned char*
		);

	pos = retval;

	memcpy(pos, rcpt->ConsumerCert->TeamPubKey->data, rcpt->ConsumerCert->TeamPubKey->datalen);
	pos += rcpt->ConsumerCert->TeamPubKey->datalen;

	memcpy(pos, rcpt->ConsumerCert->UserPubKey->data, rcpt->ConsumerCert->UserPubKey->datalen);
	pos += rcpt->ConsumerCert->UserPubKey->datalen;

	memcpy(pos, rcpt->ConsumerCert->TeamSig->data, rcpt->ConsumerCert->TeamSig->datalen);
	pos += rcpt->ConsumerCert->TeamSig->datalen;

	memcpy(pos, rcpt->ProviderPubKey->data, rcpt->ProviderPubKey->datalen);
	pos += rcpt->ProviderPubKey->datalen;

	memcpy(pos, rcpt->RcptSig->data, rcpt->RcptSig->datalen);

	return retval;
}

//takes a preallocated (using PWNC_RCPT_new()) rcpt and serialized crypto data
//deserializes data, and sets the fields of the receipts
//data is in raw form (not b64)
PWNC_RCPT *deserialize_receipt_enc_data(PWNC_RCPT *rcpt, unsigned char *data, int len, unsigned short receipt_algorithm, unsigned short bits) {
	unsigned char *pos = NULL;

	rcpt->ConsumerCert->TeamPubKey->algorithm = receipt_algorithm;
	rcpt->ConsumerCert->UserPubKey->algorithm = receipt_algorithm;
	rcpt->ConsumerCert->TeamSig->algorithm = receipt_algorithm;
	rcpt->RcptSig->algorithm = receipt_algorithm;

	rcpt->ConsumerCert->TeamPubKey->bits = bits;
	rcpt->ConsumerCert->UserPubKey->bits = bits;
	rcpt->ConsumerCert->TeamSig->bits = bits;
	rcpt->RcptSig->bits = bits;

	if (receipt_algorithm == PWNC_ALGO_RSA) {
		//the rcpt signature is RSA1024
		//so, we know that the receipt signature are the last 128 bytes of the serialzed data
		//also, the first 3*128 bytes are the consumer's cert
		rcpt->ConsumerCert->TeamPubKey->datalen = bits/8;
		rcpt->ConsumerCert->UserPubKey->datalen = bits/8;
		rcpt->ConsumerCert->TeamSig->datalen = bits/8;
		rcpt->RcptSig->datalen = bits/8;

		ALLOC_AND_CLEAN(rcpt->ConsumerCert->TeamPubKey->data, bits/8, unsigned char*);
		ALLOC_AND_CLEAN(rcpt->ConsumerCert->UserPubKey->data, bits/8, unsigned char*);
		ALLOC_AND_CLEAN(rcpt->ConsumerCert->TeamSig->data, bits/8, unsigned char*);
		ALLOC_AND_CLEAN(rcpt->RcptSig->data, bits/8, unsigned char*);

		//now copy the actual data to their fields in the structs...
		pos = data;
		memcpy(rcpt->ConsumerCert->TeamPubKey->data, pos, bits/8);
		pos += bits/8;

		memcpy(rcpt->ConsumerCert->UserPubKey->data, pos, bits/8);
		pos += bits/8;

		memcpy(rcpt->ConsumerCert->TeamSig->data, pos, bits/8);

		pos = data + (len - bits/8); //skip provider pub key and read Rcpt signature
		memcpy(rcpt->RcptSig->data, pos, bits/8);

		//set provider public key datalen and data
		rcpt->ProviderPubKey->datalen = len - 4*(bits/8);
		pos = data + 3*(bits/8);
		ALLOC_AND_CLEAN(rcpt->ProviderPubKey->data, rcpt->ProviderPubKey->datalen, unsigned char*);
		memcpy(rcpt->ProviderPubKey->data, pos, rcpt->ProviderPubKey->datalen);
	}
	else
	if (receipt_algorithm == PWNC_ALGO_ECC) {
		//ECC (160/192)
		rcpt->ConsumerCert->TeamPubKey->datalen = bits/4 + 1;
		rcpt->ConsumerCert->UserPubKey->datalen = bits/4 + 1;
		rcpt->ConsumerCert->TeamSig->datalen = bits/4;
		rcpt->RcptSig->datalen = bits/4;

		ALLOC_AND_CLEAN(rcpt->ConsumerCert->TeamPubKey->data, bits/4 + 1, unsigned char*);
		ALLOC_AND_CLEAN(rcpt->ConsumerCert->UserPubKey->data, bits/4 + 1, unsigned char*);
		ALLOC_AND_CLEAN(rcpt->ConsumerCert->TeamSig->data, bits/4, unsigned char*);
		ALLOC_AND_CLEAN(rcpt->RcptSig->data, bits/4, unsigned char*);

		pos = data;
		memcpy(rcpt->ConsumerCert->TeamPubKey->data, pos, bits/4 + 1);
		pos += (bits/4 + 1);

		memcpy(rcpt->ConsumerCert->UserPubKey->data, pos, bits/4 + 1);
		pos += (bits/4 + 1);

		memcpy(rcpt->ConsumerCert->TeamSig->data, pos, bits/4);

		pos = data + (len - (bits/4) ); //skip provider pub key and read Rcpt signature
		memcpy(rcpt->RcptSig->data, pos, bits/4);

		rcpt->ProviderPubKey->datalen = len - ( 2*(bits/4) + 2*(bits/4 + 1) );
		pos = data + 2*(bits/4 + 1) + (bits/4); //2 pubkeys + 1 ECDSA sig
		ALLOC_AND_CLEAN(rcpt->ProviderPubKey->data, rcpt->ProviderPubKey->datalen, unsigned char*);
		memcpy(rcpt->ProviderPubKey->data, pos, rcpt->ProviderPubKey->datalen);
	}
	else {
		return NULL;
	}

	//check providerpubkey length to determine the algorithm/bits of the key
	//moufa...- heuristic
	if (rcpt->ProviderPubKey->datalen == 41) {
		//ECC 160
		rcpt->ProviderPubKey->algorithm = PWNC_ALGO_ECC;
		rcpt->ProviderPubKey->bits = 160;
	}
	else
	if (rcpt->ProviderPubKey->datalen == 49) {
		//ECC 192
		rcpt->ProviderPubKey->algorithm = PWNC_ALGO_ECC;
		rcpt->ProviderPubKey->bits = 192;
	}
	else {
		//RSAxxxx
		rcpt->ProviderPubKey->algorithm = PWNC_ALGO_RSA;
		rcpt->ProviderPubKey->bits = rcpt->ProviderPubKey->datalen*8;
	}

	return rcpt;
}

//Serializes a certificate's info (teampk - userpk - team  sig)
//(To be used when constructing a CONN msg)
unsigned char *serialize_cert_enc_data(PWNC_CERT *cert, int *rawlen) {
	unsigned char *retval = NULL;
	unsigned char *pos = NULL;

	*rawlen = cert->TeamPubKey->datalen + cert->UserPubKey->datalen + cert->TeamSig->datalen;

	ALLOC_AND_CLEAN(retval, *rawlen, unsigned char*);

	pos = retval;

	memcpy(pos, cert->TeamPubKey->data, cert->TeamPubKey->datalen);
	pos += cert->TeamPubKey->datalen;

	memcpy(pos, cert->UserPubKey->data, cert->UserPubKey->datalen);
	pos += cert->UserPubKey->datalen;

	memcpy(pos, cert->TeamSig->data, cert->TeamSig->datalen);
	pos += cert->TeamSig->datalen;

	return retval;
}

//takes the serialized data as a parameter, as well as the cert algorithm and bits
//Allocates a new PWNC_CERT struct, deserializes data, sets its fields accordingly and returns the cert
//data are in raw form (not b64 encoded)
PWNC_CERT *deserialize_cert_enc_data(unsigned char *data, int len, unsigned short algorithm, unsigned short bits) {
	unsigned char *pos = NULL;
	PWNC_CERT *cert = PWNC_CERT_new();

	cert->TeamPubKey->algorithm = algorithm;
	cert->UserPubKey->algorithm = algorithm;
	cert->TeamSig->algorithm = algorithm;

	cert->TeamPubKey->bits = bits;
	cert->UserPubKey->bits = bits;
	cert->TeamSig->bits = bits;

	if (algorithm == PWNC_ALGO_RSA) {
		cert->TeamPubKey->datalen = bits/8;
		cert->UserPubKey->datalen = bits/8;
		cert->TeamSig->datalen = bits/8;

		ALLOC_AND_CLEAN(cert->TeamPubKey->data, bits/8, unsigned char*);
		ALLOC_AND_CLEAN(cert->UserPubKey->data, bits/8, unsigned char*);
		ALLOC_AND_CLEAN(cert->TeamSig->data, bits/8, unsigned char*);

		//copy data
		pos = data;
		memcpy(cert->TeamPubKey->data, pos, bits/8);
		pos += bits/8;

		memcpy(cert->UserPubKey->data, pos, bits/8);
		pos += bits/8;

		memcpy(cert->TeamSig->data, pos, bits/8);
	}
	else
	if (algorithm == PWNC_ALGO_ECC) {
		//ECC (160/192)
		cert->TeamPubKey->datalen = bits/4 + 1;
		cert->UserPubKey->datalen = bits/4 + 1;
		cert->TeamSig->datalen = bits/4;

		ALLOC_AND_CLEAN(cert->TeamPubKey->data, bits/4 + 1, unsigned char*);
		ALLOC_AND_CLEAN(cert->UserPubKey->data, bits/4 + 1, unsigned char*);
		ALLOC_AND_CLEAN(cert->TeamSig->data, bits/4, unsigned char*);

		pos = data;
		memcpy(cert->TeamPubKey->data, pos, bits/4 + 1);
		pos += (bits/4 + 1);

		memcpy(cert->UserPubKey->data, pos, bits/4 + 1);
		pos += (bits/4 + 1);

		memcpy(cert->TeamSig->data, pos, bits/4);
	}
	else {
		return NULL;
	}

	return cert;
}
