/**
* io.c - Input/Output operations
* (e.g. reading P2PWNC messages from sockets, reading/writing
* public keys from disk, etc)
*
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "io.h"

//################################################################################
//reads up to the next crlf  sequence in fp
int move_to_end_of_line(FILE *fp) {
	unsigned char cr = 0;
	unsigned char lf = 0;

	if (!fp) {
		return 0;
	}

	while ( 1 ) {
		if (fread(&cr, 1, 1, fp) == feof(fp)) { \
			return 0;
		}
		if (fread(&lf, 1, 1, fp) == feof(fp)) {
			return 0;
		}
		if ( (cr == '\r') && (lf == '\n') ){
			return 1;
		}
		else
		if (lf == '\r') {
			if (fseek(fp, -1, SEEK_CUR) == -1) {
				return -1;
			}
		}
	}
	return -1;
}
//################################################################################

//################################################################################
//reads a client certificate from a file
//format of cert file:<algorithm+bits>crlf<b64 encoded team pk>crlfcrlf<b64 encoded user pk>crlfcrlf<b64 encoded team signature>
PWNC_CERT* read_cert_file(const char* path) {
	FILE *fp = NULL;
	PWNC_CERT* cert = NULL;
	unsigned char *algorithm = NULL; //string representing algorithm + bits (first line of the file)
	unsigned char *text = NULL; //here the whole data are put first
	unsigned char *b64 = NULL; //b64 encoded data field (e.g. team pub key)
	unsigned char *raw = NULL; //raw data (afted b64 decoding) - this array is temporary
	unsigned char *pos = NULL;
	unsigned char *crlfpos = NULL;
	int rawlen = 0;
	unsigned short bits = 0;
	long startofdata = 0;
	long endofdata = 0;
	long datalen = 0;
	int failed = 0; //this variable shows whether the last call to DECODE_AND_COPY has failed or not

	fp = fopen(path, "r");
	if (fp) {
		//allocate and read algorithm identifier (e.g. RSA1024)
		ALLOC_AND_CLEAN(algorithm, 10, unsigned char*);
		cert = PWNC_CERT_new();
		if (fgets(algorithm, 10, fp)) {
			//extract algorithm type
			if (!strncasecmp(algorithm, "RSA", 3) ) {
				cert->TeamPubKey->algorithm = PWNC_ALGO_RSA;
				cert->UserPubKey->algorithm = PWNC_ALGO_RSA;
				cert->TeamSig->algorithm = PWNC_ALGO_RSA;
			}
			else
			if (!strncasecmp(algorithm, "ECC", 3) ) {
				cert->TeamPubKey->algorithm = PWNC_ALGO_ECC;
				cert->UserPubKey->algorithm = PWNC_ALGO_ECC;
				cert->TeamSig->algorithm = PWNC_ALGO_ECC;
			}
			else {
				SAFE_FREE(algorithm);
				SAFE_FREE_CERT(cert);
				fclose(fp);
				return NULL;
			}

			//now extract bits
			if (sscanf(algorithm + 3, "%hd", &bits) ) {
				cert->TeamPubKey->bits = bits;
				cert->UserPubKey->bits = bits;
				cert->TeamSig->bits = bits;
			}
			else {
				SAFE_FREE(algorithm);
				SAFE_FREE_CERT(cert);
				fclose(fp);
				return NULL;
			}
			SAFE_FREE(algorithm);
			//now we have to read data (team pk, user pk, team signature)
			//first calculate data lenght: endofdata - startofdata
			startofdata = ftell(fp) + 1;

			if (fseek(fp, 0L, SEEK_END)) {
				SAFE_FREE_CERT(cert);
				fclose(fp);
				return NULL;
			}
			else {
				endofdata = ftell(fp);
				datalen = endofdata - startofdata + 1;
				fseek(fp, startofdata - 1, SEEK_SET); //move pointer back to the start of data
			}

			//read all data from file
			//we allocate "datalen" bytes for the char array that will store the b64 encoded data (plus the crlfs between fields)
			ALLOC_AND_CLEAN(text, datalen, unsigned char*);
			int len = fread (text, 1, datalen, fp);
			if (len != datalen) {
				//error reading data
				SAFE_FREE_CERT(cert);
				SAFE_FREE(text);
				fclose(fp);
				return NULL;
			}
			//we are done with the file
			fclose(fp);

			//now parse data
			//find first crlfcrlf sequence (separates team pk from user pk)
			pos = text;
			//printf("%s\n", pos);
			if (!(crlfpos = strstr(pos, "\r\n\r\n"))) {
				SAFE_FREE_CERT(cert);
				SAFE_FREE(text);
				return NULL;
			}
			DECODE_AND_COPY(cert->TeamPubKey, pos, crlfpos, raw, b64, rawlen, cert, failed);
			if (failed) {
				return NULL;
			}

			pos = crlfpos + 4;//move pos to the next character after the 2 crlfs
			//find next 2*crlf
			if (!(crlfpos = strstr(pos, "\r\n\r\n")) ) {
				//error...
				SAFE_FREE(text);
				//SAFE_FREE(cert->TeamPubKey->data);
				SAFE_FREE_CERT(cert);
				return NULL;
			}

			DECODE_AND_COPY(cert->UserPubKey, pos, crlfpos, raw, b64, rawlen, cert, failed);

			if (failed) {
				SAFE_FREE(text);
				SAFE_FREE_CERT(cert);
				return NULL;
			}

			//decode and copy cert signature to cert struct
			pos = crlfpos + 4;//move pos to the next character after the 2 crlfs
			datalen -= (pos - text);
			DECODE_AND_COPY(cert->TeamSig, pos, pos + datalen, raw, b64, rawlen, cert, failed);

			if (failed) {
				SAFE_FREE(text);
				SAFE_FREE_CERT(cert);
				return NULL;
			}
		}
		else {
			fclose(fp);
			return NULL;
		}
	} //if (fp)

	SAFE_FREE(text);

	return cert;
}
//################################################################################

//################################################################################
//writes a cert struct to a file
//returns the number of characters written
int write_cert_file(PWNC_CERT *cert, const char* path) {
	FILE *fp = NULL;
	unsigned char *strcert; //text representing the certificate. It's the text to be written to the file
	unsigned char *consumerpk = NULL; //consumer public key
	unsigned char *teampk = NULL; //consumer team pk
	unsigned char *certsig = NULL; //user cert signature

	int teampklen, consumerpklen, certsiglen;
	int len = 0; //length of text to be written to file

	if (!cert || !path) {
		//NULL cert or path
		return -1;
	}
	fp = fopen(path, "w");

	unsigned char stralgo[5]; //string repr. of encryption algorithm
	memset(stralgo, 0, 5);
	unsigned char strbits[5]; //str representation of  key length in bits
	memset(strbits, 0, 5);

	//we suppose that the team pubkey "algorithm" and "bits" fields are the same for the user pub key and the cert signature
	switch (cert->TeamPubKey->algorithm) {
		case PWNC_ALGO_ECC:
			strcpy(stralgo, "ECC");
			break;
		case PWNC_ALGO_RSA: default:
			strcpy(stralgo, "RSA");
			break;
	}
	sprintf(strbits, "%hd", cert->TeamPubKey->bits);

	teampklen = calculate_b64_size(cert->TeamPubKey->datalen);
	consumerpklen = calculate_b64_size(cert->UserPubKey->datalen);
	certsiglen = calculate_b64_size(cert->TeamSig->datalen);

	teampk = (unsigned char*)malloc( teampklen );
	consumerpk = (unsigned char*)malloc( consumerpklen );
	certsig = (unsigned char*)malloc( certsiglen );

	memset(teampk, 0, teampklen);
	memset(consumerpk, 0, consumerpklen);
	memset(certsig, 0, certsiglen);

	//Generate b64 encoded byte arrays
	teampklen = b64_encode( cert->TeamPubKey->data, teampk, cert->TeamPubKey->datalen );
	consumerpklen = b64_encode( cert->UserPubKey->data, consumerpk, cert->UserPubKey->datalen );
	certsiglen = b64_encode( cert->TeamSig->data, certsig, cert->TeamSig->datalen );

	//Calculate total length of cert (text)
	len += strlen(stralgo) + strlen(strbits) + 2 + teampklen + 4 + consumerpklen + 4 + certsiglen;

	ALLOC_AND_CLEAN(strcert, len, unsigned char*);
	//sprintf data to cert
	sprintf(
		strcert,
		"%s%s\r\n%s\r\n\r\n%s\r\n\r\n%s",
		stralgo, strbits, teampk, consumerpk, certsig
		);

	SAFE_FREE(consumerpk);
	SAFE_FREE(teampk);
	SAFE_FREE(certsig);

	//now write "cert" to file
	len = fwrite(strcert, 1, len, fp);
	fclose(fp);
	SAFE_FREE(strcert);

	return len;
}
//################################################################################

//################################################################################
//Reads a public or private key from a file and returns the pwnc_key struct that has been constructed
//First line of the file: Algorithm identifier(name+bits, e.g. RSA1024)
struct pwnc_key* read_key_file(const char *path) {
	struct pwnc_key * pk = NULL;
	FILE *fp = NULL;
	unsigned char *algorithm = NULL; //string representing algorithm + bits (first line of the file)
	unsigned char *text = NULL; //here the whole data are put first
	unsigned char *b64 = NULL; //b64 encoded data field (e.g. team pub key)
	unsigned char *raw = NULL; //raw data (afted b64 decoding) - this array is temporary
	int rawlen = 0;
	unsigned short bits = 0;
	long startofdata = 0;
	long endofdata = 0;
	long datalen = 0;
	int failed = 0; //this variable shows whether the last call to DECODE_AND_COPY has failed or not

	fp = fopen(path, "r");
	if (fp) {
		//allocate and read algorithm identifier (e.g. RSA1024)
		ALLOC_AND_CLEAN(algorithm, 10, unsigned char*);
		ALLOC_AND_CLEAN(pk, sizeof(struct pwnc_key), struct pwnc_key*);
		if (fgets(algorithm, 10, fp)) {
			//extract algorithm type
			if (!strncasecmp(algorithm, "RSA", 3) ) {
				pk->algorithm = PWNC_ALGO_RSA;
			}
			else
			if (!strncasecmp(algorithm, "ECC", 3) ) {
				pk->algorithm = PWNC_ALGO_ECC;
			}
			else {
				SAFE_FREE(algorithm);
				SAFE_FREE_KEY(pk);
				fclose(fp);
				return NULL;
			}

			//now extract bits
			if (sscanf(algorithm + 3, "%hd", &bits) ) {
				pk->bits = bits;
			}
			else {
				SAFE_FREE(algorithm);
				SAFE_FREE_KEY(pk);
				fclose(fp);
				return NULL;
			}
			SAFE_FREE(algorithm);

			//read pub key data (b64 encoded)
			//first calculate data lenght: endofdata - startofdata
			startofdata = ftell(fp) + 1;

			if (fseek(fp, 0L, SEEK_END)) {
				SAFE_FREE_KEY(pk);
				fclose(fp);
				return NULL;
			}
			else {
				endofdata = ftell(fp);
				datalen = endofdata - startofdata + 1;
				fseek(fp, startofdata - 1, SEEK_SET); //move pointer back to the start of data
				//fseek(fp, startofdata, SEEK_SET); //move pointer back to the start of data
			}

			//read data from file
			//we allocate "datalen" bytes for the char array that will store the b64 encoded data
			ALLOC_AND_CLEAN(text, datalen, unsigned char*);
			int len = fread (text, 1, datalen, fp);
			if (len != datalen) {
				//error reading data
				SAFE_FREE_KEY(pk);
				SAFE_FREE(text);
				fclose(fp);
				return NULL;
			}
			//we are done with the file
			fclose(fp);

			DECODE_AND_COPY(pk, text, text + datalen, raw, b64, rawlen, pk, failed);

			if (failed) {
				SAFE_FREE(text);
				//SAFE_FREE(pk->data);
				SAFE_FREE_KEY(pk);
				return NULL;
			}
		}
		else {
			fclose(fp);
			return NULL;
		}
	} //if (fp)

	SAFE_FREE(text);

	return pk;
}
//################################################################################

//################################################################################
//Writes a pwnc_key structure to file
//returns number of bytes written
int write_key_file(struct pwnc_key *key, const char *path) {
	FILE *fp = NULL;
	unsigned char *strkey = NULL; //text to be written to the file
	unsigned char *pk = NULL; //consumer public key

	int pklen = 0;
	int len = 0; //length of text to be written to file

	if (!key || !path) {
		//NULL key or path
		return -1;
	}
	fp = fopen(path, "w");

	unsigned char stralgo[5]; //string repr. of encryption algorithm
	memset(stralgo, 0, 5);
	unsigned char strbits[5]; //str representation of  key length in bits
	memset(strbits, 0, 5);

	switch (key->algorithm) {
		case PWNC_ALGO_ECC:
			strncpy(stralgo, "ECC", 3);
			break;
		case PWNC_ALGO_RSA: default:
			strncpy(stralgo, "RSA", 3);
			break;
	}
	sprintf(strbits, "%hd", key->bits);

	pklen = calculate_b64_size(key->datalen);

	ALLOC_AND_CLEAN(pk, pklen, unsigned char*);

	//Generate b64 encoded byte array
	pklen = b64_encode( key->data, pk, key->datalen );

	//Calculate total length of key (text)
	len += strlen(stralgo) + strlen(strbits) + 2 + pklen;
//printf("len: %d\n", len);
	ALLOC_AND_CLEAN(strkey, len+2, unsigned char*);
	//sprintf data to strkey
	sprintf(
		strkey,
		"%s%s\r\n%s",
		stralgo, strbits, pk
		);

	SAFE_FREE(pk);

	//now write "strkey" to file
	len = fwrite(strkey, 1, len, fp);
	fclose(fp);
	SAFE_FREE(strkey);

	return len;
}
//################################################################################

//################################################################################
//Reads a message from a socket (sockfd), returns it (text)
//and sets mtype to the message type
unsigned char* read_message(int sockfd, int *mtype) {
	//first read the first 4 bytes (message type) and set mtype
	//then, read content length and the remaining message
	unsigned char *text_message = NULL;

	unsigned char *header = NULL; //content length header (this array is read from the socket)

	//message type string(read from the socket)
	unsigned char strtype[30];
	memset(strtype, 0, 30);

	unsigned char strcl[10]; //content length of message
	int clen = 0;
	int bytesread = 0;
	int totlen = 0; //total len of the text message

	int i = 0;

	//read message type (the first 4 bytes)
	bytesread = read(sockfd, strtype, 4 + strlen(" P2PWNC/1.0"));
	if (bytesread != -1) {
		//save message type
		if (!strncasecmp(strtype, "CACK P2PWNC/1.0", 4 + strlen(" P2PWNC/1.0") )) {
			*mtype = PWNC_MSG_CACK;
		}
		else
		if (!strncasecmp(strtype, "QRSP P2PWNC/1.0", 4 + strlen(" P2PWNC/1.0") )) {
			*mtype = PWNC_MSG_QRSP;
		}
		else
		if (!strncasecmp(strtype, "RRSP P2PWNC/1.0", 4 + strlen(" P2PWNC/1.0") )) {
			*mtype = PWNC_MSG_RRSP;
		}
		else
		if (!strncasecmp(strtype, "RCPT P2PWNC/1.0", 4 + strlen(" P2PWNC/1.0") )) {
			*mtype = PWNC_MSG_RCPT;
		}
		else
		if (!strncasecmp(strtype, "CONN P2PWNC/1.0", 4 + strlen(" P2PWNC/1.0") )) {
			*mtype = PWNC_MSG_CONN;
		}
		else
		if (!strncasecmp(strtype, "RREQ P2PWNC/1.0", 4 + strlen(" P2PWNC/1.0") )) {
			*mtype = PWNC_MSG_RREQ;
		}
		else
		if (!strncasecmp(strtype, "QUER P2PWNC/1.0", 4 + strlen(" P2PWNC/1.0") )) {
			*mtype = PWNC_MSG_QUER;
		}
		else
		if (!strncasecmp(strtype, "UPDT P2PWNC/1.0", 4 + strlen(" P2PWNC/1.0") )) {
			*mtype = PWNC_MSG_UPDT;
		}

		ALLOC_AND_CLEAN(header, strlen("\r\nContent-length: ") + 1, unsigned char*);
		bytesread = read(sockfd, header, strlen("\r\nContent-length: "));
		if (bytesread != -1) {
			//first check header
			if (strcasecmp(header, "\r\nContent-length: ")) {
				//malformed header
				SAFE_FREE(header);
				return NULL;
			}
			//read content length
			//actually, we read byte after byte until we find a crlf
			memset(strcl, 0, 10);
			i = 0;
			while ( !strstr(strcl, "\r\n") && (i < 10) ) {
				if (read(sockfd, &strcl[i], 1) == -1) {
					//error reading character
					SAFE_FREE(header);
					return NULL;
				}
				i++;
			}
			if (!sscanf(strcl, "%d\r\n", &clen)) {
				//malformed clen
				SAFE_FREE(header);
				return NULL;
			}

			//finally, read the remaining text message into "text_message" (which will eventually be returned)
			totlen = 4 + strlen(" P2PWNC/1.0") + strlen(header) + strlen(strcl) + clen + 1; //total length of message ("4 + ...": mtype, "+1": trailing \0)

			ALLOC_AND_CLEAN(text_message, totlen, unsigned char*);

			//copy the fields that are already read
			sprintf(text_message, "%s%s%s", strtype, header, strcl);
			//"append" the remaining bytes to text_message (start writing from position text_message+sizeof(already read))
			bytesread = read(sockfd, text_message + strlen(text_message), clen);
			if (bytesread != clen) {
				//error reading data from socket: bytesread must be == with the calculated text message length
				SAFE_FREE(text_message);
				SAFE_FREE(header);
				return NULL;
			}
		}
		else {
			SAFE_FREE(header);
			return NULL;
		}
	}
	else {
		return NULL;
	}
	SAFE_FREE(header);
	return text_message;
}
//################################################################################

//################################################################################
//writes a receipt to the"path" file
//fmode: "w":new file, "a": append to "path"
//format: text format (b64 encoded fields - PWNC_FORMAT_B64), or binary format (raw byte fields - PWNC_FORMAT_RAW)
//the file starts with a special byte if it is in raw format, else its supposed to be plain text
int write_rcpt_file(PWNC_RCPT *rcpt, const char* fmode, const char* path, int format) {
	FILE *fp = NULL;
	unsigned char *text = NULL;
	unsigned char *consumerpk = NULL; //consumer public key
	unsigned char *teampk = NULL; //consumer team pk
	unsigned char *certsig = NULL; //user cert signature
	unsigned char *providerpk = NULL; //provider public key
	unsigned char *rsig = NULL; //receipt signature
	//lengths of the above char arrays
	unsigned int teampklen, consumerpklen, providerpklen, certsiglen, rsiglen;

	unsigned char stralgo[5]; //string repr. of encryption algorithm
	memset(stralgo, 0, 5);
	unsigned char strbits[5]; //str representation of  key length in bits
	memset(strbits, 0, 5);
	unsigned char strts[32]; //string representation of timestamp (UTC)
	memset(strts, 0, 32);
	unsigned char strweight[20]; //string representation of weight
	memset(strweight, 0, 20);
	unsigned char strcl[10]; //string repr. of content length
	memset(strcl, 0, 10);

	unsigned char flag = 0;
	struct tm utctime; //timestamp in UTC

	int len = 0;

	//check args
	if (!rcpt || !fmode || !path) {
		return -1;
	}

	//check fmode
	if (strcmp(fmode, "w") && strcmp(fmode, "a")) {
		//invalid mode
		return -1;
	}

	fp = fopen(path, fmode);
	if (!fp) {
		return -1;
	}

	switch (format) {
		case PWNC_FORMAT_RAW:
			//format:
			//algorithmid (ushort), bits(ushort), timestamp, weight, cert(pk1len, pk1data, pk2len, pk2data, siglen, sigdata), prov pk
			//RAW formatted receipts start with a \0 byte
			len += fwrite( &flag, sizeof(unsigned char), 1, fp);
			len += fwrite( &(rcpt->ConsumerCert->TeamPubKey->algorithm), sizeof(unsigned short), 1, fp);
			len += fwrite( &(rcpt->ConsumerCert->TeamPubKey->bits), sizeof(unsigned short), 1, fp);
			len += fwrite( &(rcpt->timestamp), sizeof(unsigned long), 1, fp);
			len += fwrite( &(rcpt->weight), sizeof(unsigned long), 1, fp);

			len += fwrite( &(rcpt->ConsumerCert->TeamPubKey->datalen), sizeof(unsigned int), 1, fp);
			len += fwrite( rcpt->ConsumerCert->TeamPubKey->data, sizeof(unsigned char), rcpt->ConsumerCert->TeamPubKey->datalen, fp);

			len += fwrite( &(rcpt->ConsumerCert->UserPubKey->datalen), sizeof(unsigned int), 1, fp);
			len += fwrite( rcpt->ConsumerCert->UserPubKey->data, sizeof(unsigned char), rcpt->ConsumerCert->UserPubKey->datalen, fp);

			len += fwrite( &(rcpt->ConsumerCert->TeamSig->datalen), sizeof(unsigned int), 1, fp);
			len += fwrite( rcpt->ConsumerCert->TeamSig->data, sizeof(unsigned char), rcpt->ConsumerCert->TeamSig->datalen, fp);

			len += fwrite( &(rcpt->ProviderPubKey->datalen), sizeof(unsigned int), 1, fp);
			len += fwrite( rcpt->ProviderPubKey->data, sizeof(unsigned char), rcpt->ProviderPubKey->datalen, fp);

			len += fwrite( &(rcpt->RcptSig->datalen), sizeof(unsigned int), 1, fp);
			len += fwrite( rcpt->RcptSig->data, sizeof(unsigned char), rcpt->RcptSig->datalen, fp);
			fclose(fp);
			return len;

		case PWNC_FORMAT_B64:
			//receipts are separated by 3crlfs

			switch (rcpt->ConsumerCert->TeamPubKey->algorithm) {
				case PWNC_ALGO_ECC:
					strcpy(stralgo, "ECC");
					break;
				case PWNC_ALGO_RSA: default:
					strcpy(stralgo, "RSA");
					break;
			}
			sprintf(strbits, "%hd", rcpt->ConsumerCert->TeamPubKey->bits);

			teampklen = calculate_b64_size(rcpt->ConsumerCert->TeamPubKey->datalen);
			consumerpklen = calculate_b64_size(rcpt->ConsumerCert->UserPubKey->datalen);
			certsiglen = calculate_b64_size(rcpt->ConsumerCert->TeamSig->datalen);
			providerpklen = calculate_b64_size(rcpt->ProviderPubKey->datalen);
			rsiglen = calculate_b64_size(rcpt->RcptSig->datalen);

			ALLOC_AND_CLEAN(teampk, teampklen, unsigned char*);
			ALLOC_AND_CLEAN(consumerpk, consumerpklen, unsigned char*);
			ALLOC_AND_CLEAN(certsig, certsiglen, unsigned char*);
			ALLOC_AND_CLEAN(providerpk, providerpklen, unsigned char*);
			ALLOC_AND_CLEAN(rsig, providerpklen, unsigned char*);

			//Generate b64 encoded byte arrays
			teampklen = b64_encode( rcpt->ConsumerCert->TeamPubKey->data, teampk, rcpt->ConsumerCert->TeamPubKey->datalen );
			consumerpklen = b64_encode( rcpt->ConsumerCert->UserPubKey->data, consumerpk, rcpt->ConsumerCert->UserPubKey->datalen );
			certsiglen = b64_encode( rcpt->ConsumerCert->TeamSig->data, certsig, rcpt->ConsumerCert->TeamSig->datalen );
			providerpklen = b64_encode( rcpt->ProviderPubKey->data, providerpk, rcpt->ProviderPubKey->datalen );
			rsiglen = b64_encode( rcpt->RcptSig->data, rsig, rcpt->RcptSig->datalen);


//			localtime_r( (const time_t*)&(rcpt->timestamp), &utctime );
//			strftime(strts, 32, "%a, %d %b %Y %H:%M:%S %z", &utctime); //string representation - RFC 822
			gmtime_r((const time_t*)&(rcpt->timestamp), &utctime);
			strftime(strts, 32, "%a, %d %b %Y %H:%M:%S %z", &utctime); //string representation - RFC 822

			//construct string representation of weight
			sprintf(strweight, "%lu", rcpt->weight);

			//Calculate length
			len = strlen(stralgo) + strlen(strbits) + 2 + strlen(strts) + 2 + strlen(strweight) + 2 + teampklen + 4 + consumerpklen + 4 + certsiglen + 4 + providerpklen + 4 + rsiglen + 6;
			//+6: 3 trailing crlfs...

			ALLOC_AND_CLEAN(text, len + 1, unsigned char*);
			//sprintf data to cert
			sprintf(
				text,
				"%s%s\r\n%s\r\n%s\r\n%s\r\n\r\n%s\r\n\r\n%s\r\n\r\n%s\r\n\r\n%s\r\n\r\n\r\n",
				stralgo, strbits, strts, strweight, teampk, consumerpk, certsig, providerpk, rsig
				);

			SAFE_FREE(consumerpk);
			SAFE_FREE(teampk);
			SAFE_FREE(certsig);
			SAFE_FREE(providerpk);
			SAFE_FREE(rsig);

			len = fwrite(text, 1, len, fp);
			fclose(fp);
			return len;
		default:
			return -1;
	}
	return -1;
}
//################################################################################

//################################################################################
//PROSOXI!!!!!!!!!!! --- DON'T USE IT
//TODO:++++++
///Provlhma otan diavazei receipts b64-encoded
//Diavazei B64_LINE_LEN bytes a time (mhkos grammhs) alla den exw ypologisei ti paizei me tis allages grammhs anamesa sta fields toy receipt
//Proswrina: mono raw encoding sta receipt files
PWNC_RCPT* read_rcpt_file_old(FILE* fp, long offset) {
	PWNC_RCPT *rcpt = NULL;
	unsigned char first_byte;
	unsigned short algorithm = 0;
	unsigned short bits = 0;
	unsigned int datalen = 0;

	unsigned char *stralgo = NULL;
	unsigned char strbits[6];
	unsigned char strts[32];
	unsigned char strcheckts[32];

	struct tm utctime;

	unsigned char *text = NULL; //text where the b64 encoded data will be stored
	unsigned char *pos = NULL;
	int curlen = 0; //current length of text. Used when reallocing memory (reading line by line from fp)
	int bytes_read = 0;//bytes read in an iteration (when reading b64-encoded data)

	int i = 0;

	if (fseek(fp, offset, SEEK_SET) == -1) {
		return NULL;
	}

	if (fread(&first_byte, sizeof(unsigned char), 1, fp) != 1) {
		return NULL;
	}

	//rcpt = (PWNC_RCPT*)malloc(sizeof(PWNC_RCPT));
	rcpt = PWNC_RCPT_new();

	if (first_byte == 0) {
		//raw mode
		fread(&algorithm, sizeof(unsigned short), 1, fp);
		fread(&bits, sizeof(unsigned short), 1, fp);
		fread( &(rcpt->timestamp), sizeof(unsigned long), 1, fp);
		fread( &(rcpt->weight), sizeof(unsigned long), 1, fp);

		fread(&datalen, sizeof(unsigned int), 1, fp);
		ALLOC_AND_CLEAN(rcpt->ConsumerCert->TeamPubKey->data, datalen+1, unsigned char*);
		fread(rcpt->ConsumerCert->TeamPubKey->data, sizeof(unsigned char), datalen, fp);
		rcpt->ConsumerCert->TeamPubKey->datalen = datalen;

		fread(&datalen, sizeof(unsigned int), 1, fp);
		ALLOC_AND_CLEAN(rcpt->ConsumerCert->UserPubKey->data, datalen+1, unsigned char*);
		fread(rcpt->ConsumerCert->UserPubKey->data, sizeof(unsigned char), datalen, fp);
		rcpt->ConsumerCert->UserPubKey->datalen = datalen;

		fread(&datalen, sizeof(unsigned int), 1, fp);
		ALLOC_AND_CLEAN(rcpt->ConsumerCert->TeamSig->data, datalen+1, unsigned char*);
		fread(rcpt->ConsumerCert->TeamSig->data, sizeof(unsigned char), datalen, fp);
		rcpt->ConsumerCert->TeamSig->datalen = datalen;

		fread(&datalen, sizeof(unsigned int), 1, fp);
		ALLOC_AND_CLEAN(rcpt->ProviderPubKey->data, datalen + 1, unsigned char*);
		fread(rcpt->ProviderPubKey->data, sizeof(unsigned char), datalen, fp);
		rcpt->ProviderPubKey->datalen = datalen;

		fread(&datalen, sizeof(unsigned int), 1, fp);
		ALLOC_AND_CLEAN(rcpt->RcptSig->data, datalen + 1, unsigned char*);
		fread(rcpt->RcptSig->data, sizeof(unsigned char), datalen, fp);
		rcpt->RcptSig->datalen = datalen;

		rcpt->ConsumerCert->TeamPubKey->algorithm = algorithm;
		rcpt->ConsumerCert->TeamPubKey->bits = bits;
		rcpt->ConsumerCert->UserPubKey->algorithm = algorithm;
		rcpt->ConsumerCert->UserPubKey->bits = bits;
		rcpt->ConsumerCert->TeamSig->algorithm = algorithm;
		rcpt->ConsumerCert->TeamSig->bits = bits;
		rcpt->ProviderPubKey->algorithm = algorithm;
		rcpt->ProviderPubKey->bits = bits;
		rcpt->RcptSig->algorithm = algorithm;
		rcpt->RcptSig->bits = bits;
	}
	else {
		ALLOC_AND_CLEAN(stralgo, 4, unsigned char*);

		memset(strbits, 0, 10);
		//text format...
		//read algorithm identifier (3bytes)
		*stralgo = first_byte;
		if (fread( stralgo + 1, 1, 2, fp) != 2) {
			SAFE_FREE(stralgo);
			SAFE_FREE_RCPT(rcpt);
			return NULL;
		}

		if (strncasecmp(stralgo, "RSA", 3)) {
			if (strncasecmp(stralgo, "ECC", 3)) {
				SAFE_FREE(stralgo);
				SAFE_FREE_RCPT(rcpt);
				return NULL;
			}
			else {
				algorithm = PWNC_ALGO_ECC;
			}
		}
		else {
			algorithm = PWNC_ALGO_RSA;
		}
		SAFE_FREE(stralgo);

		//read bits (read up to the first crlf
		i = 0;
		while ( !strstr(strbits, "\r\n") && (i < 6) ) {
			if (fread(&strbits[i], 1, 1, fp) == -1) {
				SAFE_FREE_RCPT(rcpt);
				return NULL;
			}
			i++;
		}
		if (!sscanf(strbits, "%hd\r\n", &bits)) {
			SAFE_FREE_RCPT(rcpt);
			return NULL;
		}

		//read timestamp
		memset(&utctime, 0, sizeof(struct tm));
		memset(strts, 0, 32);
		memset(strcheckts, 0, 32);

		if (fread(strts, sizeof(unsigned char), 31, fp) != 31) {
			SAFE_FREE_RCPT(rcpt);
			return NULL;
		}
		//read rfc822 timestamp from "strts" to a broken-down time format
		strptime(strts, "%a, %d %b %Y %H:%M:%S %z", &utctime);
		//set timestamp
		rcpt->timestamp = (unsigned long)my_timegm(&utctime);
		//timestamp check
//		localtime_r( (const time_t*)&(rcpt->timestamp), &utctime );
		gmtime_r((const time_t*)&(rcpt->timestamp), &utctime);
		strftime(strcheckts, 32, "%a, %d %b %Y %H:%M:%S %z", &utctime);
		if (strcasecmp(strts, strcheckts)) {
			//strts<>checkstr
			SAFE_FREE_RCPT(rcpt);
			return NULL;
		}

		//read weight
		//first read up to next crlf
		if (!move_to_end_of_line(fp)) {
			SAFE_FREE_RCPT(rcpt);
			return NULL;
		}
		//read weight
		fscanf(fp, "%lu\r\n", &(rcpt->weight));

		//now read all b64 encoded data to a big string
		//and operate on that string
		//since each line is at most B64_LINE_LEN long, we read that many bytes each time , realloc string, append bytes
		curlen = 0;
		text = (unsigned char*)malloc(B64_LINE_LEN + 2); //+2: crlf
		memset(text, 0, B64_LINE_LEN + 2);
		while(1) {
			//read from file in B64_LINE_LEN-byte chunks
			bytes_read = fread(text + curlen, sizeof(unsigned char), B64_LINE_LEN+2, fp);
			if ( (pos = strstr(text, "\r\n\r\n\r\n")) ) {
				//found end of receipt
				//rewind fp to the actual end of the data, that is right after the 3 crlfs
				fseek (fp, -(bytes_read - (pos - text - curlen) )+6 , SEEK_CUR);
				*pos = 0; //end of string
				break;
			}
			if (!(text = (unsigned char*)realloc(text, curlen + B64_LINE_LEN + 2))) {
				SAFE_FREE(text);
				SAFE_FREE_RCPT(rcpt);
				return NULL;
			}
			curlen += bytes_read;
		}

		//now we have to parse "text" to extract the information
		//decode b64 data
		unsigned char *raw_data = NULL;
		unsigned char *b64_data = NULL;
		int rawlen = 0;
		int b64_len = pos - text;
		ALLOC_AND_CLEAN(b64_data, b64_len, unsigned char*);
		ALLOC_AND_CLEAN(raw_data, b64_len, unsigned char*); //alloc more space than necessary, but anyway....
		strncpy(b64_data, pos, b64_len);
		if ( (rawlen = b64_decode(b64_data, raw_data)) == -1) {
			//actual decoding
			SAFE_FREE_RCPT(rcpt);
			SAFE_FREE(raw_data);
			SAFE_FREE(b64_data);
			SAFE_FREE(text);
			return NULL;
		}
		SAFE_FREE(b64_data);

		if (!deserialize_receipt_enc_data(rcpt, raw_data, rawlen, algorithm, bits)) {
			SAFE_FREE_RCPT(rcpt);
			SAFE_FREE(raw_data);
			SAFE_FREE(text);
			return NULL;
		}
		SAFE_FREE(text);
		SAFE_FREE(raw_data);

		return rcpt;
	}

	return rcpt;
}

//################################################################################

//stores the receipt in compact format in the fp stream starting from "offset"
int io_store_receipt(PWNC_RCPT *rcpt, FILE *fp, int offset) {
	int len = 0;
	int totlen = 0;
	unsigned char null_byte = 0;
	unsigned short rlen = 0; //total receipt length (bytes that will be written on disk)

	if (!fp) {
		return -1;
	}

	if (fseek(fp, offset, SEEK_SET) == -1) {
		return -1;
	}

	if (rcpt->RcptSig->algorithm == PWNC_ALGO_RSA) {
		null_byte = 1;
		//calculate rlen
		rlen = sizeof(unsigned char)  + 2*sizeof(unsigned short) + 2*sizeof(unsigned long) + 4*(rcpt->RcptSig->bits/8) + rcpt->ProviderPubKey->datalen;
	}
	else {
		null_byte = 0;
		//calculate rlen
		//1 byte for the algo id, 2 bytes for the bitlen, 2 bytes for the rlen, 3*(sizeof(ECC public key) + sizeof (ECC signature) + datalen of provider pk
		rlen = sizeof(unsigned char)  + 2*sizeof(unsigned short) + 2*sizeof(unsigned long) + 3*(rcpt->RcptSig->bits/4 + 1) + rcpt->RcptSig->bits/4 + rcpt->ProviderPubKey->datalen;
	}

	//write algorithm id (null_byte - 0 for ECC, 1 for RSA)
	if ( (len = fwrite( &null_byte, sizeof(unsigned char), 1, fp)) != 1) {
		return -1;
	}
	totlen += len;

	//write bits
	if ( (len = fwrite( &(rcpt->ConsumerCert->TeamPubKey->bits), sizeof(unsigned short), 1, fp)) != 1 ) {
		return -1;
	}
	totlen += len;

	//write receipt len
	if ( (len = fwrite( &rlen, sizeof(unsigned short), 1, fp)) != 1 ) {
		return -1;
	}
	totlen += len;

	//write timestamp
	if ( (len = fwrite( &(rcpt->timestamp), sizeof(unsigned long), 1, fp)) != 1) {
		return -1;
	}
	totlen += len;

	//write receipt weight
	if ( (len = fwrite( &(rcpt->weight), sizeof(unsigned long), 1, fp)) != 1) {
		return -1;
	}
	totlen += len;

	//write consumer team pub key
	if ( (len = fwrite( rcpt->ConsumerCert->TeamPubKey->data, sizeof(unsigned char), rcpt->ConsumerCert->TeamPubKey->datalen, fp))
		!= rcpt->ConsumerCert->TeamPubKey->datalen) {
		return -1;
	}
	totlen += len;

	//write consumer public key
	if ( (len = fwrite( rcpt->ConsumerCert->UserPubKey->data, sizeof(unsigned char), rcpt->ConsumerCert->UserPubKey->datalen, fp))
		 != rcpt->ConsumerCert->UserPubKey->datalen ) {
	 	return -1;
	}
	totlen += len;

	//write cert signature
	if ( (len = fwrite( rcpt->ConsumerCert->TeamSig->data, sizeof(unsigned char), rcpt->ConsumerCert->TeamSig->datalen, fp))
		!= rcpt->ConsumerCert->TeamSig->datalen ) {
		return -1;
	}
	totlen += len;

	//write provider public key
	if ( (len = fwrite( rcpt->ProviderPubKey->data, sizeof(unsigned char), rcpt->ProviderPubKey->datalen, fp))
		!= rcpt->ProviderPubKey->datalen ) {
		return -1;
	}
	totlen += len;

	//write receipt signature
	if ( (len = fwrite( rcpt->RcptSig->data, sizeof(unsigned char), rcpt->RcptSig->datalen, fp))
		!= rcpt->RcptSig->datalen ) {
		return -1;
	}
	totlen += len;

	return totlen;
}

//reads a receipt from disk
//A receipt in its compact (disk) format is as follows
//Algorithm id (1st byte, 0 for ECC, 1 for RSA)
//Keys bit length (e.g. 1024 for RSA1024) (2 bytes)
//Receipt total length (2 bytes)
//timestamp (4 bytes)
//weight (4 bytes)
//consumer cert
//provider pubkey
//receipt signature
PWNC_RCPT* read_rcpt_file(FILE* fp, long offset) {
	PWNC_RCPT *rcpt = NULL;
	unsigned short algorithm = 0;
	unsigned char algo_byte = 0; //first byte of the receipt, indicates the algo used
	unsigned short bits = 0;
	unsigned int datalen = 0;
	int field_len = 0;
	int sig_len = 0;
	int prov_pkey_len = 0;
	unsigned short rlen = 0;
	int bytes_read = 0;
	struct tm utctime;

	if (fseek(fp, offset, SEEK_SET) == -1) {
		return NULL;
	}

	//read algorithm id (unsigned char)
	//0 for ecc, 1 for rsa
	if (fread(&algo_byte, sizeof(unsigned char), 1, fp) != 1) {
		return NULL;
	}
	if (algo_byte) {
		algorithm = PWNC_ALGO_RSA;
	}
	else {
		algorithm = PWNC_ALGO_ECC;
	}

	rcpt = PWNC_RCPT_new();

	fread(&bits, sizeof(unsigned short), 1, fp); //keys, etc bit length
	fread(&rlen, sizeof(unsigned short), 1, fp); //read size of receipt (== total bytes of data that will be read)

	fread( &(rcpt->timestamp), sizeof(unsigned long), 1, fp); //receipt timestamp
	fread( &(rcpt->weight), sizeof(unsigned long), 1, fp); //receipt weight

	if (algo_byte) {
		//rsa public keys and signatures: if (RSA1024) ==> length = 1024/8
		field_len = bits/8;
		sig_len = bits/8;
		//the provider's public key may be of unknown algorithm
		//Thus, subtract the length of the other fields from "rlen"
		prov_pkey_len = rlen - sizeof(unsigned char) - 2*sizeof(unsigned short) - 2*sizeof(unsigned long) - 4*field_len;
	}
	else {
		//ecc public keys: e.g. if (ECC160) => length of public keys; 160/4 + 1
		field_len = bits/4 + 1;
		sig_len = bits/4;
		//same as before
		prov_pkey_len = rlen - sizeof(unsigned char) - 2*sizeof(unsigned short) - 2*sizeof(unsigned long) - 3*field_len - sig_len;
	}

	//read consumer team pub key
	ALLOC_AND_CLEAN(rcpt->ConsumerCert->TeamPubKey->data, field_len, unsigned char*);
	bytes_read = fread(rcpt->ConsumerCert->TeamPubKey->data, sizeof(unsigned char), field_len, fp);
	rcpt->ConsumerCert->TeamPubKey->datalen = bytes_read;

	//read consumer pub key
	ALLOC_AND_CLEAN(rcpt->ConsumerCert->UserPubKey->data, field_len, unsigned char*);
	bytes_read = fread(rcpt->ConsumerCert->UserPubKey->data, sizeof(unsigned char), field_len, fp);
	rcpt->ConsumerCert->UserPubKey->datalen = bytes_read;

	//read certificate signature
	ALLOC_AND_CLEAN(rcpt->ConsumerCert->TeamSig->data, sig_len, unsigned char*);
	bytes_read = fread(rcpt->ConsumerCert->TeamSig->data, sizeof(unsigned char), sig_len, fp);
	rcpt->ConsumerCert->TeamSig->datalen = bytes_read;

	//read provider public key
	ALLOC_AND_CLEAN(rcpt->ProviderPubKey->data, prov_pkey_len, unsigned char*);
	bytes_read = fread(rcpt->ProviderPubKey->data, sizeof(unsigned char), prov_pkey_len, fp);
	rcpt->ProviderPubKey->datalen = bytes_read;

	//finally, read receipt signature
	ALLOC_AND_CLEAN(rcpt->RcptSig->data, sig_len, unsigned char*);
	bytes_read = fread(rcpt->RcptSig->data, sizeof(unsigned char), sig_len, fp);
	rcpt->RcptSig->datalen = bytes_read;

	rcpt->ConsumerCert->TeamPubKey->algorithm = algorithm;
	rcpt->ConsumerCert->TeamPubKey->bits = bits;
	rcpt->ConsumerCert->UserPubKey->algorithm = algorithm;
	rcpt->ConsumerCert->UserPubKey->bits = bits;
	rcpt->ConsumerCert->TeamSig->algorithm = algorithm;
	rcpt->ConsumerCert->TeamSig->bits = bits;
	rcpt->ProviderPubKey->algorithm = algorithm; //ignored...
	rcpt->ProviderPubKey->bits = bits; //ignored
	rcpt->RcptSig->algorithm = algorithm;
	rcpt->RcptSig->bits = bits;

	return rcpt;
}
