/**
* rdb_engine.c - Receipt repository module
*
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*
* In addition, as a special exception, the copyright holders
* give permission to link the code of this program with the
* OpenSSL toolkit (or with modified versions of OpenSSL that use the
* same license as OpenSSL, version 0.9.8a), and distribute linked
* combinations including the two.  You must obey the GNU General Public License
* in all respects for all of the code used other than OpenSSL.  If you modify
* this file, you may extend this exception to your version of the
* file, but you are not obligated to do so.  If you do not wish to
* do so, delete this exception statement from your version.
*/

#include "rdb_engine.h"

int sat_push_ops = 0; long sat_time = 0;
int non_sat_push_ops = 0; long non_sat_time = 0;
int relabel_ops = 0; long relabel_time =0;

int global_relabel_threshold = 0;

//generates a hash
unsigned long GenerateHash(unsigned char *tobehashed, int datalen) {
	unsigned long hash = 5381;
	int i = 0;
	for (i=0; i<datalen; i++) {
		hash = ((hash << 5) + hash) + tobehashed[i]; // hash*33 + c
	}
	return hash;
}

unsigned long team_hash(const PWNC_TEAM *to_hash) {
	return GenerateHash( to_hash->team_pk->data, to_hash->team_pk->datalen);
}

int team_cmp(const PWNC_TEAM *arg1, const PWNC_TEAM *arg2) {
	return ( GenerateHash(arg1->team_pk->data, arg1->team_pk->datalen) - GenerateHash(arg2->team_pk->data, arg2->team_pk->datalen) );
}

unsigned char *serialize_rcpt_identifier(const PWNC_RCPT *rcpt, int *len) {
	unsigned char *receipt_info = NULL;
	unsigned char *ts = NULL; //big endian timestamp representation
	unsigned char *pos = NULL;
	int rlen = 0; //length of the receipt data that will be hashed

	rlen = rcpt->ConsumerCert->TeamPubKey->datalen + rcpt->ConsumerCert->UserPubKey->datalen + rcpt->ProviderPubKey->datalen + sizeof(unsigned long);
	ts = endian_conv( (unsigned char*)&rcpt->timestamp, sizeof(unsigned long));

	ALLOC_AND_CLEAN(receipt_info, rlen, unsigned char*);

	memcpy(receipt_info, rcpt->ConsumerCert->TeamPubKey->data, rcpt->ConsumerCert->TeamPubKey->datalen);
	pos = receipt_info + rcpt->ConsumerCert->TeamPubKey->datalen;

	memcpy(pos, rcpt->ConsumerCert->UserPubKey->data, rcpt->ConsumerCert->UserPubKey->datalen);
	pos += rcpt->ConsumerCert->UserPubKey->datalen;

	memcpy(pos, rcpt->ProviderPubKey->data, rcpt->ProviderPubKey->datalen);
	pos += rcpt->ProviderPubKey->datalen;

	memcpy(pos, ts, sizeof(unsigned long));
	SAFE_FREE(ts);
	*len = rlen;
	return receipt_info;
}

//calculates, allocates space for and returns a receipt's SHA-1 hash
unsigned char *rcpt_SHA1(const PWNC_RCPT *rcpt) {
	unsigned char *rinfo = NULL;
	unsigned char *message_digest = NULL;
	int rlen = 0;
	ALLOC_AND_CLEAN(message_digest, 20, unsigned char*);

	rinfo = serialize_rcpt_identifier(rcpt, &rlen);

	//calculate receipt info SHA-1 hash
	if (!SHA1(rinfo, rlen, message_digest)) {
		fprintf(stderr, "Error generating SHA1 digest\n");
		SAFE_FREE(message_digest);
		return NULL;
	}
	SAFE_FREE(rinfo);
	return message_digest;
}

//hashes the SHA1-hash...
unsigned long receipt_hash(const PWNC_RCPT *to_hash) {
	unsigned char *rhash = NULL;
	unsigned long retval = 0;
	rhash = rcpt_SHA1(to_hash); //generates receipt sha-1 hash {consumer cert, providerpk, timestamp}
	retval = GenerateHash (rhash, RHASH_LEN);
	SAFE_FREE(rhash);
	return retval;
}

int receipt_cmp(const PWNC_RCPT *arg1, const PWNC_RCPT *arg2) {
	unsigned char *rhash1 = NULL;
	unsigned char *rhash2 = NULL;
	int retval = 0;
	rhash1 = rcpt_SHA1(arg1);
	rhash2 = rcpt_SHA1(arg2);
	retval = (int)(GenerateHash(rhash1, RHASH_LEN) - GenerateHash(rhash2, RHASH_LEN));
	SAFE_FREE(rhash1);
	SAFE_FREE(rhash2);
	return retval;
}


#include "rbtree.h"

//constructs a new RBTREE_RCPT object and returns it
//this is to be placed in an rbtnode->data field
RBTREE_RCPT* RBTREE_RCPT_new(PWNC_RCPT *rcpt, AL_NODE *gnode) {
	RBTREE_RCPT *retval = (RBTREE_RCPT*)malloc(sizeof(RBTREE_RCPT));
	memset(retval, 0, sizeof(RBTREE_RCPT));
	retval->rcpt = rcpt;
	retval->graph_entry = gnode;
	retval->disk_pos = -1;
	return retval;
}

//RB tree CALLBACKS
//returns the key for an RBTNODE storing receipts
void *rcpt_key_func(RBTNODE *node) {
	if (node == NULL_NODE) {
		return 0;
	}
	return (void*)((RBTREE_RCPT*)node->data)->rcpt->timestamp;
}

//sets key value to an RBTNODE storing receipts
void rcpt_set_key(RBTNODE *node, void *keyval) {
	if (node == NULL_NODE) {
		return;
	}
	((RBTREE_RCPT*)node->data)->rcpt->timestamp = (unsigned long)keyval;
}

//compares (the keys of) two RBTNODEs storing receipts
int rcpt_comp(RBTNODE *node1, RBTNODE *node2) {
	if (node1 == NULL_NODE) {
		if (node2 == NULL_NODE) {
			return 0;
		}
		else {
			return -1;
		}
	}
	else
	if (node2 == NULL_NODE) {
		return 1;
	}

	if ( ((RBTREE_RCPT*)node1->data)->rcpt->timestamp > ((RBTREE_RCPT*)node2->data)->rcpt->timestamp) {
		return 1;
	}
	if ( ((RBTREE_RCPT*)node1->data)->rcpt->timestamp < ((RBTREE_RCPT*)node2->data)->rcpt->timestamp) {
		return -1;
	}
	if ( ((RBTREE_RCPT*)node1->data)->rcpt->timestamp == ((RBTREE_RCPT*)node2->data)->rcpt->timestamp) {
		return 0;
	}
	return 0;
}

//constructs a new RBTNODE and returns it
//the new node contains the rcpt in the data field (that is an RBTREE_RCPT object containing rcpt)
RBTNODE *RBTNODE_new(PWNC_RCPT *rcpt, const void *key_func, const void *set_key_func, const void *comp_func) {
	RBTNODE *retval = (RBTNODE*)malloc(sizeof(RBTNODE));
	memset(retval, 0, sizeof(RBTNODE));
	retval->data = (void*)RBTREE_RCPT_new(rcpt, NULL);
	retval->left = NULL_NODE;
	retval->right = NULL_NODE;
	retval->parent = NULL_NODE;
	retval->key_func = key_func;
	retval->comp_func = comp_func;
	retval->set_key_func = set_key_func;
	return retval;
}


//################################################################
//Funcs that add/remove a team from the database
//They are meant to be called, e.g. when a new team is generated by the tca and
//the process using the rdb module is sent a msg by another process through an mqueue
//instructing it to add a new team
void rdb_add_team(PWNC_PUBKEY *teampk) {
	PWNC_TEAM *team = NULL;

	pthread_mutex_lock(&teams_mutex);

	//lookup team
	ALLOC_AND_CLEAN(team, sizeof(PWNC_TEAM), PWNC_TEAM*);
	team->team_pk = teampk;
	if ( (PWNC_TEAM*)lh_retrieve(teams_ht, (void*)team) ){ //retrieves team
		//if team already exits --> do nothing
		SAFE_FREE(team);
		pthread_mutex_unlock(&teams_mutex);
		return;
	}

	//fix credits
	team->credits = 0;
	team->served = 0;
	team->consumed = 0;
	lh_insert(teams_ht, (void*)team);

#ifdef MAXFLOW
	//add the team to the graph
	graph_add_team(team);
#endif

	pthread_mutex_unlock(&teams_mutex);
}

//remove team
//TODO: check it --> na elegksw ti paizei otan h omada pou einai na diagrafei exei apodeikseis
//p.x. isws prepei na diagrafoune apo to receipt graph (kai isws kai apo to disko)
void rdb_remove_team(PWNC_PUBKEY *teampk) {
}


//Looks up a team. If it exists --> return sth. (e.g. 1), else return 0
int rdb_team_exists(PWNC_PUBKEY *teampk) {
	PWNC_TEAM *team = NULL;
	int retval = 0;

	pthread_mutex_lock(&teams_mutex);

	//lookup team
	ALLOC_AND_CLEAN(team, sizeof(PWNC_TEAM), PWNC_TEAM*);
	team->team_pk = teampk;
	if ( (PWNC_TEAM*)lh_retrieve(teams_ht, (void*)team) ){ //retrieves team
		//team already exits
		retval = 1;
	}
	else {
		retval = 0;
	}

	SAFE_FREE(team);
	pthread_mutex_unlock(&teams_mutex);

	return retval;
}

//################################################################

//inserts a receipt in the hash table and in the tree
//if a receipt with the same SHA1 value exists, we keep the one with the greater timestamp.
//The pointer to the old receipt is removed from the hash table. Then, a new node is inserted in the hash table, and the old receipt
//is replaced in  the tree (calling replace_rcpt - the old receipt memory is freed)
int insert_receipt(PWNC_RCPT *rcpt, int store_it) {
	PWNC_RCPT *temp = NULL;
	RBTNODE *oldest_rcpt = NULL_NODE;
	RBTNODE *new_tree_node = NULL_NODE;

	int retval = 0;
	if (store_it) {
		//TODO: Na to elegksw ayto. An store_it == true ---> new receipt to be stored
		//An store_it == false ----> rcpt loading stage (de xreiazontai checks, ypoti8etai oti exoune hdh ginei (kerdizoume para poly xrono ap' ta verifications)
		//DONE (nomizw...)
		if ( (retval = check_rcpt(rcpt)) ) {
			//error - receipt check failed
			return retval;
		}
	}
	//if provider==consumer ignore receipt
	if ( (retval = check_equal_teams(rcpt)) ) {
		return retval;
	}

	//lock receipts mutex (that is, prevent other threads from accessing the receipts hash table and the rbtree
	pthread_mutex_lock(&rcpt_mutex);

	if ( (temp = (PWNC_RCPT*)lh_retrieve(receipts_ht, (void*)rcpt)) ) {
		//receipt hash already exists
		//check weights and keep the one with the greatest weight
		//also, adjusts team credits accordingly
		if (temp->weight < rcpt->weight) {
			lh_delete(receipts_ht, (void*)temp); //remove old receipt
			remove_rcpt_credits(temp); //remove old receipt credits
			replace_rcpt(rcpt_tree, rcpt, store_it); //replace old receipt with new one
			add_rcpt_credits(rcpt); //add the credits, etc due to the new receipt
		}
	}
	else {//receipt hash doesn't exist
		//else just append rcpt to rcpt database and ++ provider credits, etc
		if (num_of_receipts_stored >= RECEIPT_THRESHOLD) {
			//threshold reached ===> remove oldest receipt, add new
			oldest_rcpt = rbtree_find_min(rcpt_tree);
			//if rcpt older than minnode ===> do nothing else ...
			if ( (unsigned long)oldest_rcpt->key_func(oldest_rcpt) < rcpt->timestamp ) {
				//remove from hash table
#ifdef MAXFLOW
				//if there is a graph, the respective node must be removed
				graph_remove_rcpt( ((RBTREE_RCPT*)oldest_rcpt->data)->graph_entry, ((RBTREE_RCPT*)oldest_rcpt->data)->rcpt->weight );
#endif
				lh_delete(receipts_ht, ((RBTREE_RCPT*)oldest_rcpt->data)->rcpt);
				//remove from tree
				rbtree_delete(&rcpt_tree, oldest_rcpt);
#ifdef PERM_STORAGE
				if (store_it) {
					handle_receipt_storage( ((RBTREE_RCPT*)oldest_rcpt->data), RDB_ACTION_REMOVE ); //remove receipt from disk
				}
#endif
				//free memory of old receipt
				SAFE_FREE_RCPT(((RBTREE_RCPT*)oldest_rcpt->data)->rcpt);
				SAFE_FREE(oldest_rcpt);
				//decrement num of receipts
				num_of_receipts_stored--;
			}
		}
		//check if teams exist, else add them and update credits, etc
		check_rcpt_teams(rcpt);
		//now, add new receipt to hash table
		lh_insert(receipts_ht, (void*)rcpt);
		//add new receipt to tree
		new_tree_node = RBTNODE_new(rcpt, rcpt_key_func, rcpt_set_key, rcpt_comp);
		if (num_of_receipts_stored == 0) {
			//if the tree is empty (once in a lifetime...)
			rcpt_tree = new_tree_node;
		}
		else {
			rbtree_insert(&rcpt_tree, new_tree_node );
		}
#ifdef PERM_STORAGE
		if (store_it) {
			if (handle_receipt_storage( ((RBTREE_RCPT*)new_tree_node->data), RDB_ACTION_INSERT) == -1) {
				fprintf(stderr, "insert_receipt: Could not insert. handle_receipt_storage returned -1\n");
				//insert receipt to disk
				//TODO:"Error writing receipt to disk" --+++kati prepei na kanw isws...
			}
		}
#endif

#ifdef MAXFLOW
		//add receipt to the graph
		//this function is responsible for setting up the graph_entry field of new_tree_node
		graph_add_rcpt(new_tree_node);
#endif

		num_of_receipts_stored++;
	}

	pthread_mutex_unlock(&rcpt_mutex);
	return 0;
}

//traverses the tree, and replaces the tree node ('s data) with the rcpt
//If there are more than one receipts with the same timestamp, it checks the receipt's sha1 hash and replaces this one
//if there is a receipt graph, this func doesn't have anything to do with it, since no new edges are removed/inserted (only the edge data are changed)
void replace_rcpt(RBTNODE *tree, PWNC_RCPT *rcpt, int store_it) {
	unsigned char *rcpt_sha1_hash_1 = NULL;
	unsigned char *rcpt_sha1_hash_2 = NULL;
	PWNC_RCPT *rcpt_to_be_replaced = NULL;
	FILE *fp = NULL;

	if (tree == NULL_NODE)
		return;

	if (tree->key_func(tree) == (void*)rcpt->timestamp) {
		rcpt_sha1_hash_1 = rcpt_SHA1(rcpt);
		rcpt_sha1_hash_2 = rcpt_SHA1( (PWNC_RCPT*)tree->data );
		if (!memcmp(rcpt_sha1_hash_1, rcpt_sha1_hash_2, 20)) {
			rcpt_to_be_replaced = ((RBTREE_RCPT*)tree->data)->rcpt;
			SAFE_FREE_RCPT( rcpt_to_be_replaced ); //free old receipt
			((RBTREE_RCPT*)tree->data)->rcpt = rcpt; //replace it with a new one
#ifdef PERM_STORAGE
			//if permanent storage is supported, the old receipt is overwritten by the new one
			//no special checks are performed, since the only thing that changes is the receipt data (e.g. we don't need to
			//modify the list of the free positions in the receipts db).
			if (store_it) {
				fp = fopen(RECEIPTS_DB_FILE, "r+");
				if (fp) {
					//overwrite old receipt
					io_store_receipt( ((RBTREE_RCPT*)tree->data)->rcpt, fp, ((RBTREE_RCPT*)tree->data)->disk_pos*size_of_rcpt);
					fclose(fp);
				}
			}
#endif
			SAFE_FREE(rcpt_sha1_hash_1);
			SAFE_FREE(rcpt_sha1_hash_2);
			return;
		}
		//else
		//check to see if there are more receipts with equal timestamp to be checked
		if (tree->left->key_func(tree) == (void*)rcpt->timestamp )
			replace_rcpt(tree->left, rcpt, store_it);
		if (tree->right->key_func(tree) == (void*)rcpt->timestamp )
			replace_rcpt(tree->right, rcpt, store_it);
	}

	if (tree->key_func(tree) > (void*)rcpt->timestamp)
		replace_rcpt(tree->left, rcpt, store_it);
	if (tree->key_func(tree) < (void*)rcpt->timestamp)
		replace_rcpt(tree->right, rcpt, store_it);
}

//check receipt validity
//1. verify certificate
//2. verify receipt signature
//returns 0 if checks are successful or some error (otherwise)
int check_rcpt(PWNC_RCPT *rcpt) {
	unsigned char *cert_info = NULL;
	unsigned char *rinfo = NULL;
	int info_len = 0;

	//verify consumer cert
	cert_info =  serialize_cert_info(rcpt->ConsumerCert->TeamPubKey, rcpt->ConsumerCert->UserPubKey, &info_len);
	if (!verify_pwnc_signature(rcpt->ConsumerCert->TeamSig, rcpt->ConsumerCert->TeamPubKey, cert_info, info_len, 1)) {
		//check failed --> error
		SAFE_FREE(cert_info);
		return PWNC_ERR_CERT;
	}
	SAFE_FREE(cert_info);
	info_len = 0;

	// verify rcpt signature
	rinfo = serialize_receipt_info(rcpt->ConsumerCert, rcpt->ProviderPubKey, rcpt->timestamp, rcpt->weight, &info_len);
	if (!verify_pwnc_signature(rcpt->RcptSig, rcpt->ConsumerCert->UserPubKey, rinfo, info_len, 1)) {
		//check failed --> error
		SAFE_FREE(rinfo);
		return PWNC_ERR_NOT_VERIFIED;
	}
	SAFE_FREE(rinfo);

	return 0;
}

//checks whether the teams involved in a receipt already exist
//if not, they are added to the hash table storing teams
//we suppose that receipt verification has already taken place
void check_rcpt_teams(PWNC_RCPT *rcpt) {
//	printf("CHECK 1: %d - %d\n", rcpt->ConsumerCert->TeamPubKey->datalen, rcpt->ProviderPubKey->datalen);
	PWNC_TEAM *team, *provider_team, *consumer_team;

	pthread_mutex_lock(&teams_mutex);

	//team == provider
	ALLOC_AND_CLEAN(team, sizeof(PWNC_TEAM), PWNC_TEAM*);
	team->team_pk = rcpt->ProviderPubKey;
	provider_team = (PWNC_TEAM*)lh_retrieve(teams_ht, (void*)team); //retrieves provider team
	SAFE_FREE(team);

	ALLOC_AND_CLEAN(team, sizeof(PWNC_TEAM), PWNC_TEAM*);
	team->team_pk = rcpt->ConsumerCert->TeamPubKey;
	consumer_team = (PWNC_TEAM*)lh_retrieve(teams_ht, (void*)team); //retrieves provider team
	SAFE_FREE(team);

#ifdef FIXED_IDS
	//if FIXED_IDS is set, then we do not add receipts involving teams that do not exist
	if (!provider_team || !consumer_team) {
		pthread_mutex_unlock(&teams_mutex);
		return;
	}
#endif

	if (!provider_team) {
		//create new provider team and add it to the hash table storing teams
		ALLOC_AND_CLEAN(provider_team, sizeof(PWNC_TEAM), PWNC_TEAM*);
		provider_team->team_pk = PWNC_KEY_new();
		provider_team->team_pk->algorithm = rcpt->ProviderPubKey->algorithm;
		provider_team->team_pk->bits = rcpt->ProviderPubKey->bits;
		ALLOC_AND_CLEAN(provider_team->team_pk->data, rcpt->ProviderPubKey->datalen, unsigned char*);
		memcpy(provider_team->team_pk->data, rcpt->ProviderPubKey->data, rcpt->ProviderPubKey->datalen);
		provider_team->team_pk->datalen = rcpt->ProviderPubKey->datalen;

		provider_team->credits = 0;
		provider_team->served = 0;
		provider_team->consumed = 0;

		if (lh_insert(teams_ht, (void*)provider_team)) {
			fprintf(stderr, "REPLACED provider team!\n");
		}

#ifdef MAXFLOW
		//the teams must be added to the graph
		graph_add_team(provider_team);
#endif
	}
	if (!consumer_team) {
		//create new consumer team - add it to hash table
		ALLOC_AND_CLEAN(consumer_team, sizeof(PWNC_TEAM), PWNC_TEAM*);
		consumer_team->team_pk = PWNC_KEY_new();
		consumer_team->team_pk->algorithm = rcpt->ConsumerCert->TeamPubKey->algorithm;
		consumer_team->team_pk->bits = rcpt->ConsumerCert->TeamPubKey->bits;
		ALLOC_AND_CLEAN(consumer_team->team_pk->data, rcpt->ConsumerCert->TeamPubKey->datalen, unsigned char*);
		memcpy(consumer_team->team_pk->data, rcpt->ConsumerCert->TeamPubKey->data, rcpt->ConsumerCert->TeamPubKey->datalen);
		consumer_team->team_pk->datalen = rcpt->ConsumerCert->TeamPubKey->datalen;

		consumer_team->credits = 0;
		consumer_team->served = 0;
		consumer_team->consumed = 0;

		if (lh_insert(teams_ht, (void*)consumer_team)) {
			fprintf(stderr, "REPLACED consumer team!\n");
		}
#ifdef MAXFLOW
		graph_add_team(consumer_team);
#endif

	}

	provider_team->credits += ((double)rcpt->weight)/1000.0;
	provider_team->served++;
	consumer_team->consumed++;

	pthread_mutex_unlock(&teams_mutex);
}

//check if provider==consumer
//in such a case the receipt is not stored
int check_equal_teams(PWNC_RCPT *rcpt) {
	if (!rcpt) {
		return 0;
	}

	if (rcpt->ConsumerCert->TeamPubKey->datalen != rcpt->ProviderPubKey->datalen) {
		return 0;
	}

	if (memcmp(rcpt->ConsumerCert->TeamPubKey->data, rcpt->ProviderPubKey->data, rcpt->ProviderPubKey->datalen)) {
		return 0;
	}

	return 1;
}

//removes the credits due to this receipt
//in this func we lock the teams mutex
void remove_rcpt_credits(PWNC_RCPT *rcpt) {
	PWNC_TEAM *team, *provider_team, *consumer_team;

	pthread_mutex_lock(&teams_mutex);

	//team == provider
	ALLOC_AND_CLEAN(team, sizeof(PWNC_TEAM), PWNC_TEAM*);
	team->team_pk = rcpt->ProviderPubKey;
	provider_team = (PWNC_TEAM*)lh_retrieve(teams_ht, (void*)team); //retrieves provider team
	SAFE_FREE(team);

	ALLOC_AND_CLEAN(team, sizeof(PWNC_TEAM), PWNC_TEAM*);
	team->team_pk = rcpt->ConsumerCert->TeamPubKey;
	consumer_team = (PWNC_TEAM*)lh_retrieve(teams_ht, (void*)team); //retrieves provider team
	SAFE_FREE(team);

	if (provider_team && consumer_team) {
		provider_team->credits -= ((double)rcpt->weight)/1000.0;
		provider_team->served--;
		consumer_team->consumed--;
	}
	pthread_mutex_unlock(&teams_mutex);
}

//adds the credits the provider deserves from this rcpt
void add_rcpt_credits(PWNC_RCPT *rcpt) {
	PWNC_TEAM *team, *provider_team, *consumer_team;

	pthread_mutex_lock(&teams_mutex);

	//team == provider
	ALLOC_AND_CLEAN(team, sizeof(PWNC_TEAM), PWNC_TEAM*);
	team->team_pk = rcpt->ProviderPubKey;
	provider_team = (PWNC_TEAM*)lh_retrieve(teams_ht, (void*)team); //retrieves provider team
	SAFE_FREE(team);

	ALLOC_AND_CLEAN(team, sizeof(PWNC_TEAM), PWNC_TEAM*);
	team->team_pk = rcpt->ConsumerCert->TeamPubKey;
	consumer_team = (PWNC_TEAM*)lh_retrieve(teams_ht, (void*)team); //retrieves provider team
	SAFE_FREE(team);

	if (provider_team && consumer_team) {
		//maybe this check is not necessary since, if a receipt exists, then the teams it involves will have been already added
		provider_team->credits += ((double)rcpt->weight)/1000.0;
		provider_team->served++;
		consumer_team->consumed++;
	}

	pthread_mutex_unlock(&teams_mutex);
}

void rdb_init(const char *receipts_db_file, const char *receipts_db_free_list) {
	//Hash tables' initialization
	//Teams hash table
	teams_ht = lh_new(LHASH_HASH_FN(team_hash), LHASH_COMP_FN(team_cmp));

	tree_init();

	//Receipts hash table
	receipts_ht = lh_new(LHASH_HASH_FN(receipt_hash), LHASH_COMP_FN(receipt_cmp));
	//receipts_ht = lh_new(NULL, NULL);

	pthread_mutex_init(&teams_mutex, NULL);
	pthread_mutex_init(&rcpt_mutex, NULL);

#ifdef MAXFLOW
	receipt_graph = NULL;
#endif

#ifdef PERM_STORAGE
	memset(RECEIPTS_DB_FILE, 0, 256);
	strncpy(RECEIPTS_DB_FILE, receipts_db_file, 255);
	memset(RECEIPTS_DB_FREE_LIST, 0, 256);
	strncpy(RECEIPTS_DB_FREE_LIST, receipts_db_free_list, 255);
#endif
}


//Decides whether the consumer should be granted access or not
//Algorithm:
//Give access with probability = consumer.credits/provider.credits
//If (consumer.credits > provider.credits => always gives access
int rdb_judge(PWNC_PUBKEY *consumer, PWNC_PUBKEY *provider) {
//	printf("JUDGE: %d - %d\n", consumer->datalen, provider->datalen);
//	printf("About to judge...\n");
	PWNC_TEAM *provider_team, *consumer_team, *testteam;

	pthread_mutex_lock(&teams_mutex);

	//retrieve provider
	ALLOC_AND_CLEAN(testteam, sizeof(PWNC_TEAM), PWNC_TEAM*);
	testteam->team_pk = provider;
	if (! (provider_team = (PWNC_TEAM*)lh_retrieve(teams_ht, (void*)testteam)) ) {
		fprintf(stderr, "Could not retrieve provider\n");
	}
	SAFE_FREE(testteam);

	//retrieve consumer
	ALLOC_AND_CLEAN(testteam, sizeof(PWNC_TEAM), PWNC_TEAM*);
	testteam->team_pk = consumer;
	if (! (consumer_team = (PWNC_TEAM*)lh_retrieve(teams_ht, (void*)testteam)) ) {
		fprintf(stderr, "Could not retrieve consumer\n");
	}
	SAFE_FREE(testteam);

	//Compare provider - consumer credits
	//if provider or consumer unknown ===> flip coin
	//else
	//if rand() > consumer/provider =>give access ( 0 < rand < 1)
	double testratio = 0.0;
	if (!provider_team || !consumer_team) {
		pthread_mutex_unlock(&teams_mutex);
		fprintf(stderr, "UNKNOWN provider or consumer...Random choice\n");
		if (rand()/(RAND_MAX + 1.0) > 0.5) {
			return 1;
		}
		else {
			return 0;
		}
	}

	//if provider == consumer ==> always give access (same team)
	//Normally, this check is done by the server. Anyway....
	if ( (provider_team->team_pk->datalen == consumer_team->team_pk->datalen) &&
		!memcmp(provider_team->team_pk->data, consumer_team->team_pk->data, consumer_team->team_pk->datalen) ) {

		pthread_mutex_unlock(&teams_mutex);
		return 1;

	}

#ifdef MAXFLOW
	long prov_to_cons = maxflow_run(provider_team->graph_node, consumer_team->graph_node);
	long cons_to_prov = maxflow_run(consumer_team->graph_node, provider_team->graph_node);

	if ( (cons_to_prov != (long)0) && ( (testratio = ((double)prov_to_cons) / ((double)cons_to_prov) ) < rand()/(RAND_MAX + 1.0)) ) {
		//give access with probability = maxflow(prov-->cons) / maxflow(cons-->prov)
		pthread_mutex_unlock(&teams_mutex);
		return 0;
	}
	else {
		//access granted - either the consumer doesn't "owe" anything to the provider or the ratio is >= rand-between-0-and-1
		pthread_mutex_unlock(&teams_mutex);
		return 1;
	}

#else /* no maxflow-based decision function support --> use scores*/

	if ( (provider_team->credits != 0.0) && ( (testratio = consumer_team->credits / provider_team->credits) < rand()/(RAND_MAX + 1.0)) ) {
		//provider > consumer --> no access
		pthread_mutex_unlock(&teams_mutex);
		return 0;
	}
	else {
		//grant access
		pthread_mutex_unlock(&teams_mutex);
		return 1;
	}
#endif /* MAXFLOW */

	pthread_mutex_unlock(&teams_mutex);
	return 0;
}
//##################################################################

void rdb_range_query( void (*do_stuff)(RBTNODE*, void *), void *do_stuff_args, unsigned long timestamp ) {
	//lock receipts mutex
	pthread_mutex_lock(&rcpt_mutex);

	//construct a test node, that is a dumb node whose key is set to timestamp (will be used for comparison)
	PWNC_RCPT *testrcpt = (PWNC_RCPT*)malloc(sizeof(PWNC_RCPT));
	RBTREE_RCPT *testdata = RBTREE_RCPT_new(testrcpt, NULL);
	RBTNODE testnode  = {testdata, NULL_NODE, NULL_NODE, NULL_NODE, NODE_COLOR_BLACK, (void*)&rcpt_key_func, (void*)&rcpt_comp, (void*)&rcpt_set_key};
	testnode.set_key_func(&testnode, (void*)timestamp);

	//perform a range query on the tree
	rbtree_range_query(rcpt_tree, do_stuff, do_stuff_args, &testnode);

	//now that we're done, unlock mutex
	pthread_mutex_unlock(&rcpt_mutex);

	//finally, free memory
	SAFE_FREE(testrcpt);
	SAFE_FREE(testdata);
}


#ifdef PERM_STORAGE

//handles addition/removal/replacement/etc of receipts on disk
//When a receipt is inserted, it is added in the rbtree and it has to be added in a file on disk
//Once a receipt has been inserted in the file, its position is never changed
//There is an index of the receipt on disk inside the RBTREE_RCPT struct (disk_pos)
//When a receipt is removed, its disk_pos is added in the list of free positions, so than when a new receipt is to be added, it is
//written on a free position (data on disk are not removed when a node is deleted, they're just overwritten by the new node).
//to_be_handler: node that is to be inserted, removed, etc
//action: RDB_ACTION_INSERT, etc... (see rdb_engine.h)
int handle_receipt_storage(RBTREE_RCPT *to_be_handled, int action) {
	FILE *fp = NULL;
	int pos = 0; //the free position as returned by pop_free_pos()
	int final_pos = 0; //the position the rcpt will finally be written (it may be different from pos in case pop_free_pos returns FL_EMPTY_FILE)

	fp = fopen(RECEIPTS_DB_FILE, "r+");
	if (!fp) {
		fprintf(stderr, "Error opening rdb file\n");
		return -1;
	}

	switch (action) {
		case RDB_ACTION_INSERT:
			pos = pop_free_pos(RECEIPTS_DB_FREE_LIST);
			if (pos == (long)FL_ERROR) {
				fprintf(stderr, "Error getting free pos\n");
				return -1;
			}

			if (pos == (long)FL_EMPTY_FILE) {
				if (num_of_receipts_stored != RECEIPT_THRESHOLD) {
					//if file is empty and threshold isn't reached, we can just append
					//the receipt in the receipts db.
					//In this case, we set disk_pos to num_of_receipts_stored
					final_pos = num_of_receipts_stored;
				}
				else {
					//cannot store receipt because db is full
					return -1;
				}
			}
			else {
				//else: the new receipt will overwrite a deleted one (will be written on pos)
				final_pos = 0;
			}

			//insert receipt in a free pos in the file
			if (io_store_receipt( to_be_handled->rcpt, fp, final_pos*size_of_rcpt) == -1) {
				//store receipt failed, so we can push the free pos back into the file and return
				if (pos != (long)FL_EMPTY_FILE) {
					push_free_pos(RECEIPTS_DB_FREE_LIST, pos, RECEIPT_THRESHOLD);
				}
				return -1;
			}
			//set node's disk_pos
			to_be_handled->disk_pos = final_pos;
		break;

		case RDB_ACTION_REMOVE:
			//in case of a deletion, we just add the pos that the receipt occupied in the list of free receipts
			push_free_pos( RECEIPTS_DB_FREE_LIST, to_be_handled->disk_pos, RECEIPT_THRESHOLD );
		break;

		default:
			return -1;
		break;
	} //endsw
	return 0;
}

//reads the receipts from the receipt db and builds a red black tree and hash table
//This func is supposed to be called when db is initialized (init_db)
//First, the free positions are read from the free list file (using get_free_list) (TODO: load them in a hash table for O(1) access - maybe not necessary, because free list will contain very
//few items) so that the data contained in these positions in the db file
//will not be loaded
void rdb_load_data() {
	long *free_positions = NULL; //list of free positions
	int flist_size = 0; //number of free positions as read by the file
	FILE *fp = NULL;
	long offset = 0; //receipt position in the file (measured in bytes)
	int pos = 0; //receipt position in the file (measured in records)
	PWNC_RCPT *rcpt = NULL;
	int fd = 0;

	free_positions = get_free_list(RECEIPTS_DB_FREE_LIST, &flist_size);
	if ( (free_positions == NULL) && (flist_size == -1) ){
		//error retrieving free list
		fprintf(stderr, "Error retrieving list of free positions\n");
		return;
	}

	//opens receipts file
	fd = open(RECEIPTS_DB_FILE, O_RDWR);
	if (fd == -1) {
		if (errno == ENOENT) {
			fprintf(stderr, "Receipts file does not exist. Creating...\n");
			if ( (fd = creat(RECEIPTS_DB_FILE, (mode_t)0755)) == -1) {
				fprintf(stderr, "Error creating receipts file\n");
				SAFE_FREE(free_positions);
				return;
			}
		}
		else {
			fprintf(stderr, "Error opening rdb file\n");
			SAFE_FREE(free_positions);
			return;
		}
	}

	fp = fdopen(fd, "r");
	if (!fp) {
		fprintf(stderr, "Error opening rdb file\n");
		SAFE_FREE(free_positions);
		close(fd);
		return;
	}


	//if flist is NULL, do not check for free positions, just load the receipt read
	while (!feof(fp)) {
		//if the position is free, we skip it
		if (free_positions && is_free_pos(free_positions, flist_size, pos)) {
			offset += size_of_rcpt;
			pos++;
			continue;
		}
		else {
			//receipt is valid
			//read it
			rcpt = read_rcpt_file(fp, offset);
			//insert it into tree and hash table and adjust teams
			if (rcpt) {
				insert_receipt(rcpt, 0); //store_it = 0 ---> we do not want to store the receipt (we are in the loading process);
			}
			offset += size_of_rcpt;
			pos++;
		}
	}
	fclose(fp);

#ifdef MAXFLOW
	//after data has been loaded, label graph nodes
	AL_HEAD *cur_head = receipt_graph;
	int label = 0;
	while (cur_head != NULL) {
		//fix labels
		cur_head->label = label;
		label++;
		cur_head = cur_head->next_head;
	}
#endif

}

#endif /* PERM_STORAGE */



//##################################################################################

//GRAPH functions

//neighbor hash table callbacks
unsigned long neighbor_hash(const AL_NODE *to_hash) {
	return (unsigned long)to_hash->myhead->label;
//	return GenerateHash( (unsigned char*)&to_hash->myhead->label, 4);
//	return (unsigned long)(to_hash->myhead->team->team_pk->data);
/*	unsigned char *retval = (unsigned char*)malloc(4);
	memset(retval, 0, 4);
	memcpy(to_hash->myhead->team->team_pk->data, 4);
	return (unsigned long)(*retval;
*/}

int neighbor_cmp(const AL_NODE *arg1, const AL_NODE *arg2) {
	return (arg1->myhead->label - arg2->myhead->label);
//	return (GenerateHash( (unsigned char*)&arg1->myhead->label, 4) - GenerateHash( (unsigned char*)&arg2->myhead->label, 4) );
	//return ( (unsigned long)(arg1->myhead->team->team_pk->data) - (unsigned long)(arg2->myhead->team->team_pk->data) );
}

//disconnect a node from the adjacency list it belongs
void adj_list_remove(AL_NODE *to_be_removed) {
	//extreme cases:
	//1. cur is the sole node in list ===> Finally, neighbors==last_neighbor == NULL
	//2. cur is the 1st node in list (cur==neighbors, prev==NULL, next!=NULL)  ===> neighbors==cur->next
	//3. cur is the last node in list (cur==last_neighbor, prev != NULL, next == NULL) ===> last_neighbor == cur->prev
	//4. cur is an itermediate node in list ===> cur->prev->next = cur->next; cur->next->prev = cur->prev;
	//*new*
	if (to_be_removed->prev) {
		to_be_removed->prev->next = to_be_removed->next;
	}
	else {
		//this means that the node to be removed is the first in the adjacency list (==to_be_removed->consumer->neighbors)
		to_be_removed->consumer->neighbors = to_be_removed->next;
	}

	if (to_be_removed->next) {
		to_be_removed->next->prev = to_be_removed->prev;
	}
	else {
		//this means that to_be_removed == last_neighbor
		to_be_removed->consumer->last_neighbor = to_be_removed->prev;
	}

	//remove node from the neighbors ht
#ifdef _USE_HT_
	lh_delete(to_be_removed->consumer->neighbors_ht, (void*)to_be_removed);
#else
#ifndef _SLOW_
	to_be_removed->consumer->neighbors_m[to_be_removed->myhead->label].removed = 1;
	//to_be_removed->consumer->neighbors_m[to_be_removed->myhead->label].node = NULL;
#endif
#endif

}

#ifdef _GLOBAL_RELABEL_
//Adds an entry in the "incoming" list of node->myhead
void add_incoming(AL_NODE *node) {
	INCOMING *new_node = (INCOMING*)malloc(sizeof(INCOMING));
	memset(new_node, 0, sizeof(INCOMING));
	new_node->node = node;
	node->incoming_ptr = new_node;

	//add new node on top of the list (O(1))
	INCOMING *cur = node->myhead->incoming; //traverse the list of v's neighbor until we reach the end
	if (cur == NULL) {
		//the new node will be the 1st in the list, so v->neighbors == v->last_neighbor == new_node
		node->myhead->incoming = new_node;
		node->myhead->last_incoming = new_node;
		new_node->next = NULL;
		new_node->prev = NULL;
	}
	else {
		new_node->next = node->myhead->incoming;
		new_node->prev = NULL;
		node->myhead->incoming->prev = new_node;
		node->myhead->incoming = new_node;
	}
}

//Removes the respective "incoming" entry from the incoming list of node->myhead
void remove_incoming(AL_NODE *to_be_removed) {
	INCOMING *to_be_freed = to_be_removed->incoming_ptr;
	if (to_be_freed->prev) {
		to_be_freed->prev->next = to_be_freed->next;
	}
	else {
		to_be_removed->myhead->incoming = to_be_freed->next;
	}

	if (to_be_freed->next) {
		to_be_freed->next->prev = to_be_freed->prev;
	}
	else {
		//to_be_removed == last_incoming
		to_be_removed->myhead->last_incoming = to_be_freed->prev;
	}
}

//Runs a backward bfs from the sink to all the nodes using the "incoming" lists.
//Updates the distances of all nodes from the sink.
//This func is O(V + E) so it is run periodically (e.g every V relabel operations)
void global_relabel(AL_HEAD *sink) {
	AL_HEAD *cur = NULL;
	FIFO_NODE *qhead = NULL;
	FIFO_NODE *qtail = NULL;
	INCOMING *inc = NULL;

	//initialize nodes
	cur = receipt_graph;
	while (cur) {
		if (cur != sink) {
			cur->node_state = MF_NODE_COLOR_WHITE;
			//cur->height = INFINITY;
			cur->height = num_of_teams_stored;
		}
		cur = cur->next_head;
	}

	sink->node_state = MF_NODE_COLOR_GRAY;
	sink->height = 0;
	//Enqueue source (source of the BFS <==> sink of maxflow
	FIFO_NODE *fnode = (FIFO_NODE*)malloc(sizeof(FIFO_NODE));
	memset(fnode, 0, sizeof(FIFO_NODE));
	fnode->data = (void*)sink;
	fifo_push(fnode, &qhead, &qtail);

	AL_HEAD *u = NULL;
	while ( qhead || qtail ) {
		fnode = fifo_pop(&qhead, &qtail);
		u = (AL_HEAD*)fnode->data;
		SAFE_FREE(fnode);
		inc = u->incoming;
		while (inc && (inc->node->saturated==0) ) {
			if (INCOMING_HEAD(inc)->node_state == MF_NODE_COLOR_WHITE) {
				INCOMING_HEAD(inc)->node_state = MF_NODE_COLOR_GRAY;
				INCOMING_HEAD(inc)->height = u->height + 1;

				fnode = (FIFO_NODE*)malloc(sizeof(FIFO_NODE));
				memset(fnode, 0, sizeof(FIFO_NODE));
				fnode->data = (void*)INCOMING_HEAD(inc);
				fifo_push(fnode, &qhead, &qtail);
			}
			inc = inc->next;
		}
		u->node_state = MF_NODE_COLOR_BLACK;
	}
}
#endif /* _GLOBAL_RELABEL_ */

//adds a new team to the graph
//Constructs a new AL_HEAD and appends it to the end of the list of heads
void graph_add_team(PWNC_TEAM *team) {
	AL_HEAD *cur = receipt_graph;
	//alloc new node
	AL_HEAD *new_node = (AL_HEAD*)malloc(sizeof(AL_HEAD));
	memset(new_node, 0, sizeof(AL_HEAD));
	new_node->color = 0;
	new_node->height = 0;
	new_node->excess = 0;
	new_node->label = 0;
//	new_node->discharged = 0;
	new_node->neighbors = NULL;
	new_node->last_neighbor = NULL;//*new*
	new_node->next_head = NULL;
	new_node->team = team;
	team->graph_node = new_node;

	num_of_teams_stored++;

#ifdef _USE_HT_
	new_node->neighbors_ht = lh_new(LHASH_HASH_FN(neighbor_hash), LHASH_COMP_FN(neighbor_cmp));
#else
#ifndef _SLOW_
	//we keep track of current number of teams
	//so, having appended the new team, we, go through all treams and reallon each head node's "neighbors_m" array
	//adding a new position. This will be the pos of the new team. Also, alloc space for the new head node's neighbors_m
	new_node->neighbors_m = (NMATRIX_ENTRY*)malloc(num_of_teams_stored * sizeof(NMATRIX_ENTRY) );
	memset(new_node->neighbors_m, 0, num_of_teams_stored * sizeof(NMATRIX_ENTRY));
#endif
#endif

#ifdef _GLOBAL_RELABEL_
	new_node->incoming = NULL;
	new_node->last_incoming = NULL;
#endif

	//move to the end of the list and append the new node
	if (cur == NULL) {
		receipt_graph = new_node;
	}
	else {
		while(cur->next_head != NULL) {
#ifndef _USE_HT_
#ifndef _SLOW_
			//realloc neighbors_m
			cur->neighbors_m = (NMATRIX_ENTRY*)realloc(cur->neighbors_m, num_of_teams_stored * sizeof(NMATRIX_ENTRY) );
			memset(&cur->neighbors_m[num_of_teams_stored-1], 0, sizeof(NMATRIX_ENTRY));
#endif
#endif
			cur = cur->next_head;
		}
		cur->next_head = new_node;
	}
}

//adds a receipt to the graph
//If there already exists an edge A->B, just ++the edge's weight
//If not, append new node to the end of the neighbor list
void graph_add_rcpt(RBTNODE *tree_node) {
	PWNC_TEAM *consumer_team = NULL;
	PWNC_TEAM *provider_team = NULL;

	PWNC_TEAM *temp_team = (PWNC_TEAM*)malloc(sizeof(PWNC_TEAM));
	memset(temp_team, 0, sizeof(PWNC_TEAM));

	temp_team->team_pk = ((RBTREE_RCPT*)tree_node->data)->rcpt->ConsumerCert->TeamPubKey;
	consumer_team = (PWNC_TEAM*)lh_retrieve(teams_ht, (void*)temp_team);
	temp_team->team_pk = ((RBTREE_RCPT*)tree_node->data)->rcpt->ProviderPubKey;
	provider_team = (PWNC_TEAM*)lh_retrieve(teams_ht, (void*)temp_team);

	if (!consumer_team || !provider_team) {
		fprintf(stderr, "graph_add_rcpt: Error retrieving teams from hash table\n");
		SAFE_FREE(temp_team);
		return;
	}
	SAFE_FREE(temp_team);

	//now that we've retrieved the team (and consequently the head node) we scan the neighbor list and
	//check if the edge exists. Else --> append the new receipt to the neighbor list

	//construct new adjacency list node (<==> receipt)
	AL_NODE *new_node = NULL;
	//now search the consumer's list and append it
	AL_NODE *cur = consumer_team->graph_node->neighbors;
	if (cur != NULL) {
		while(cur != NULL) {
			if (cur->myhead == provider_team->graph_node) {
				//found node-->update edge weight
				cur->weight += ((RBTREE_RCPT*)tree_node->data)->rcpt->weight;
				cur->residual_weight = cur->weight;
				return; //nothing left to do --> exit
			}//else advance pointer
			cur = cur->next;
		}

		//Edge not found--->append the new node and update head->last_neighbor pointer
		new_node = (AL_NODE*)malloc(sizeof(AL_NODE));
		memset(new_node, 0, sizeof(AL_NODE));

		new_node->next = NULL; //set next to NULL (since the new node will be appended)
		new_node->prev = consumer_team->graph_node->last_neighbor; //*new*
		consumer_team->graph_node->last_neighbor->next = new_node;//*new*
		consumer_team->graph_node->last_neighbor = new_node;//*new*

		new_node->myhead = provider_team->graph_node;
		new_node->consumer = consumer_team->graph_node;
		new_node->weight = ((RBTREE_RCPT*)tree_node->data)->rcpt->weight;
		new_node->residual_weight = new_node->weight;
	}
	else {
		//no neighbors..
		new_node = (AL_NODE*)malloc(sizeof(AL_NODE));
		memset(new_node, 0, sizeof(AL_NODE));
		new_node->next = NULL; //set next to NULL (since the new node will be appended)
		new_node->prev = NULL; //*new* - prev == NULL since its the sole node in the list
		new_node->myhead = provider_team->graph_node;
		new_node->consumer = consumer_team->graph_node;
		new_node->weight = ((RBTREE_RCPT*)tree_node->data)->rcpt->weight;
		new_node->residual_weight = new_node->weight;

		consumer_team->graph_node->neighbors = new_node;
		consumer_team->graph_node->last_neighbor = new_node; //*new*
	}

#ifdef _GLOBAL_RELABEL_
	//the new receipt must be also added in the "incoming" list of the provider - this only applies when the edge doesn't already exist
	//in the residual graph
	add_incoming(new_node);
#endif
}

//"remove" a node from the receipts graph
//Find the list where this node belongs (rcpt_node->consumer)
//Traverse the neighbor list until yoo find the node to be deleted
//Reduces the edge's weight by rweight (<-->receipt weight) and if weight becomes 0 removes the edge
//Returns the node to be deleted (its mem will be freed by the calling func??)
AL_NODE *graph_remove_rcpt(AL_NODE *rcpt_node, unsigned long rweight) {
	AL_NODE *cur = rcpt_node->consumer->neighbors;

	while ( (cur != rcpt_node) && (cur != NULL) ) {
		cur = cur->next;
	}

	if (cur == NULL) {
		//some err occured - rcpt_node not found!
		fprintf(stderr, "graph_remove_rcpt: Node not found\n");
		return NULL;
	}

	cur->weight -= rweight;
	cur->residual_weight = cur->weight;
	if (cur->weight <= 0) {
		//the edge must be removed
		adj_list_remove(cur);

#ifdef _GLOBAL_RELABEL_
		remove_incoming(cur);
		SAFE_FREE(cur->incoming_ptr);
#endif

	}
	return cur;
}

//TODO: Delete TEAM

//adds a "foo" edge to the graph (v->u)
//This edge is directed from the consumer to the provider. It is appended to the "consumer"'s (v)
//neighbor list, so its "consumer" is v and its "myhead" is u->myhead
//Returns the newly-inserted edge
//TODO: ayth einai proxeirh lysh, prepei na allaksw tis parametrous, anti gia u p.x. na vazw to
//u->myhead 'h kai tipota (isws...)
AL_NODE *add_foo_edge(AL_HEAD *v, AL_HEAD *u, unsigned long edge_weight) {
	AL_NODE *new_node = (AL_NODE*)malloc(sizeof(AL_NODE));
	memset(new_node, 0, sizeof(AL_NODE));
	new_node->myhead = u;
	new_node->consumer = v;
	new_node->saturated = 0; //it's not yet saturated
	new_node->weight = edge_weight; //weight of the edge
	new_node->residual_weight = edge_weight;
	new_node->is_foo = 1; //foo edge
	//append to list
	AL_NODE *cur = v->neighbors; //traverse the list of v's neighbor until we reach the end
	if (cur == NULL) {
		//the new node will be the 1st in the list, so v->neighbors == v->last_neighbor == new_node
		v->neighbors = new_node;
		v->last_neighbor = new_node;
		new_node->next = NULL;
		new_node->prev = NULL;//*new*
	}
	else {
		//instead of appending to the end of  list (O(V) -see right above) , put it on top of list (O(1))
		new_node->next = v->neighbors;
		new_node->prev = NULL; //*new*
		v->neighbors->prev = new_node; //*new*
		v->neighbors = new_node;
	}

#ifdef _USE_HT_
	//insert new foo edge to head node's neighbors hash table
	lh_insert(v->neighbors_ht, (void*)new_node); //**new**
#else
#ifndef _SLOW_
	v->neighbors_m[new_node->myhead->label].removed = 0;
	v->neighbors_m[new_node->myhead->label].node = new_node;
#endif
#endif

#ifdef _GLOBAL_RELABEL_
	add_incoming(new_node);
#endif
	return new_node;
}

//returns the neighboring node  if v is in the neighbor list of u, else NULL
//Saturated edges are placed at the end of the list, so as soon as a saturated edge is found --> exit loop
AL_NODE *is_neighbor(AL_HEAD *u, AL_HEAD *v) {
#ifdef _USE_HT_
	AL_NODE *retval = NULL;
	//construct a foo node with ->myhead == v and pass it as a param to lh_retrieve
	AL_NODE *temp = (AL_NODE*)malloc(sizeof(AL_NODE));
	memset(temp, 0, sizeof(AL_NODE));
	temp->myhead = v;
	retval = (AL_NODE*)lh_retrieve(u->neighbors_ht, (void*)temp);
	//free temporary AL_NODE pointer
	SAFE_FREE(temp);
	return retval;
#else
#ifdef _SLOW_
	AL_NODE *cur = u->neighbors;
	while ( (cur != NULL) && (!cur->saturated) ) {
		if ( cur->myhead == v ){
			return cur; //found a neighbor==v
		}
		cur = cur->next;
	}
	return NULL;
#else
	if (u->neighbors_m[v->label].removed) {
		return NULL;
	}
	else {
		return u->neighbors_m[v->label].node;
	}
#endif
#endif
}

void maxflow_push(AL_HEAD *u, AL_NODE *v) {
	struct timeval start, end;
	memset(&start, 0, sizeof(struct timeval));
	memset(&end, 0, sizeof(struct timeval));
	gettimeofday(&start, NULL);
	int satted = 0;

	long df = 0;
	df = ( (u->excess <= v->residual_weight) ? u->excess:v->residual_weight);

	v->residual_weight -= df;
	if (v->residual_weight == 0) {
		v->saturated = 1;
		//disconnect v and re-insert it at the end of the list
		adj_list_remove(v); //O(1)
#ifdef _GLOBAL_RELABEL_
		remove_incoming(v);
		if (!v->myhead->last_incoming) { //empty list
			v->myhead->last_incoming = v->incoming_ptr;
			v->myhead->incoming = v->incoming_ptr;
			v->myhead->incoming->prev = NULL;
			v->myhead->incoming->next = NULL;
		}
		else {
			v->myhead->last_incoming->next = v->incoming_ptr;
			v->incoming_ptr->prev = v->myhead->last_incoming;
			v->incoming_ptr->next = NULL;
			v->myhead->last_incoming = v->incoming_ptr;
		}

#endif
		//append v at the end of list - O(1)
		if (!v->consumer->last_neighbor) { //empty list
			v->consumer->last_neighbor = v;
			v->consumer->neighbors = v;
			v->prev = NULL;
			v->next = NULL;
		}
		else {
			v->consumer->last_neighbor->next = v;
			v->prev = v->consumer->last_neighbor;
			v->next = NULL;
			v->consumer->last_neighbor = v;
		}

		sat_push_ops++;satted = 1;
	}
	else {
		non_sat_push_ops++; satted = 0;
	}

	AL_NODE *neigh = NULL;
	if (!(neigh = is_neighbor(v->myhead, u)) ) {
		add_foo_edge(v->myhead, u, df); //adds a temp edge on the residual graph (add_incoming is called by add_foo_edge)
	}
	else {
		//v->weight += df;
//TODO: 10/11: Edw pera isws prepei na kanw to discounting
//na mhn prow8w df alla df*(out_degree(source)/out_degree(neigh))
//Dhladh, to varos mias akmhs *den* einai weight alla weight*(out_degree(source)/out_degree(neigh))
		neigh->residual_weight += df;
	}
	u->excess -= df;
	v->myhead->excess += df;
}

void maxflow_relabel(AL_HEAD *u, int min_height) {
	u->height = min_height + 1;
	relabel_ops++;
#ifdef _GLOBAL_RELABEL_
	global_relabel_threshold++;
#endif
}

void maxflow_discharge(AL_HEAD *u, AL_HEAD *sink, FIFO_NODE **q_head, FIFO_NODE **q_tail) {
	AL_NODE *cur = NULL;
	int min_height = INFINITY;
	FIFO_NODE *fnode = NULL;

	while ( (u->excess > 0) && (!u->neighbors->saturated) ) {
		cur = u->neighbors;
		min_height = INFINITY;
		while ( cur && (!cur->saturated) && (u->excess > 0) ) {
			//if (u->height > cur->myhead->height) {
			if (u->height == cur->myhead->height + 1) {
				maxflow_push(u, cur);

				//enqueue the nodes that have become overflowing back to Q
				if ( (cur->myhead->color != MF_NODE_COLOR_BLACK) && (cur->myhead != sink) ) {
					cur->myhead->color = MF_NODE_COLOR_BLACK;
					fnode = (FIFO_NODE*)malloc(sizeof(FIFO_NODE));
					memset(fnode, 0, sizeof(FIFO_NODE));
					fnode->data = (void*)cur->myhead;
					//push v back to Q
					fifo_push(fnode, q_head, q_tail);
				}

			}

			if (cur->myhead->height < min_height) {
				min_height = cur->myhead->height;
			}
			cur = cur->next;
		}

		if ( (!cur || cur->saturated) && (u->excess > 0) ) {
			//that is, we have scanned all nodes ("end_of_list")
			cur = u->neighbors;
			min_height = INFINITY;
			while (cur && (!cur->saturated) ) {
				if (min_height > cur->myhead->height) {
					min_height = cur->myhead->height;
				}
				cur = cur->next;
			}
			maxflow_relabel(u, min_height);
#ifdef _GLOBAL_RELABEL_
			if (global_relabel_threshold == num_of_teams_stored) {
				global_relabel_threshold = 0;
				global_relabel(sink);
				sink->height = 0;
			}
#endif
		}
	}
}

//Creates an initial preflow in the flow network
//worst case bound: O(V^3) (rather pessimistic)
void maxflow_initialize_preflow(AL_HEAD *source, AL_HEAD *sink) {
	AL_HEAD *cur_head = receipt_graph;
	AL_NODE *cur_node = NULL;
	AL_NODE *temp_cur_node = NULL;
	long flow = 0;
	int node_count = 0;

	//initialize excess/height of every node to 0
	while (cur_head != NULL) {
		cur_head->excess = 0;
#ifndef _GLOBAL_RELABEL_
		//if NOT defined
		cur_head->height = 0;
		node_count++; //at the same time, count the number of teams(==nodes)
#endif
		cur_head = cur_head->next_head;
	}

#ifndef _GLOBAL_RELABEL_
	//if NOT defined
	source->height = node_count; //set source node height to #nodes
#endif

#ifdef _GLOBAL_RELABEL_
	global_relabel(sink);
#endif
	//now push as much flow as possible to the source's neighbors
	cur_node = source->neighbors;
	while (cur_node != NULL && !cur_node->saturated) { //O(V)
		flow = cur_node->residual_weight; //flow == weight of the edge under consideration in this loop
		AL_NODE  *s_node = NULL;
		if ( (s_node = is_neighbor(cur_node->myhead, source)) == NULL) { //O(V)
			//add new foo edge
			s_node = add_foo_edge(cur_node->myhead, source, flow); //O(V)
		}
		else {
			//just push flow on the opposite direction of s->u
			s_node->residual_weight += flow;
		}
		//keep track of this cur node
		temp_cur_node = cur_node->next;
		//saturate source
		cur_node->saturated = 1; //we cannot remove the edge from the graph so as not to lose data
		//we just mark it as saturated and the algorithm will ignore it (will consider it as
		//not belonging to the residual graph) - also, put sat/ed edge at the end of the adj. list
		adj_list_remove(cur_node);
#ifdef _GLOBAL_RELABEL_
		remove_incoming(cur_node);
		if (!cur_node->myhead->last_incoming) { //empty list
			cur_node->myhead->last_incoming = cur_node->incoming_ptr;
			cur_node->myhead->incoming = cur_node->incoming_ptr;
			cur_node->myhead->incoming->prev = NULL;
			cur_node->myhead->incoming->next = NULL;
		}
		else {
			cur_node->myhead->last_incoming->next = cur_node->incoming_ptr;
			cur_node->incoming_ptr->prev = cur_node->myhead->last_incoming;
			cur_node->incoming_ptr->next = NULL;
			cur_node->myhead->last_incoming = cur_node->incoming_ptr;
		}

#endif
		//append v at the end of list - O(1)
		if (!cur_node->consumer->last_neighbor) { //empty list
			cur_node->consumer->last_neighbor = cur_node;
			cur_node->consumer->neighbors = cur_node;
			cur_node->prev = NULL;
			cur_node->next = NULL;
		}
		else {
			cur_node->consumer->last_neighbor->next = cur_node;
			cur_node->prev = cur_node->consumer->last_neighbor;
			cur_node->next = NULL;
			cur_node->consumer->last_neighbor = cur_node;
		}

		//decr source's excess
		source->excess -= flow;

		//incr cur's excess
		cur_node->myhead->excess = flow;

		//advance pointer
		cur_node = temp_cur_node;
	}
#ifdef _GLOBAL_RELABEL_
	//relabel source
	source->height = num_of_teams_stored;
#endif
}

//this function restores the graph to its original state, that is before calling maxflow_run
//It scans all nodes and edges and removes those with is_foo = 1 and sets edge->residual_weight = edge->weight,
//saturated=0, excess=0, height=0
void maxflow_restore_graph() {
//	printf("Num of nodes: %d\n", num_of_teams_stored);
	AL_HEAD *cur_head = NULL; //current head node
	AL_NODE *cur_node = NULL; //curent neighbor
	AL_NODE *to_be_freed = NULL; //temp pointer to a node to be removed

	sat_push_ops = 0; sat_time = 0;
	non_sat_push_ops = 0; non_sat_time = 0;
	relabel_ops = 0; relabel_time = 0;

	int label = 0;

	cur_head = receipt_graph; //start from the first head node and check 'em all

	cur_head = receipt_graph;
	while (cur_head != NULL) {
		cur_head->height = 0;
		cur_head->excess = 0;
		cur_head->color = MF_NODE_COLOR_WHITE;

#ifdef _USE_HT_
		//free neighbors ht
		lh_free(cur_head->neighbors_ht);
		//reallocate it
		cur_head->neighbors_ht = lh_new(LHASH_HASH_FN(neighbor_hash), LHASH_COMP_FN(neighbor_cmp));
#endif

		cur_node = cur_head->neighbors;
		while (cur_node != NULL) {
			cur_node->saturated = 0;
			//scan list of neighbors to find foo edges and remove them (and free their memory)
			if (cur_node->is_foo) {
				//if the edge is foo ---> remove it from graph
				adj_list_remove(cur_node);
				to_be_freed = cur_node;
#ifdef _GLOBAL_RELABEL_
				remove_incoming(to_be_freed);
				SAFE_FREE(to_be_freed->incoming_ptr);
#endif
				cur_node = cur_node->next;
				SAFE_FREE(to_be_freed);
			}
			else {
				//cur_node->saturated = 0;
				cur_node->residual_weight = cur_node->weight;
				cur_node = cur_node->next;

#ifdef _USE_HT_
				//insert node to neighbors ht
				if (cur_node) {
					lh_insert(cur_head->neighbors_ht, (void*)cur_node);
				}
#else
#ifndef _SLOW_
				if (cur_node) {
					cur_head->neighbors_m[cur_node->myhead->label].removed = 0;
					cur_head->neighbors_m[cur_node->myhead->label].node = cur_node;
				}
#endif
#endif
			}
		}//while (cur_node)
		cur_head = cur_head->next_head;
	}//while (cur_head)
}

//TODO: mallon de xreiazetai ---> prepei na fygei
int has_valid_neighbors(AL_HEAD* u) {
	AL_NODE *cur = u->neighbors;
	while (cur) {
		if (!cur->saturated) {
			return 1;
		}
		cur = cur->next;
	}
	return 0;
}

int fifo_is_empty(FIFO_NODE *head, FIFO_NODE *tail)  {
	if ((head == NULL) || (tail == NULL))
		return 1;
	return 0;
}


//discharge version - TODO
long maxflow_run(AL_HEAD *source, AL_HEAD *sink) {
	struct timeval start, end;

	//initialize queues
	FIFO_NODE *adj_fifo_head, *adj_fifo_tail;
	FIFO_NODE *Q_head, *Q_tail;
	adj_fifo_head = NULL;
	adj_fifo_tail = NULL;
	Q_head = NULL;
	Q_tail = NULL;

	AL_HEAD *u = NULL;

	//restores graph in case it has been modified by a prior call to maxflow_run
	maxflow_restore_graph();

	//push all adjacent nodes of u
	FIFO_NODE *fnode = NULL;
	AL_NODE *to_be_enqueued = source->neighbors;
	while ( (to_be_enqueued != NULL) && (!to_be_enqueued->saturated) ) {
		//*new*: since all saturated edges are placed at the end of the list, when the first sat/ed adge is encountered --> exit loop
		to_be_enqueued->myhead->color = MF_NODE_COLOR_BLACK;
		if (to_be_enqueued->myhead != sink) {
			fnode = (FIFO_NODE*)malloc(sizeof(FIFO_NODE));
			memset(fnode, 0, sizeof(FIFO_NODE));
			fnode->data = (void*)to_be_enqueued->myhead;
			fifo_push(fnode, &Q_head, &Q_tail);
		}
		to_be_enqueued = to_be_enqueued->next;
	}

	//initialize preflow
	maxflow_initialize_preflow(source, sink);

	//main execution loop
	FIFO_NODE *u_fnode = NULL;
	while ( (Q_head != NULL) || (Q_tail != NULL) ) {
		//while !empty(Q)
		u_fnode = fifo_pop(&Q_head, &Q_tail); //pop node from Q
		u = (AL_HEAD*)u_fnode->data;
		SAFE_FREE(u_fnode);
		u->color = MF_NODE_COLOR_WHITE; //and color it white
		//now find the min height of all adjacent nodes of u - O(V)

		maxflow_discharge(u, sink, &Q_head, &Q_tail);
	}
	return sink->excess;
}


void print_graph() {
	AL_NODE *cur_node = NULL;
	AL_HEAD *cur_head = receipt_graph;
	int node_count = 0;
	int edge_count = 0;

	printf("####################################################################\n");
	while (cur_head != NULL) {
		fprintf(stderr, "%d #  ", cur_head->label);
		node_count++; //at the same time, count the number of teams(==nodes)

		//count edges:
		cur_node = cur_head->neighbors;
		while(cur_node) {
			edge_count++;
			fprintf(stderr, "%d(%ld) -> ", cur_node->myhead->label, cur_node->weight);
			cur_node = cur_node->next;
		}
		fprintf(stderr, "NULL\n");
		cur_head = cur_head->next_head;
	}
	fprintf(stderr, "Node count: %d\n", node_count);
	fprintf(stderr, "Edge count: %d\n", edge_count);
	printf("####################################################################\n");
}

#ifdef _GLOBAL_RELABEL_
void print_inverse_graph() {
	INCOMING *cur_node = NULL;
	AL_HEAD *cur_head = receipt_graph;
	int node_count = 0;
	int edge_count = 0;

	fprintf(stderr, "####################################################################\n");
	while (cur_head != NULL) {
		fprintf(stderr, "%d #  ", cur_head->label);
		node_count++; //at the same time, count the number of teams(==nodes)

		//count edges:
		cur_node = cur_head->incoming;
		while(cur_node) {
			edge_count++;
			fprintf(stderr, "%d(%ld) -> ", cur_node->node->myhead->label, cur_node->node->weight);
			cur_node = cur_node->next;
		}
		fprintf(stderr, "NULL\n");
		cur_head = cur_head->next_head;
	}
	fprintf(stderr, "Node count: %d\n", node_count);
	fprintf(stderr, "Edge count: %d\n", edge_count);
	fprintf(stderr, "####################################################################\n");
}
#endif

#define _GNU_SOURCE
#include <stdio.h>

void export_graph(const char *fname) {
	FILE *fp = NULL;

	fp = fopen(fname, "w+");
	if (!fp) {
		fprintf(stderr, "Could not open %s\n", fname);
		return;
	}

	AL_HEAD *cur_head = NULL;
	AL_NODE *cur_node =  NULL;

	//add labels and count nodes
	int nodecount = 0;
	cur_head = receipt_graph;
	while(cur_head) {
		cur_head->label = nodecount;
		nodecount++;
		cur_head = cur_head->next_head;
	}
	fprintf(fp, "%d\n", nodecount);

	cur_head = receipt_graph;
	while(cur_head) {
		fprintf(fp, "%d#", cur_head->label);
		cur_node = cur_head->neighbors;
		while (cur_node) {
			fprintf(fp, " %d(%ld)", cur_node->myhead->label, cur_node->weight);
			cur_node = cur_node->next;
		}
		fprintf(fp, "\n");
		cur_head = cur_head->next_head;
	}

	fclose(fp);
}

AL_HEAD *import_graph(const char *fname) {
	AL_HEAD *graph = NULL;
	AL_NODE *new_node = NULL;
	AL_HEAD *new_head = NULL;
	AL_NODE *prev_node = NULL;
	AL_HEAD *prev_head = NULL;
	int nodecount = 0;

	FILE *fp = NULL;
	char *line = NULL;
	int len = 0;
	int bytes_read = 0;

	int label = 0;
	long weight = 0;
	int pos = 0;

	int i = 0;
	int j = 0;

	if ( !(fp = fopen(fname, "r")) ) {
		fprintf(stderr, "Could not open graph file\n");
		return NULL;
	}

	//read node count
	if ( (bytes_read = getline(&line, &len, fp)) == -1) {
		fprintf(stderr, "Error reading node count\n");
		fclose(fp);
		return NULL;
	}
	else {
		if (sscanf(line, "%d", &nodecount) == 0) {
			fprintf(stderr, "Error reading node count\n");
			fclose(fp);
			return NULL;
		}
	}

	//create list of heads
	for (i=0; i<nodecount; i++) {
		new_head = (AL_HEAD*)malloc(sizeof(AL_HEAD));
		memset(new_head, 0, sizeof(AL_HEAD));
		if (!graph) {
			graph = new_head;
		}
		else {
			prev_head->next_head = new_head; //append to the end of list
		}

		prev_head = new_head;
		new_head->neighbors = NULL;
		new_head->excess = 0;
		new_head->height = 0;
		new_head->color = 0;
		new_head->next_head = NULL;
		new_head->label = i;
		new_head->team = NULL;
	}

	//now start adding edges
	new_head = graph;
	while( (bytes_read = getline(&line, &len, fp)) != -1) {
		bytes_read = sscanf(line, "%d#", &new_head->label); //read (head) node label
		pos += bytes_read;

		//read edges
		while ( (bytes_read = sscanf(line+pos, " %d(%ld)", &label, &weight)) != 0 ) {
			pos += bytes_read;
			new_node = (AL_NODE*)malloc(sizeof(AL_NODE));
			memset(new_node, 0, sizeof(AL_NODE));
			if (!new_head->neighbors) {
				new_head->neighbors = new_node;
			}
			else {
				prev_node->next = new_node;
			}
			prev_node = new_node;

			new_node->consumer = new_head;
			//find edge's "myhead" pointer
			new_node->myhead = graph;
			for (j = 0; j<label; j++) {
				new_node->myhead = new_node->myhead->next_head;
			}
			new_node->weight = weight;
			new_node->saturated = 0;
			new_node->is_foo = 0;
		}
		SAFE_FREE(line);
		new_head = new_head->next_head;
	}

	if (fp) {
		fclose(fp);
	}

	return graph;
}
