/**
* storage.c - Receipt storage routines
*
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "storage.h"

//The functions that follow are not thread-safe
//We suppose that locking has been already performed by other funcs

//finds a free position where a new receipt can be stored (in perm storage)
//if -1 is returned, either an error has occured or no space is left (threshold reached) so the receipt cannot be stored
long pop_free_pos(const char *receipts_db_free_list) {
	FILE *fp = NULL;
	int fd = 0;

	int retval = 0;
	long start_of_file, end_of_file, file_size;

	fd = open(receipts_db_free_list, O_RDWR);
	if (fd == -1) {
		fprintf(stderr, "Error opening free list file\n");
		return (long)FL_ERROR;
	}

	fp = fdopen(fd, "r+");
	if (!fp) {
		fprintf(stderr, "Error opening free list file\n");
		close(fd);
		return (long)FL_ERROR;
	}

	end_of_file = ftell(fp);
	fseek(fp, 0L, SEEK_SET);
	start_of_file = ftell(fp);
	file_size = end_of_file - start_of_file;

	if (file_size == 0) {
		//empty file <==> no free positions
		return (long)FL_EMPTY_FILE;
	}

	//if file is not empty -> move to the last record
	if (fseek(fp, -1L, SEEK_END) == -1) {
		fprintf(stderr, "Failed to get last record\n");
		return (long)FL_ERROR;
	}

	//read the free pos
	if (fread(&retval, sizeof(long), 1, fp) != sizeof(long)) {
		fprintf(stderr, "Failed to read record\n");
		return (long)FL_ERROR;
	}

	//remove last record - truncate file
	ftruncate( fd, end_of_file - start_of_file - sizeof(long) );

	fclose(fp);

	return retval;
}

//writes a new free pos at the end of the free-pos list
int push_free_pos(const char * receipts_db_free_list, long free_pos, long max_receipts) {
	FILE *fp = NULL;

	//check if free_pos is outside the valid range
	if ( (free_pos < 0) && (free_pos > max_receipts) ) {
		fprintf(stderr, "Invalid free_pos value\n");
		return FL_ERROR;
	}

	fp = fopen(receipts_db_free_list, "r+");
	if (!fp) {
		fprintf(stderr, "Error opening free list file\n");
		return FL_ERROR;
	}

	//move past the last record
	fseek(fp, 0L, SEEK_END);

	fwrite(&free_pos, sizeof(long), 1, fp);
	return 0;
}

//gets a list of the free positions in the db
//On error ---> num_of_items = -1
//If the file is empty ---> num_of_items = 0
//(In both above cases NULL is returned, else an array of longs is returned)
long *get_free_list(const char *receipts_db_free_list, int *num_of_items) {
	FILE *fp = NULL;
	long end_of_file, start_of_file, file_size;
	long *retval = NULL;
	int i = 0;

	fp = fopen(receipts_db_free_list, "r");
	if (!fp) {
		fprintf(stderr, "Error opening free list file\n");
		*num_of_items = -1;
		return NULL;
	}

	end_of_file = ftell(fp);
	fseek(fp, 0L, SEEK_SET);
	start_of_file = ftell(fp);
	file_size = end_of_file - start_of_file;

	//if file is empty --> set num_of_items to 0 and return NULL
	if (file_size == 0) {
		*num_of_items = 0;
		fclose(fp);
		return NULL;
	}

	retval = (long*)malloc( file_size/sizeof(long) );
	memset(retval, 0, file_size/sizeof(long) );

	for (i=0; i<file_size/sizeof(long); i++) {
		//read free list (it stores long integers which represent free positions in the db)
		if ( fread(&(retval[i]), 1, sizeof(long), fp) != sizeof(long) ) {
			SAFE_FREE(retval);
			*num_of_items = -1;
			fclose(fp);
			return NULL;
		}
	}
	fclose(fp);
	*num_of_items = file_size/sizeof(long);
	return retval;
}

int is_free_pos(long *free_list, int list_len, long pos) {
	int i = 0;
	for (i=0; i<list_len; i++) {
		if (free_list[i] == pos) {
			return 1;
		}
	}
	return 0;
}
